#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include "dither.h"

void Build555Cube(PALETTEENTRY *p)
{
	int r, g, b;
	for (r=0; (r < 5); r++) {
		for (g=0; (g < 5); g++) {
			for (b=0; (b < 5); b++) {
				p->peRed = 255 * r / 4;
				p->peGreen = 255 * g / 4;
				p->peBlue = 255 * b / 4;
				p->peFlags = 0;
				p++;
			}
		}
	}
}

void BuildSyspCube(PALETTEENTRY *sysp, int syspCube[4][4][4])
{
	int r, g, b;
	for (r = 0; (r < 4); r++) {
		for (g = 0; (g < 4); g++) {
			for (b = 0; (b < 4); b++) {
        		int c;
				int mind;
				int best;
				int rpix = (r << 6) + (r << 4) + (r << 2) + r;
				int gpix = (g << 6) + (g << 4) + (g << 2) + g;
				int bpix = (b << 6) + (b << 4) + (b << 2) + b; 	
				for (c=0; (c < 16); c++) {
					int d = abs(sysp[c].peRed - rpix) +
					 	    abs(sysp[c].peGreen - gpix) +
					 	    abs(sysp[c].peBlue - bpix);
					if ((!c) || (d < mind)) {
						mind = d;
						best = c;
					}
				}
				syspCube[r][g][b] = best;
			}
		}
	}				
}

void DitherRowTruecolorRGB(
	BYTE *dibBitsC, BYTE *scanlineC, unsigned int i)
{
	while (i) {
		/* Blue, Green, Red. */
		dibBitsC[2] = scanlineC[0];
		dibBitsC[1] = scanlineC[1];
		dibBitsC[0] = scanlineC[2];
		dibBitsC += 3;
		scanlineC += 3;
		i--;
	}
}

void DitherRowTruecolorGray(
	BYTE *dibBitsC, BYTE *scanlineC, unsigned int i)
{
	while (i) {
		/* Grayscale */
		*dibBitsC = *scanlineC;
		dibBitsC++;
		*dibBitsC = *scanlineC;
		dibBitsC++;
		*dibBitsC = *scanlineC;
		dibBitsC++;
		scanlineC++;
		i--;
	}
}

void DitherRowTruecolorPalette(
	BYTE *dibBitsC, BYTE *scanlineC, unsigned int w, 
	PALETTEENTRY *sourcep)
{
	while (w) {
		int ind = *scanlineC;
		*dibBitsC = sourcep[ind].peBlue;
		dibBitsC ++;
		*dibBitsC = sourcep[ind].peGreen;
		dibBitsC ++;
		*dibBitsC = sourcep[ind].peRed;
		dibBitsC ++;
		scanlineC ++;
		w--;
	}
}

void DitherRow555RGB(
	BYTE *dibBitsC, BYTE *scanlineC, unsigned int i)
{
	int r = rand() & 255, g = rand() & 255, b = rand() & 255;
	int rg, gg, bg;
	while (i) {
		r += scanlineC[0];
		g += scanlineC[1];
		b += scanlineC[2];
		rg = (r + 1) >> 6;
		if (rg > 4) {
			rg = 4;
		}
		gg = (g + 1) >> 6;
		if (gg > 4) {
			gg = 4;
		}
		bg = (b + 1) >> 6;
		if (bg > 4) {
			bg = 4;
		}
		*dibBitsC = rg * 25 + gg * 5 + bg;
		rg <<= 6;
		gg <<= 6;
		bg <<= 6;
		r -= rg;
		g -= gg;
		b -= bg;
		dibBitsC ++;
		scanlineC += 3;
		i--;
	}
}

void DitherRowPalettePalette(
	BYTE *dibBitsC, BYTE *scanlineC, unsigned int i)
{
	/* This is the easy case. */
	memcpy(dibBitsC, scanlineC, i);
}

void DitherRow555Gray(
	BYTE *dibBitsC, BYTE *scanlineC, unsigned int i)
{
	int r = rand() & 255, g = rand() & 255, b = rand() & 255;
	int rg;

	while (i) {
		/* Just one channel. */
		r += *scanlineC;
		rg = (r + 1) >> 6;
		if (rg > 4) {
			rg = 4;
		}
		*dibBitsC = rg * 25 + rg * 5 + rg;
		rg <<= 6;
		r -= rg;
		dibBitsC ++;
		scanlineC ++;
		i--;
	}
}

void DitherRowSyspPalette(
	BYTE *dibBitsC, BYTE *scanlineC, unsigned int w,
	PALETTEENTRY *sourcep, int syspCube[4][4][4],
	PALETTEENTRY *sysp)
{
	int x;
	int r = rand() & 255, g = rand() & 255, b = rand() & 255;
	for (x = 0; (x < w); x++) {
		int i;
		int rc, gc, bc;
		r += sourcep[*scanlineC].peRed;
		g += sourcep[*scanlineC].peGreen;
		b += sourcep[*scanlineC].peBlue;
		scanlineC++;
		if (r < 0) {
			rc = 0;
		} else if (r > 255) {
			rc = 3;
		} else {
			rc = r >> 6;
		}
		if (g < 0) {
			gc = 0;
		} else if (g > 255) {
			gc = 3;
		} else {
			gc = g >> 6;
		}
		if (b < 0) {
			bc = 0;
		} else if (b > 255) {
			bc = 3;
		} else {
			bc = b >> 6;
		}
		i = syspCube[rc][gc][bc];
		r -= (dibBitsC[2] = sysp[i].peRed); 	    
		g -= (dibBitsC[1] = sysp[i].peGreen); 	    
		b -= (dibBitsC[0] = sysp[i].peBlue); 	    
		dibBitsC += 3;
	}
}

void DitherRowSyspRGB(
	BYTE *dibBitsC, BYTE *scanlineC, unsigned int w,
	int syspCube[4][4][4],
	PALETTEENTRY *sysp)
{
	int x;
	int r = rand() & 255, g = rand() & 255, b = rand() & 255;
	for (x = 0; (x < w); x++) {
		int i;
		int rc, gc, bc;
		r += *scanlineC;
		scanlineC++;
		g += *scanlineC;
		scanlineC++;
		b += *scanlineC;
		scanlineC++;
		if (r < 0) {
			rc = 0;
		} else if (r > 255) {
			rc = 3;
		} else {
			rc = r >> 6;
		}
		if (g < 0) {
			gc = 0;
		} else if (g > 255) {
			gc = 3;
		} else {
			gc = g >> 6;
		}
		if (b < 0) {
			bc = 0;
		} else if (b > 255) {
			bc = 3;
		} else {
			bc = b >> 6;
		}
		i = syspCube[rc][gc][bc];
		r -= (dibBitsC[2] = sysp[i].peRed); 	    
		g -= (dibBitsC[1] = sysp[i].peGreen); 	    
		b -= (dibBitsC[0] = sysp[i].peBlue); 	    
		dibBitsC += 3;
	}
}

void DitherRowSyspGray(
	BYTE *dibBitsC, BYTE *scanlineC, unsigned int w,
	int syspCube[4][4][4],
	PALETTEENTRY *sysp)
{
	int x;
	int r = rand() & 255, g = rand() & 255, b = rand() & 255;
	for (x = 0; (x < w); x++) {
		int i;
		int rc, gc, bc;
		r += *scanlineC;
		g += *scanlineC;
		b += *scanlineC;
		scanlineC++;
		if (r < 0) {
			rc = 0;
		} else if (r > 255) {
			rc = 3;
		} else {
			rc = r >> 6;
		}
		if (g < 0) {
			gc = 0;
		} else if (g > 255) {
			gc = 3;
		} else {
			gc = g >> 6;
		}
		if (b < 0) {
			bc = 0;
		} else if (b > 255) {
			bc = 3;
		} else {
			bc = b >> 6;
		}
		i = syspCube[rc][gc][bc];
		r -= (dibBitsC[2] = sysp[i].peRed); 	    
		g -= (dibBitsC[1] = sysp[i].peGreen); 	    
		b -= (dibBitsC[0] = sysp[i].peBlue); 	    
		dibBitsC += 3;
	}
}
