#include "gui.h"
#include "guim.h"
#include "guit.h"
#include "guib.h"
#include "guif.h"

int testEventHandler(guiCanvasPtr c, guiEventPtr e);
int frameEventHandler(guiCanvasPtr c, guiEventPtr e);

guiSize upperbars;
guiCanvasPtr frame, canvas;
guiTextPtr text, line;
guiMenuItemPtr saveItem;

void menuAction(char *text, void *clientData);
void saveAction(char *text, void *clientData);
void quitHandler(char *text, void *clientData);
void openHandler(char *text, void *clientData);

extern unsigned char opossum[];

int main(int argc, char *argv[])
{
	guiSize s, bs;
	guiMenuPtr m;
	guiMenuDataPtr file;
	guiMenuDataPtr edit;
	guiMenuDataPtr help;
	guiMenuItemPtr item;
	guiButtonPtr button, leader;
	guiBitmapPtr icon;
	int half;
	if (!guiSetup()) {
		fprintf(stderr, "guiSetup() failed.\n");
		return 1;
	}
	s = guiGetScreenSize();
	frame = guiCanvasCreate(0, s.w / 6, s.h / 6, s.w / 3 * 2, s.h / 3 * 2);
	guiCanvasSetName(frame, "libgui test application");
	icon = guiBitmapBuiltin(opossum, 1);
	guiCanvasSetIcon(frame, icon);
	guiCanvasSetEventHandler(frame, frameEventHandler);
	guiCanvasSetEventMask(frame, guiEventResize);
	s = guiCanvasGetPhysicalSize(frame);
	m = guiMenuCreate(frame, 0, 0);
	file = guiMenuAdd(m, "&File");
	guiMenuAddItem(file, "&New", (void *) 0, menuAction);	
	guiMenuAddItem(file, "&Open", (void *) 1, openHandler);
	saveItem = guiMenuAddItem(file, "&Save", (void *) 2, saveAction);	
	guiMenuAddItem(file, "Save &As", (void *) 3, menuAction);	
	guiMenuAddItem(file, "E&xit", 0, quitHandler);	
	edit = guiMenuAdd(m, "&Edit");
	guiMenuAddItem(edit, "&Cut", (void *) 0, menuAction);	
	guiMenuAddItem(edit, "C&opy", (void *) 1, menuAction);	
	item = guiMenuAddItem(edit, "&Paste", (void *) 2, menuAction);	
	guiMenuItemSetActive(item, 0);
	item = guiMenuAddItem(edit, "&Undo", (void *) 3, menuAction);	
	guiMenuItemSetActive(item, 0);
	help = guiMenuAdd(m, "&Help");
	guiMenuAddItem(help, "&Contents", (void *) 0, menuAction);	
	guiMenuAddItem(help, "&Search", (void *) 1, menuAction);	
	upperbars = guiMenuGetPhysicalSize(m);
	button = guiButtonCreate(frame, 0, upperbars.h, 
		"Go", guiButtonSimple);
	guiButtonShow(button, 1);
	bs = guiButtonGetPhysicalSize(button);
	bs.w += 16;
	button = guiButtonCreate(frame, bs.w, 
		upperbars.h, "Slow", guiButtonCheck);
	guiButtonShow(button, 1);
	bs.w += guiButtonGetPhysicalSize(button).w;
	bs.w += 16;
	leader = guiButtonCreate(frame, bs.w, 
		upperbars.h, "One", guiButtonRadio);
	guiButtonShow(leader, 1);
	guiButtonSetLeader(leader, leader);
	bs.w += guiButtonGetPhysicalSize(leader).w;
	bs.w += 16;
	button = guiButtonCreate(frame, bs.w, 
		upperbars.h, "Two", guiButtonRadio);
	guiButtonShow(button, 1);
	guiButtonSetLeader(button, leader);
	bs.w += guiButtonGetPhysicalSize(leader).w;
	bs.w += 16;
	button = guiButtonCreate(frame, bs.w, 
		upperbars.h, "Three", guiButtonRadio);
	guiButtonShow(button, 1);
	guiButtonSetLeader(button, leader);
	upperbars.h += guiButtonEstimate("", guiButtonSimple).h;
	half = (s.h - upperbars.h) / 2;
	canvas = guiCanvasCreate(frame,
		0, upperbars.h,
		s.w, half);
	guiCanvasSetBorder(canvas, 2, guiBorderIn);
	line = guiTextCreateSingle(frame,
		0, upperbars.h + half,
		s.w,
		guiTextEstimateSingle(1).h, guiNone);
	text = guiTextCreateMultiple(frame, 
		0, upperbars.h + half + guiTextGetPhysicalSize(line).h,
		s.w, s.h - half - upperbars.h - guiTextGetPhysicalSize(line).h,
		120);
	s = guiCanvasGetPhysicalSize(canvas);
	guiCanvasSetVirtualSize(canvas, 
		s.w * 2, s.h * 2);
	guiCanvasSetEventMask(canvas, guiEventRepaint | guiEventKeyDown); 
	guiCanvasSetEventHandler(canvas, testEventHandler);
	guiCanvasSetTabbing(frame, 1);
	guiCanvasShow(canvas, 1);
	guiMenuShow(m, 1);
	guiCanvasShow(frame, 1);
	guiTextShow(text, 1);
	guiTextShow(line, 1);
	guiCanvasSetTimer(canvas, 0, 5000);
	guiMainLoop();
	return 0;
}

guiColor color = { 255, 255, 255 };

int testEventHandler(guiCanvasPtr c, guiEventPtr e)
{
	switch(e->t) {
		case guiEventRepaint:
		{
			guiSize s = guiCanvasGetVirtualSize(c);
			guiCanvasSetDrawColor(c, color);
			guiCanvasDrawLine(c, 0, 0, s.w - 1, s.h - 1);
			guiCanvasDrawLine(c, s.w - 1, 0, 0, s.h - 1); 
			break;
		}
		case guiEventKeyDown:
		{
			printf("%c %d %d %d %d\n", 
				e->key, e->key,
				e->shift, e->control, e->alt);
			fflush(stdout);
			break;
		}
		case guiEventTimer:
		{
			color.r = rand() & 255;	
			color.g = rand() & 255;	
			color.b = rand() & 255;	
			guiCanvasSetTimer(c, 0, 5000);
			guiCanvasInvalidate(c, 0);
			break;
		}
		default:
		break;
	}	
	/* Don't block the default behaviors, which are useful */
	return 0;
}

int frameEventHandler(guiCanvasPtr c, guiEventPtr e)
{
	int half;
	switch(e->t) {
		case guiEventResize:
		half = (e->s.h - upperbars.h) / 2;
		guiCanvasSetPhysicalSize(canvas,
			e->s.w, half);
		guiTextMove(line,
			0, half + upperbars.h);
		guiTextSetPhysicalSize(line,
			e->s.w, guiTextGetPhysicalSize(line).h);
		guiTextMove(text,
			0, half + upperbars.h + guiTextGetPhysicalSize(line).h);
		guiTextSetPhysicalSize(text,
			e->s.w, e->s.h - upperbars.h - half - 
			guiTextGetPhysicalSize(line).h);
		break;
		default:
		break;
	}
	return 0;
}

void menuAction(char *text, void *clientData)
{
	printf("Text: %s Client Data: %d\n", text, (int) clientData);
	printf("Toggling active state of text line (just for laughs)\n");
	guiTextSetActive(line, !guiTextGetActive(line));
}

void saveAction(char *text, void *clientData)
{
	guiMenuItemSetCheck(saveItem, !guiMenuItemGetCheck(saveItem));
}

void openHandler(char *text, void *clientData)
{
	guiFileOptions options;
	options.from = frame;
	options.initialFilename = "";
	options.filter = "*.gif|*.jpg|*.jpeg|*.png|*.GIF|*.JPG|*.JPEG|*.PNG";
	options.title = "Select Image to Map";
	options.label = "Please select a GIF, JPEG, or PNG format image.";
	options.helpHandler = 0;
	options.mustExist = 1;
	if (guiFileSelect(&options) != guiFileCanceled) {
		fprintf(stderr, "Result: %s\n", options.outputFilename);
		free(options.outputFilename);
	}
}

void quitHandler(char *text, void *clientData)
{
	int result = guiMessage(frame, "Exit the Program",
		"There was a time when your mother wouldn't let "
		"you have any candy. I'm going to remind you of this "
		"in order to occupy space and upset you.\n\n"
		"Do you want to save your work first?",
		guiMessageYesNoCancel);
	printf("Result returned: %d\n", result);
	if (result == guiMessageCancel) {
		printf("Exit cancelled\n");
		return;
	} 
	printf("Exit chosen from the file menu\n");
	guiShutdown();
}

