#include "gui.h"
#include "guix.h"
#include "guixb.h"
#include "imaget.h"
#include "fonteq.h"
#include "guib.h"
#include "guila.h"
#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>

guiRectangle guiRectangleEmpty = { 0, 0, 0, 0 };

int guiXDone = 0;

Display *guiXDisplay;
int guiXScreen;
int guiXDepth;
XContext guiXContext;
Visual *guiXVisual;
Window guiXLeader;
guiCanvasPtr guiXLastCanvas;
guiCanvasPtr guiXMouseDownCanvas = 0;
int guiXButtonsDown = 0;
GC guiXDrawGC;
GC guiXFillGC;
GC guiXScrollbarBackgroundGC;
GC guiXScrollbarBodyGC;
GC guiXBorderHighGC;
GC guiXBorderLowGC;
XPoint *guiXPoints;
int guiXPointsAllocated;
guiCanvasPtr guiXFocusCanvas;

guiColor guiDefaultBackgroundColor = { 192, 192, 192 };

guiColor guiScrollbarBodyColor = { 64, 64, 220 };

guiColor guiScrollbarBackgroundColor = { 82, 148, 189 }; 

guiColor guiBorderHighColor = { 227, 227, 227 };

guiColor guiBorderLowColor = { 101, 101, 101 };

guiBitmapPtr guiXArrowBitmaps[2][4];
guiBitmapPtr guiXThumbBitmaps[2][3][2];
guiBitmapPtr guiXDingusBitmap;

guiCanvasPtr guiXArrows[2][4];
guiCanvasPtr guiXThumbs[2][3][2];
guiCanvasPtr guiXDingus;

typedef struct {
	guiCanvasPtr canvas;
	int id;
	long delay;
} guiXTimer, *guiXTimerPtr;

guiXTimerPtr guiXTimers;
int guiXTimersAllocated;
int guiXTimersTotal;

typedef struct guiXKeyGrabStruct {
	guiCanvasPtr c;
	guiKey k;
	int shift;
	int control;
	int alt;
	struct guiXKeyGrabStruct *next;
} guiXKeyGrab, *guiXKeyGrabPtr;

guiXKeyGrabPtr guiXFirstGrab = 0;

typedef struct guiXModalItemStruct {
	guiCanvasPtr c;
	struct guiXModalItemStruct *next;
} guiXModalItem, *guiXModalItemPtr;

guiXModalItemPtr guiXModalFirst = 0;

int guiXModalDepth;
	
char *guiXPSFonts[guiFontFamiliesTotal][2][2] = {
	{ 
		{ "Courier", "Courier-Bold" }, 
		{ "Courier-Oblique", "Courier-BoldOblique" }
	},
	{ 
		{ "Times-Roman", "Times-Bold" }, 
		{ "Times-Italic", "Times-BoldItalic" }
	},
	{
		{ "Helvetica", "Helvetica-Bold" }, 
		{ "Helvetica-Oblique", "Helvetica-BoldOblique" }
	}
};		

char *guiXFonts[guiFontFamiliesTotal][2][2] = {
	{ 
		{ "-adobe-courier-medium-r", "-adobe-courier-bold-r" }, 
		{ "-adobe-courier-medium-o", "-adobe-courier-bold-o" }, 
	},
	{ 
		{ "-adobe-times-medium-r", "-adobe-times-bold-r" }, 
		{ "-adobe-times-medium-i", "-adobe-times-bold-i" }, 
	},
	{
		{ "-adobe-helvetica-medium-r", "-adobe-helvetica-bold-r" }, 
		{ "-adobe-helvetica-medium-o", "-adobe-helvetica-bold-o" }, 
	}
};	

#define guiXFontCacheSize 8

typedef struct {
	XFontStruct *data;	
	guiFont spec;
	guiFont actual;
	int used;
} guiXFontCacheEntry;

guiXFontCacheEntry guiXFontCache[guiXFontCacheSize];

int guiXToGray(unsigned long xp);

typedef enum {
	guiXTrueColor,
	guiXCubeColor,
	guiXGrayScale
} guiXColorTypes;

guiXColorTypes guiXColorType;

/* Conversion tables for pseudocolor visuals. */

int guiXToGrayTable[256];
guiColor guiXCubeToColorTable[8][8][8];
unsigned long guiXCubeToIndexTable[8][8][8];
guiColor guiXGrayToColorTable[256];
unsigned long guiXGrayToIndexTable[256];

unsigned long guiXRedMask;
unsigned long guiXGreenMask;
unsigned long guiXBlueMask;
int guiXShiftR;
int guiXBitsR;
int guiXShiftG;
int guiXBitsG;
int guiXShiftB;
int guiXBitsB;

static void guiXDrawFrame(guiCanvasPtr c, 
	int state, int x, int y, int w, int h);

int guiXFindShiftAndBits(unsigned long mask, int *shiftP, int *bitsP);

void guiXCanvasSetGC(guiCanvasPtr c);
void guiXCanvasPSDraw(guiCanvasPtr c);
void guiXCanvasPSFill(guiCanvasPtr c);
void guiXCanvasInvalidateBody(guiCanvasPtr c, guiRectanglePtr r, 
	int repaintNow);
guiSize guiCanvasGetVisibleSize(guiCanvasPtr c);
int guiXCanvasSetTimerBody(guiCanvasPtr c, int id, long milliseconds);
void guiXCanvasScrollbarVerticalMouseDown(guiCanvasPtr c, int x, int y);
void guiXCanvasScrollbarHorizontalMouseDown(guiCanvasPtr c, int x, int y);
void guiXCanvasScrollbarMouseUp();
void guiXCanvasScrollbarMouseDrag(int x, int y);
void guiXCanvasScrollbarAct();
void guiXCanvasFirstFocus(guiCanvasPtr c);

guiCanvasPtr guiXScrollbarButtonCanvas;

/* Checks whether scrollbars are needed */
void guiXCanvasCheckScrollbars(guiCanvasPtr c);

guiCanvasPtr guiXCanvasGetTopParent(guiCanvasPtr c);
guiCanvasPtr guiXCanvasGetRootBenefactor(guiCanvasPtr c);
int guiXCanvasResolveKey(guiCanvasPtr *cp, guiEventPtr e);

guiColor guiXColorAverage(guiColor a, guiColor b);

static void canvasDestroyed(guiCanvasPtr c);

/* Internal timer for scrollbars */ 

#define guiXScrollbarTimer (-1)

enum {
	guiXUpArrow,
	guiXDownArrow,
	guiXPageUp,
	guiXPageDown,
	guiXVerticalThumb,
	guiXLeftArrow,
	guiXRightArrow,
	guiXPageLeft,
	guiXPageRight,
	guiXHorizontalThumb
};
 
int guiXScrollbarButtonType;
int guiXScrollbarDragOffset;

int guiCanvasDefaultEventHandler(guiCanvasPtr c, guiEventPtr e)
{
	return 0;
}

guiCanvasPtr guiCanvasCreate(guiCanvasPtr parent, int x, int y, int w, int h)
{
	guiCanvasPtr c = (guiCanvasPtr) malloc(sizeof(guiCanvas));
       	guiColor bt;
	guiCanvasPtr p;
	XWMHints xwmh;
	XGCValues gcvalues;
	if (!c) {
		return 0;
	}
	c->parent = parent;
	if (parent) {
		p = parent->first;
		if (p) {
			while (p->next) {
				p = p->next;
			}
			p->next = c;
		} else {
			parent->first = c;
		}	
	}
	c->next = 0;
	c->first = 0;	
	c->childFocused = 0;
	c->l.x = x;
	c->l.y = y;
	if (c->parent) {
		c->l.x = x + c->parent->border;
		c->l.y = y + c->parent->border;
	}	
	c->physicalSize.w = w;
	c->physicalSize.h = h;
	c->virtualSize.w = w;
	c->virtualSize.h = h;
	c->visibleTop.x = 0;
	c->visibleTop.y = 0;
	c->drawColor = guiDefaultBackgroundColor;
	c->fillColor = guiDefaultBackgroundColor;
	c->textBackgroundColor = guiDefaultBackgroundColor;
	c->barV = 0;
	c->barH = 0;
	c->repaintV = 0;
	c->repaintH = 0;
	c->repaintB = 0;
	c->noScrollV = 0;
	c->noScrollH = 0;
	c->visible = 0;
	c->repainting = 0;
	c->eventHandler = guiCanvasDefaultEventHandler;
	c->invalid = guiRectangleEmpty;
	c->offscreen = 0;
	c->psoutput = 0;
	c->pslastfill = 0;
	c->clientData = 0;
	c->eventMask = 0;
	c->virtualized = 0;
	c->fontSet = 0;
	c->border = 0;
	c->clearing = 1;
	c->wantFocus = 0;
	c->onBehalfOf = 0;
	c->focused = 0;
	c->icon = 0;
	c->dying = 0;
	bt = guiDefaultBackgroundColor;
	c->handle = (guiNativeHandle) XCreateSimpleWindow(
		guiXDisplay,
		parent ? (Window) parent->handle :
			DefaultRootWindow(guiXDisplay),
		c->l.x, c->l.y, w, h, 0,
		WhitePixel(guiXDisplay, guiXScreen),
        	guiXColorGet(&bt));
	c->backgroundColor = bt;
	c->drawContext = (guiNativeHandle) XCreateGC(
		guiXDisplay, 
		(Window) c->handle,
		0, 
		&gcvalues);
	XCopyGC(guiXDisplay, guiXDrawGC, ~0, (GC) c->drawContext);
	c->fillContext = (guiNativeHandle) XCreateGC(
		guiXDisplay, 
		(Window) c->handle,
		0, 
		&gcvalues);
	XCopyGC(guiXDisplay, guiXFillGC, ~0, (GC) c->fillContext);
        /* Tell window manager we need the focus; indicate the group leader */
	if (!parent) {
		xwmh.flags = (InputHint | StateHint | WindowGroupHint);
		xwmh.input = True;
		xwmh.initial_state = NormalState;
		xwmh.window_group = guiXLeader;
		XSetWMHints(guiXDisplay, (Window) c->handle, &xwmh);
	}
        /* Save the context pointer */
        XSaveContext(guiXDisplay, (Window) c->handle,
                guiXContext, (caddr_t) c);
	/* Set the minimal event mask (including events needed
		by the infrastructure...) */
	guiCanvasSetEventMask(c, 0);
	c->t = guiCanvasApplication;
	c->dim = 0;
	c->tabbing = 0;
	return c;	
}

void guiCanvasSetBorder(guiCanvasPtr c, int b, guiBorderStyle style)
{
	c->border = b;
	c->borderStyle = style;
	c->repaintB = 1;
	guiCanvasInvalidate(c, 0);
}

void guiCanvasSetType(guiCanvasPtr c, guiCanvasType t, guiCanvasPtr onBehalfOf)
{
	XSetWindowAttributes xswa;
	Atom ATOM_WM_TRANSIENT_FOR;
	Window w;
	if (c->parent) {
		/* Can't change set the type of a child canvas */
		return;
	}
	if (c->t == t) {
		return;
	}
	ATOM_WM_TRANSIENT_FOR = XInternAtom(guiXDisplay, 
		"WM_TRANSIENT_FOR", False);
	switch(t) {
		case guiCanvasApplication:
		xswa.override_redirect = False;
		xswa.save_under = False;	
		XChangeWindowAttributes(guiXDisplay, (Window) c->handle, 
			CWSaveUnder | CWOverrideRedirect, &xswa);
		w = None;
		break;
		case guiCanvasDialog:
		xswa.override_redirect = False;
		xswa.save_under = False;	
		XChangeWindowAttributes(guiXDisplay, (Window) c->handle, 
			CWSaveUnder | CWOverrideRedirect, &xswa);
		if (onBehalfOf) {
			w = (Window) onBehalfOf->handle;
		} else {
			w = None;
		}
		c->onBehalfOf = onBehalfOf;
		break;
		case guiCanvasUndecorated:
		xswa.override_redirect = True;
		xswa.save_under = True;
		XChangeWindowAttributes(guiXDisplay, (Window) c->handle, 
			CWSaveUnder | CWOverrideRedirect, &xswa);
		w = None;
		c->onBehalfOf = onBehalfOf;
		break;
	}
	XChangeProperty(guiXDisplay, (Window) c->handle,
		ATOM_WM_TRANSIENT_FOR, XA_WINDOW, 32,
		PropModeReplace, 
		(unsigned char *) &w, 1);
	c->t = t;
}

void guiCanvasShow(guiCanvasPtr c, int flag)
{
	guiXModalItemPtr m, prev;
	c->visible = flag;
	if (flag) {
		c->invalid.x = 0;
		c->invalid.y = 0;
		c->invalid.w = c->virtualSize.w;
		c->invalid.h = c->virtualSize.h;
		XRaiseWindow(guiXDisplay, (Window) c->handle);
		XMapWindow(guiXDisplay, (Window) c->handle);
		if (!c->parent) {
			guiXCanvasFirstFocus(c);
		}
	} else {
		guiCanvasPtr md = guiXMouseDownCanvas;
		while (md) {
			if (md == c) {
				/* Fake button releases */
				guiEvent e;
				e.l.x = 0;
				e.l.y = 0;
				e.t = guiEventMouseUp;	
				if (guiXButtonsDown & guiButtonLeft) {
					e.button = guiButtonLeft;
					guiXMouseDownCanvas->
					eventHandler(guiXMouseDownCanvas, &e);
				} 
				if (guiXButtonsDown & guiButtonMiddle) {
					e.button = guiButtonMiddle;
					guiXMouseDownCanvas->
					eventHandler(guiXMouseDownCanvas, &e);
				} 
				if (guiXButtonsDown & guiButtonRight) {
					e.button = guiButtonRight;
					guiXMouseDownCanvas->
					eventHandler(guiXMouseDownCanvas, &e);
				}
				guiXMouseDownCanvas = 0;
				guiXButtonsDown = 0;
				break;
			} else {
				md = md->parent;
			}
		}
		c->invalid.x = 0;
		c->invalid.y = 0;
		c->invalid.w = c->virtualSize.w;
		c->invalid.h = c->virtualSize.h;
		XUnmapWindow(guiXDisplay, (Window) c->handle);
		m = guiXModalFirst;
		prev = 0;
		while (m) {
			if (m->c == c) {
				if (prev) {
					prev->next = m->next;
				} else {
					guiXModalFirst = m->next;						}
				free(m);
				guiXModalDepth--;
				break;
			}
			prev = m;
			m = m->next;		
		}	
	}
}

int guiCanvasIsOffscreen(guiCanvasPtr c)
{
	return c->offscreen;
}

int guiCanvasIsVisible(guiCanvasPtr c)
{
	int visible;
	do {
		visible = c->visible;
		c = c->parent;
	} while(c);		
	return visible;
}

int guiCanvasIsPrinting(guiCanvasPtr c)
{
	if (c->psoutput) {
		return 1;
	}
	return 0;
}

guiCanvasPtr guiCanvasCreateOffscreen(int w, int h)
{
	guiCanvasPtr c = (guiCanvasPtr) malloc(sizeof(guiCanvas));
	XGCValues gcvalues;
	if (!c) {
		return 0;
	}
	c->parent = 0;
	c->first = 0;
	c->childFocused = 0;
	c->next = 0;
	c->tabbing = 0;
	c->l.x = 0;
	c->l.y = 0;
	c->physicalSize.w = w;
	c->physicalSize.h = h;
	c->virtualSize.w = w;
	c->virtualSize.h = h;
	c->visibleTop.x = 0;
	c->visibleTop.y = 0;
	c->barV = 0;
	c->barH = 0;
	c->visible = 0;
	c->eventHandler = guiCanvasDefaultEventHandler;
	c->invalid = guiRectangleEmpty;
	c->offscreen = 1;
	c->psoutput = 0;
	c->border = 0;
	c->noScrollV = 0;
	c->noScrollH = 0;
	c->clearing = 1;
	c->wantFocus = 0;
	c->onBehalfOf = 0;
	c->focused = 0;
	c->icon = 0;
	c->dying = 0;
	c->t = guiCanvasApplication; /* Not really */
	c->drawColor = guiDefaultBackgroundColor;
	c->fillColor = guiDefaultBackgroundColor;
	c->textBackgroundColor = guiDefaultBackgroundColor;
	c->handle = (guiNativeHandle) XCreatePixmap(guiXDisplay,
		guiXLeader, w, h, guiXDepth);
	c->drawContext = (guiNativeHandle) XCreateGC(
		guiXDisplay, 
		(Window) c->handle,
		0, 
		&gcvalues);
	XCopyGC(guiXDisplay, guiXDrawGC, ~0, (GC) c->drawContext);
	c->fillContext = (guiNativeHandle) XCreateGC(
		guiXDisplay, 
		(Window) c->handle,
		0, 
		&gcvalues);
	XCopyGC(guiXDisplay, guiXFillGC, ~0, (GC) c->fillContext);	
	c->dim = 0;
	return c;
}

static void canvasMarkChildrenDying(guiCanvasPtr c);

void guiCanvasDestroy(guiCanvasPtr c)
{
	if (c->offscreen) {
		XFreePixmap(guiXDisplay, (Pixmap) c->handle);
		canvasDestroyed(c);
	} else {
		c->dying = 1;
		canvasMarkChildrenDying(c);
		XDestroyWindow(guiXDisplay, (Window) c->handle);
	}
}

static void canvasMarkChildrenDying(guiCanvasPtr c)
{
	guiCanvasPtr p;
	p = c->first;
	while (p) {
		p->dying = 1;
		canvasMarkChildrenDying(p);
		p = p->next;
	}	
}

void guiCanvasSetEventHandler(guiCanvasPtr c,
        int (*eventHandlerArg) (guiCanvasPtr c, guiEventPtr e))
{
	if (!eventHandlerArg) {
		c->eventHandler = guiCanvasDefaultEventHandler;
	} else {
		c->eventHandler = eventHandlerArg;
	}
}

void guiCanvasSetEventMask(guiCanvasPtr c, guiEventType t)
{
	long xmask = 0;
	c->eventMask = t;
	/* Always set the masks necessary for scrollbars and
		other infrastructure features to work. */
	xmask += PointerMotionMask;
	xmask += ButtonPressMask;
	xmask += ButtonReleaseMask;
	xmask += ExposureMask;
	xmask += StructureNotifyMask;
	xmask += KeyPressMask;
	xmask += FocusChangeMask;
	/* We don't subscribe to the "always grab the pointer on
		button press" school of thought. Instead, we ensure
		that both the owner and the canvas that originally
		saw the button press get the release event. This 
		means that MouseUp events can certainly happen
		without corresponding MouseDown events. It is also
		a really fab simplification. */
	xmask += OwnerGrabButtonMask;
	if (c->offscreen) {
		/* No events for offscreen canvases */
		return;
	}
	if (t & guiEventMouseDrag) {
		xmask += ButtonMotionMask;
	}
	if (t & guiEventMouseEnter) {
		xmask += EnterWindowMask;
	}
	if (t & guiEventMouseExit) {
		xmask += LeaveWindowMask;
	}
	XSelectInput(guiXDisplay, (Window) c->handle,
		xmask);
}

int guiCanvasSetTimer(guiCanvasPtr c, int id, long milliseconds)
{
	if (id < 0) {
		/* Negative timer ids are reserved for internal use. */
		return 0;
	}
	return guiXCanvasSetTimerBody(c, id, milliseconds);
}

int guiXCanvasSetTimerBody(guiCanvasPtr c, int id, long milliseconds)
{
	int i;
	int slot = -1;
	for (i = 0; (i < guiXTimersTotal); i++) {
		if (!guiXTimers[i].canvas) {
			slot = i;
			break;
		}
		/* If that id is already in use for this canvas,
			reset it to avoid race conditions. */
		if ((guiXTimers[i].canvas == c) && (guiXTimers[i].id == id)) {
			slot = i;
			break;
		}
	}		
	if (slot == -1) {
		if (guiXTimersTotal == guiXTimersAllocated) {
			guiXTimerPtr list;
			int allocated;
			allocated = guiXTimersAllocated * 2;
			list = (guiXTimerPtr) realloc(guiXTimers,
				sizeof(guiXTimer) * guiXTimersAllocated);
			if (!list) {
				return 0;
			}
			guiXTimersAllocated = allocated;
			guiXTimers = list;	
		}
		slot = guiXTimersTotal;
		guiXTimersTotal++;
	}
	guiXTimers[slot].canvas = c;
	guiXTimers[slot].id = id;
	guiXTimers[slot].delay = milliseconds;
	return 1;
}

void guiCanvasClearTimer(guiCanvasPtr c, int id)
{
	int i;
	for (i = 0; (i < guiXTimersTotal); i++) {
		if ((guiXTimers[i].canvas == c) &&
			(guiXTimers[i].id == id)) {
			guiXTimers[i].canvas = 0;
			break;
		}
	}		
}

void guiCanvasMove(guiCanvasPtr c, int x, int y)
{
	if (c->offscreen) {
		return;
	}
	XMoveWindow(guiXDisplay,
		(Window) c->handle, x, y);
}

void guiCanvasSetSize(guiCanvasPtr c, int w, int h)
{
	if (c->offscreen) {
		return;
	}
	c->virtualized = 0;
	c->virtualSize.w = w;	
	c->virtualSize.h = h;	
	XResizeWindow(guiXDisplay,
		(Window) c->handle, w, h);
}

void guiCanvasSetVirtualSize(guiCanvasPtr c, int w, int h)
{
	guiSize s;
	if (c->offscreen) {
		return;
	}
	c->virtualized = 1;
	c->virtualSize.w = w;
	c->virtualSize.h = h;	
	c->barH = 0;
	c->barV = 0;
	guiXCanvasCheckScrollbars(c);
	s = guiCanvasGetVisibleSize(c);
	if (c->visibleTop.x + s.w > c->virtualSize.w) {
		c->visibleTop.x = c->virtualSize.w - s.w;
	}
	c->visibleTop.x = guiMax(c->visibleTop.x, 0);
	if (c->visibleTop.y + s.h > c->virtualSize.h) {
		c->visibleTop.y = c->virtualSize.h - s.h;
	}
	c->visibleTop.y = guiMax(c->visibleTop.y, 0);
	if (guiCanvasIsVisible(c)) {
		guiCanvasInvalidate(c, 0);
	}
}

void guiXCanvasCheckScrollbars(guiCanvasPtr c)
{
	guiSize s;
	if (!c->virtualized) {
		return;
	}
	c->barH = 0;
	c->barV = 0;	
	s = guiCanvasGetVisibleSize(c);
	if ((c->virtualSize.w > s.w) && (!c->noScrollH)) {
		c->barH = 1;
	} else {
		c->barH = 0;
	}
	if ((c->virtualSize.h > s.h) && (!c->noScrollV)) {
		c->barV = 1;
	} else {
		c->barV = 0;
	}
	/* Now see if the addition of one scrollbar
		forces the addition of the other */
	if ((c->virtualSize.w > s.w) && (!c->noScrollH)) {
		c->barH = 1;
	} else {
		c->barH = 0;
	}
	if ((c->virtualSize.h > s.h) && (!c->noScrollV)) {
		c->barV = 1;
	} else {
		c->barV = 0;
	}	
}

void guiCanvasSetPhysicalSize(guiCanvasPtr c, int w, int h)
{
	if (c->offscreen) {
		return;
	}
	c->virtualized = 1;
	XResizeWindow(guiXDisplay,
		(Window) c->handle, w, h);
}

guiPoint guiCanvasGetPosition(guiCanvasPtr c)
{
	return c->l;
}

guiPoint guiCanvasGetScreenPosition(guiCanvasPtr c)
{
	guiPoint p;
	Window dummy;
	XTranslateCoordinates(guiXDisplay, (Window) c->handle,
		DefaultRootWindow(guiXDisplay), 0, 0,
			&p.x, &p.y, &dummy);
	return p;
}

guiSize guiCanvasGetVirtualSize(guiCanvasPtr c)
{
	return c->virtualSize;
}

guiSize guiCanvasGetPhysicalSize(guiCanvasPtr c)
{
	return c->physicalSize;
}

guiCanvasPtr guiCanvasGetParent(guiCanvasPtr c)
{
	return c->parent;
}

guiCanvasPtr guiCanvasGetRootParent(guiCanvasPtr c)
{
	while (c->parent) {
		c = c->parent;
	}
	return c;
}

/* Size of a printer page */                 
guiSize guiGetPageSize()
{
	guiSize s;
	/* Assume 8.5x11 for now, with a half-inch margin.
		Work at 72 dpi (Mac style). */
	s.w = 72 * 7.5;
	s.h = 72 * 10;
	return s;
}

void guiXCanvasRepaintBorder(guiCanvasPtr c);

void guiXCanvasRepaintVerticalScrollbar(guiCanvasPtr c);

void guiXCanvasRepaintHorizontalScrollbar(guiCanvasPtr c);
                         
void guiCanvasInvalidate(guiCanvasPtr c, guiRectanglePtr r)
{
	guiSize s = guiCanvasGetVisibleSize(c);
	if (c->dying) {
		return;
	}
	if (!c->visible) {
		guiXCanvasInvalidateBody(c, r, 0);
		return;
	}
	if (r) {
		int w, h;
		w = guiMin(r->w, s.w - r->x + c->visibleTop.x); 
		h = guiMin(r->h, s.h - r->y + c->visibleTop.y);
		if ((w > 0) && (h > 0)) {
			if (c->clearing) {
				XClearArea(guiXDisplay, 
					(Window) c->handle,
					r->x - c->visibleTop.x + c->border,
					r->y - c->visibleTop.y + c->border,
					w, h, 0);
			}
		}
	} else {
		if (c->clearing) {
			XClearArea(guiXDisplay, 
				(Window) c->handle,
				c->border, c->border, 
				s.w,
				s.h,
				0);
		}
	}
	guiXCanvasInvalidateBody(c, r, 1);
}

void guiXCanvasInvalidateBody(guiCanvasPtr c, guiRectanglePtr r, int repaintNow)
{
	guiRectanglePtr i = &c->invalid;
	guiSize s;
	if (c->dying) {
		return;
	}
	/* Now constrain further to protect scrollbars and borders */
	s = guiCanvasGetVisibleSize(c);
	if (!r) {
		i->x = c->visibleTop.x;
		i->y = c->visibleTop.y;
		i->w = s.w;
		i->h = s.h;
	} else {
		/* The initial case */
		if ((i->w == 0) && (i->h == 0)) {
			c->invalid = *r;
		} else { 
			/* Left/top */
			if (r->x < i->x) {
				i->w += (i->x - r->x);
				i->x = r->x;
			}
			if (r->y < i->y) {
				i->h += (i->y - r->y);
				i->y = r->y;
			}		
			/* Right/bottom */
			if ((r->x + r->w) > (i->x + i->w)) {
				i->w += ((r->x + r->w) - (i->x + i->w));
			}
			if ((r->y + r->h) > (i->y + i->h)) {
				i->h += ((r->y + r->h) - (i->y + i->h));
			}
		}
	}
	if (repaintNow) {
		XRectangle rectangle;
		guiEvent e;
		e.t = guiEventRepaint;
		e.area = *i;
		c->repainting = 1;
		/* Clear stuff that's beyond the virtual document. */
		if (c->visibleTop.x + s.w > c->virtualSize.w) {
			int extra = (c->visibleTop.x + s.w) - c->virtualSize.w;
			XClearArea(guiXDisplay, 
				(Window) c->handle,
				c->border + s.w - extra, c->border, 
				extra, s.h,
				0);
		}
		if (c->visibleTop.y + s.h > c->virtualSize.h) {
			int extra = (c->visibleTop.y + s.h) - c->virtualSize.h;
			XClearArea(guiXDisplay, 
				(Window) c->handle,
				c->border, c->border + s.h - extra, 
				s.w, extra,
				0);
		}
		/* We're repainting, so constrain to the invalid area... */
		rectangle.x = c->invalid.x - c->visibleTop.x + c->border;
		rectangle.y = c->invalid.y - c->visibleTop.y + c->border;
		rectangle.width = c->invalid.w;
		rectangle.height = c->invalid.h;
		if (rectangle.x < c->border) {
			rectangle.width -= (c->border - rectangle.x);
			rectangle.x = c->border;
		}
		if (rectangle.y < c->border) {
			rectangle.height -= (c->border - rectangle.y);
			rectangle.y = c->border;
		}
		rectangle.width = guiMin(
			rectangle.width, guiMax(
				s.w - rectangle.x + c->border, 0));  
		rectangle.height = guiMin(
			rectangle.height, guiMax(
				s.h - rectangle.y + c->border, 0));  
		XSetClipRectangles(guiXDisplay, ((GC) c->drawContext),
			0, 0, &rectangle, 1, Unsorted);
		XSetClipRectangles(guiXDisplay, ((GC) c->fillContext),
			0, 0, &rectangle, 1, Unsorted);
		if ((c->repaintV) || (c->repaintH)) {
			XCopyArea(guiXDisplay, 
				(Drawable) guiXDingus->handle,
				(Drawable) c->handle,
				guiXScrollbarBodyGC,
				0, 0,
				guiXDingus->physicalSize.w, 
				guiXDingus->physicalSize.h,
				c->physicalSize.w - c->border -
					guiXDingus->physicalSize.w,
				c->physicalSize.h - c->border - 
					guiXDingus->physicalSize.h);
		}		
		if (c->repaintB) {
			guiXCanvasRepaintBorder(c);
		}
		if (c->repaintV) {
			guiXCanvasRepaintVerticalScrollbar(c);
		}
		if (c->repaintH) {
			guiXCanvasRepaintHorizontalScrollbar(c);
		}
		if (c->eventMask & guiEventRepaint) {
			c->eventHandler(c, &e);
		}
		c->repainting = 0;
		rectangle.x = c->border;
		rectangle.y = c->border;
		rectangle.width = s.w;
		rectangle.height = s.h;
		c->repaintB = 0;
		c->repaintV = 0;
		c->repaintH = 0;
		XSetClipRectangles(guiXDisplay, ((GC) c->drawContext),
			0, 0, &rectangle, 1, Unsorted);
		XSetClipRectangles(guiXDisplay, ((GC) c->fillContext),
			0, 0, &rectangle, 1, Unsorted);
		c->invalid = guiRectangleEmpty;
	}	
}

unsigned long guiXColorGet(guiColorPtr c)
{
	int r, g, b;
	int gindex;
	c->r = guiBound(c->r, 0, 255);
	c->g = guiBound(c->g, 0, 255);
	c->b = guiBound(c->b, 0, 255);
	switch(guiXColorType) {
		case guiXTrueColor:
		return (c->r >> (8 - guiXBitsR) << guiXShiftR) +
			(c->g >> (8 - guiXBitsG) << guiXShiftG) +
			(c->b >> (8 - guiXBitsB) << guiXShiftB);
		case guiXCubeColor:
		r = guiBound(c->r, 0, 255) >> 5;
		g = guiBound(c->g, 0, 255) >> 5;	
		b = guiBound(c->b, 0, 255) >> 5;	
		/* Return the information for dithering purposes */
		*c = guiXCubeToColorTable[r][g][b];
		/* Return the X color */
		return guiXCubeToIndexTable[r][g][b]; 
		case guiXGrayScale:
		gindex = (c->r >> 2) + (c->g >> 2) + (c->b >> 1); 
		*c = guiXGrayToColorTable[gindex];
		return guiXGrayToIndexTable[gindex];
		default:
		return 0;
	}
}

void guiCanvasSetDrawColor(guiCanvasPtr c, guiColor color)
{
	if (c->dying) {
		return;
	}
	c->drawColor = color;
	if (c->dim) {
		color = guiXColorAverage(c->drawColor, c->backgroundColor);
		XSetForeground(guiXDisplay, (GC) c->drawContext,
			guiXColorGet(&color));
	} else {
		XSetForeground(guiXDisplay, (GC) c->drawContext,
			guiXColorGet(&c->drawColor));
	}
}

void guiCanvasSetTextBackgroundColor(guiCanvasPtr c, guiColor color)
{
	if (c->dying) {
		return;
	}
	c->textBackgroundColor = color;
	if (c->dim) {
		color = guiXColorAverage(c->textBackgroundColor, 
			c->backgroundColor);
		XSetBackground(guiXDisplay, (GC) c->drawContext,
			guiXColorGet(&color));
	} else {
		XSetBackground(guiXDisplay, (GC) c->drawContext,
			guiXColorGet(&c->textBackgroundColor));
	}
}

void guiCanvasSetFillColor(guiCanvasPtr c, guiColor color)
{
	if (c->dying) {
		return;
	}
	c->fillColor = color;
	if (c->dim) {
		color = guiXColorAverage(c->fillColor, c->backgroundColor);
		XSetForeground(guiXDisplay, (GC) c->fillContext,
			guiXColorGet(&color));
	} else {
		XSetForeground(guiXDisplay, (GC) c->fillContext,
			guiXColorGet(&c->fillColor));
	}
}

void guiCanvasSetXOR(guiCanvasPtr c, int xor)
{
	if (c->dying) {
		return;
	}
	c->xorFlag = xor;
	if (xor) {
		XSetFunction(guiXDisplay, ((GC) c->drawContext), GXxor);
		XSetFunction(guiXDisplay, ((GC) c->fillContext), GXxor);
	} else { 
		XSetFunction(guiXDisplay, ((GC) c->drawContext), GXcopy);
		XSetFunction(guiXDisplay, ((GC) c->fillContext), GXcopy);
	}
}

void guiCanvasSetLineWidth(guiCanvasPtr c, int width)
{
	XGCValues xgcv;
	if (c->dying) {
		return;
	}
	xgcv.line_width = width;
	c->lineWidth = width;
	XChangeGC(guiXDisplay, 
		((GC) c->drawContext), 
		GCLineWidth,
		&xgcv);
}

void guiCanvasSetFont(guiCanvasPtr c, guiFontPtr f)
{
	if (c->dying) {
		return;
	}
	c->font = *f;
	guiXCanvasSetGC(c);
	c->fontSet = 1;
}

void guiCanvasDrawLine(guiCanvasPtr c, int x1, int y1, int x2, int y2)
{
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	if (c->psoutput) {
		guiXCanvasPSDraw(c);
		fprintf(c->psoutput, "%d %d moveto\n", x1, y1);
		fprintf(c->psoutput, "%d %d lineto\n", x2, y2);
		fprintf(c->psoutput, "stroke\n");
	} else {
		XDrawLine(guiXDisplay, (Drawable) c->handle, 
			((GC) c->drawContext),
			x1 - c->visibleTop.x + c->border, 
			y1 - c->visibleTop.y + c->border, 
			x2 - c->visibleTop.x + c->border, 
			y2 - c->visibleTop.y + c->border);
	}
}

void guiCanvasDrawPolygon(guiCanvasPtr c,
        guiPoint *points, int pointsTotal)
{
	int i;
	if (c->dying) {
		return;
	}
	if (pointsTotal < 3) {
		return;
	}
	guiXCanvasSetGC(c);
	if (c->psoutput) {
		guiXCanvasPSDraw(c);
		fprintf(c->psoutput, "%d %d moveto\n", 
			points[0].x, points[0].y);
		for (i = 1; (i < pointsTotal); i++) {
			fprintf(c->psoutput, "%d %d lineto\n", 
				points[i].x, points[i].y);	
		}
		fprintf(c->psoutput, "closepath\n");
		fprintf(c->psoutput, "stroke\n");
	} else {
		for (i = 0; (i < (pointsTotal - 1)); i++) {
			XDrawLine(guiXDisplay, (Drawable) c->handle,  
				((GC) c->drawContext), 
				points[i].x - c->visibleTop.x + c->border, 
				points[i].y - c->visibleTop.y + c->border,	
				points[i + 1].x - c->visibleTop.x +
				c->border, 
				points[i + 1].y - c->visibleTop.y +
				c->border);
		}
		XDrawLine(guiXDisplay, (Drawable) c->handle,  
			((GC) c->drawContext), 
			points[pointsTotal - 1].x - c->visibleTop.x 
				+ c->border, 
			points[pointsTotal - 1].y - c->visibleTop.y
				+ c->border,	
			points[0].x - c->visibleTop.x + c->border, 
			points[0].y - c->visibleTop.y + c->border);
	}
}

void guiCanvasDrawPolyline(guiCanvasPtr c,
        guiPoint *points, int pointsTotal)
{
	int i;
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	if (c->psoutput) {
		guiXCanvasPSDraw(c);
		fprintf(c->psoutput, "%d %d moveto\n", 
			points[0].x, points[0].y);
		for (i = 1; (i < pointsTotal); i++) {
			fprintf(c->psoutput, "%d %d lineto\n", 
				points[i].x, points[i].y);	
		}
		fprintf(c->psoutput, "stroke\n");
	} else {
		for (i = 0; (i < (pointsTotal - 1)); i++) {
			XDrawLine(guiXDisplay, (Drawable) c->handle,  
				((GC) c->drawContext), 
				points[i].x - c->visibleTop.x + c->border, 
				points[i].y - c->visibleTop.y + c->border,	
				points[i + 1].x - c->visibleTop.x + c->border, 
				points[i + 1].y - c->visibleTop.y + c->border);
		}
	}
}

void guiCanvasFillPolygon(guiCanvasPtr c,
        guiPoint *points, int pointsTotal)
{
	int i;
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	if (pointsTotal < 3) {
		return;
	}
	if (c->psoutput) {
		guiXCanvasPSFill(c);
		fprintf(c->psoutput, "%d %d moveto\n", 
			points[0].x, points[0].y);
		for (i = 1; (i < pointsTotal); i++) {
			fprintf(c->psoutput, "%d %d lineto\n", 
				points[i].x, points[i].y);	
		}
		fprintf(c->psoutput, "closepath\n");
		fprintf(c->psoutput, "fill\n");
		return;
	} 	
	if (pointsTotal > guiXPointsAllocated) {
		XPoint *xPoints;
		int allocated = guiXPointsAllocated * 2;
		xPoints = realloc(guiXPoints, sizeof(XPoint) * allocated);
		if (!xPoints) {
			/* Fail gracefully */
			return;
		}
		guiXPointsAllocated = allocated;
		guiXPoints = xPoints;	
	}
	for (i = 0; (i < pointsTotal); i++) {
		guiXPoints[i].x = points[i].x - c->visibleTop.x + c->border;
		guiXPoints[i].y = points[i].y - c->visibleTop.y + c->border;
	}
	XFillPolygon(guiXDisplay, (Drawable) c->handle, ((GC) c->fillContext), 
		guiXPoints, pointsTotal, Complex, CoordModeOrigin);
}

void guiCanvasDrawRectangle(guiCanvasPtr c, int x, int y, int w, int h)
{
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	if (c->psoutput) {
		guiXCanvasPSDraw(c);
		fprintf(c->psoutput, "%d %d moveto\n", x, y);
		fprintf(c->psoutput, "%d %d lineto\n", x + w - 1, y);
		fprintf(c->psoutput, "%d %d lineto\n", x + w, y + h - 1);
		fprintf(c->psoutput, "%d %d lineto\n", x, y + h - 1);
		fprintf(c->psoutput, "closepath\n");
		fprintf(c->psoutput, "stroke\n");
	}
	XDrawRectangle(guiXDisplay, (Drawable) c->handle, 
		((GC) c->drawContext),
		x - c->visibleTop.x + c->border, 
		y - c->visibleTop.y + c->border, 
		w, h);
}

void guiCanvasFillRectangle(guiCanvasPtr c, int x, int y, int w, int h)
{
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	if (c->psoutput) {
		guiXCanvasPSFill(c);
		fprintf(c->psoutput, "%d %d moveto\n", x, y);
		fprintf(c->psoutput, "%d %d lineto\n", x + w - 1, y);
		fprintf(c->psoutput, "%d %d lineto\n", x + w, y + h - 1);
		fprintf(c->psoutput, "%d %d lineto\n", x, y + h - 1);
		fprintf(c->psoutput, "closepath\n");
		fprintf(c->psoutput, "fill\n");
	}	
	XFillRectangle(guiXDisplay, (Drawable) c->handle, 
		((GC) c->fillContext),
		x - c->visibleTop.x + c->border, 
		y - c->visibleTop.y + c->border, 
		w, h);
}

void guiCanvasDrawEllipse(guiCanvasPtr c, int x, int y, int w, int h)
{
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	if (c->psoutput) {
		double scaling;
		double radius;
		if (!w) {
			return;
		}
		if (!h) {
			return;
		}
		guiXCanvasPSDraw(c);
		fprintf(c->psoutput, "gsave\n");
		scaling = (double) h / (double) w;
		fprintf(c->psoutput, "%f scale\n", scaling);
		radius = (double) w / 2.0;
		fprintf(c->psoutput, "%f %f %f 0 360 arc stroke\n",
			(double) x + radius,
			(double) y / scaling + radius,
			radius);
		fprintf(c->psoutput, "grestore\n");
	} else {
		XDrawArc(guiXDisplay, (Drawable) c->handle, 
			((GC) c->drawContext),
			x - w / 2 - c->visibleTop.x + c->border, 
			y - h / 2 - c->visibleTop.y + c->border, 
			w, h,
			0, 360 * 64);
	}
}

void guiCanvasFillEllipse(guiCanvasPtr c, int x, int y, int w, int h)
{
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	if (c->psoutput) {
		double scaling;
		double radius;
		if (!w) {
			return;
		}
		if (!h) {
			return;
		}
		guiXCanvasPSFill(c);
		fprintf(c->psoutput, "gsave\n");
		scaling = (double) h / (double) w;
		fprintf(c->psoutput, "%f scale\n", scaling);
		radius = (double) w / 2.0;
		fprintf(c->psoutput, "%f %f %f 0 360 arc fill\n",
			(double) x + radius,
			(double) y / scaling + radius,
			radius);
		fprintf(c->psoutput, "grestore\n");
	} else {	
		XFillArc(guiXDisplay, (Drawable) c->handle, 
			((GC) c->fillContext),
			x - w / 2 - c->visibleTop.x + c->border, 
			y - h / 2 - c->visibleTop.y + c->border, w, h,
			0, 360 * 64);
	}
}

void guiCanvasDrawText(guiCanvasPtr c, int x, int y, char *text, int len)
{
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	/* Sizes the font downwards to the closest available
		pixel size and makes sure it is first up
		in the cache. If the font is already in the
		cache, this does not require any real X server work. */
	if (c->fontSet) {
		guiFontFindClosest(&c->font);	
		XSetFont(guiXDisplay, ((GC) c->drawContext), guiXFontCache[0].data->fid);
	}
	if (c->psoutput) {
		fprintf(c->psoutput, "%d %d moveto\n", x, y);
		fprintf(c->psoutput, "( %*s ) show\n", len, text);
	} else {	
		XDrawString(guiXDisplay, (Drawable) c->handle,
			((GC) c->drawContext), 
			x - c->visibleTop.x + c->border, 
			y - c->visibleTop.y + c->border, 
			text, len);	
	}
}

void guiCanvasDrawOpaqueText(guiCanvasPtr c, int x, int y, char *text, int len)
{
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	/* Sizes the font downwards to the closest available
		pixel size and makes sure it is first up
		in the cache. If the font is already in the
		cache, this does not require any real X server work. */
	if (c->fontSet) {
		guiFontFindClosest(&c->font);	
		XSetFont(guiXDisplay, ((GC) c->drawContext), guiXFontCache[0].data->fid);
	}
	if (c->psoutput) {
		fprintf(c->psoutput, "%d %d moveto\n", x, y);
		fprintf(c->psoutput, "( %*s ) show\n", len, text);
	} else {	
		XDrawImageString(guiXDisplay, (Drawable) c->handle,
			((GC) c->drawContext), 
			x - c->visibleTop.x + c->border, 
			y - c->visibleTop.y + c->border, 
			text, len);	
	}
}

void guiCanvasDrawBitmap(guiCanvasPtr c, int x, int y, guiBitmapPtr b)
{
	guiCanvasDrawBitmapSection(c, x, y, 0, 0, b->s.w, b->s.h, b);
}

void guiCanvasDrawBitmapSection(guiCanvasPtr c,
	int dstx, int dsty,
	int srcx, int srcy, int w, int h,
	guiBitmapPtr b)
{
	if (c->dying) {
		return;
	}
	guiXCanvasSetGC(c);
	if (srcx >= b->s.w) {
		return;
	}	
	if (srcy >= b->s.h) {
		return;
	}	
	if (w > (srcx + b->s.w)) {
		w = b->s.w - srcx;
	}
	if (h > (srcy + b->s.h)) {
		h = b->s.h - srcy;
	}
	if (c->psoutput) {
		int i, j;
		fprintf(c->psoutput, 	
			"gsave\n");
		fprintf(c->psoutput, 	
			"%d %d translate\n",
			dstx, dsty);
		fprintf(c->psoutput, 	
			"%d %d scale\n",
			w, h);
		fprintf(c->psoutput, 	
			"%d %d %d [%d %d %d %d %d %d] ",
			w, h, 8,
			w, 0, 0, h, 0, 0);	
		fprintf(c->psoutput, 
			"{<");
		for (j = srcy; (j < srcy + h); j++) {	
			for (i = srcx; (i < srcx + w); i++) {	
				fprintf(c->psoutput,
					"%2x", guiXToGray(
						XGetPixel((XImage *) b->handle,
							i, j)));
			}
		}
		fprintf(c->psoutput, ">} image\n");	
		fprintf(c->psoutput, 	
			"grestore\n");
	} else {
		XPutImage(guiXDisplay, (Drawable) c->handle,
			((GC) c->drawContext),
			(XImage *) b->handle,
			srcx, srcy,
			dstx - c->visibleTop.x + c->border, 
			dsty - c->visibleTop.y + c->border, w, h);
	}
}

void guiCanvasCopyArea(guiCanvasPtr dst,
	guiCanvasPtr src,
	int dstx, int dsty,
	int srcx, int srcy, int w, int h)
{
	if (dst->dying) {
		return;
	}
	if (src->dying) {
		return;
	}
	guiXCanvasSetGC(dst);
	if (srcx >= src->virtualSize.w) {
		return;
	}	
	if (srcy >= src->virtualSize.h) {
		return;
	}	
	if (w > (srcx + src->virtualSize.w)) {
		w = src->virtualSize.w - srcx;
	}
	if (h > (srcy + src->virtualSize.h)) {
		h = src->virtualSize.h - srcy;
	}
	if (dst->psoutput) {
		guiBitmapPtr b;
		 if (src->psoutput) {
			/* We can't do anything in this case 
				(same canvas or another canvas
				that is printing) */
			return;
		}
		/* We can use a temporary bitmap to transfer the area */
		b = guiCanvasGrabBitmap(src,
			srcx - src->visibleTop.x + src->border, 
			srcy - src->visibleTop.y + src->border, w, h);
		if (!b) {
			return;
		}
		guiCanvasDrawBitmap(dst, dstx, dsty, b);
		guiBitmapDestroy(b);
	} else {
		XCopyArea(guiXDisplay, 
			(Drawable) src->handle,
			(Drawable) dst->handle,
			((GC) dst->drawContext),
			srcx - src->visibleTop.x + src->border, 
			srcy - src->visibleTop.y + src->border,
			w, h,
			dstx - dst->visibleTop.x + dst->border, 
			dsty - dst->visibleTop.y + dst->border);
	}
}

guiBitmapPtr guiCanvasGrabBitmap(guiCanvasPtr c,
        int x, int y, int w, int h)
{
	guiBitmapPtr b;
	if (c->dying) {
		return 0;
	}
	b = (guiBitmapPtr) malloc(sizeof(guiBitmap));
	if (!b) {
		return 0;
	}
	b->s.w = w;
	b->s.h = h;
	b->sourceType = guiBitmapPng;
	b->handle = (guiNativeHandle) XGetImage(guiXDisplay,
		(Drawable) c->handle, (unsigned long) ~0L, 
		x - c->visibleTop.x + c->border, 
		y - c->visibleTop.y + c->border, 
		w, h,
		ZPixmap);
	if (!b->handle) {
		free(b);
		return 0;
	}
	return b;
}

void guiCanvasPrint(guiCanvasPtr c, FILE *out)
{
	guiSize s = guiGetPageSize();
	int pages;
	int page;
	guiEvent e;
	guiPoint pageTopLeft;
	if (c->dying) {
		return;
	}
	if (s.h == 0) {
		return;
	}
	pages = c->virtualSize.h / s.h;
	if (c->virtualSize.h % s.h) {
		pages++;
	}		 
	pageTopLeft.x = 0;
	pageTopLeft.y = 0;
	c->psoutput = out;
	c->pslastfill = 0;
	/* 72 dpi, 0,0 in the upper left, .5 inch margin with clipping. 
		Translate for each page; let Postscript do most of the work. */
	fprintf(out, "/BeginPage\n");
	fprintf(out, "{ 1 -1 scale\n");
	fprintf(out, "newpath\n");
	fprintf(out, "0 0 moveto\n");
	fprintf(out, "540 0 lineto\n");
	fprintf(out, "540 720 lineto\n");
	fprintf(out, "0 720 lineto\n");
	fprintf(out, "closepath\n");
	fprintf(out, "clip\n");
	fprintf(out, "} def\n");
	for (page = 0; (page < pages); page++) {
		fprintf(out, "BeginPage\n");
		fprintf(out, "36 %d translate\n",
			-756 - pageTopLeft.y);
		/* Send an event to paint the page. */
		e.t = guiEventRepaint;
		e.area.x = pageTopLeft.x;
		e.area.y = pageTopLeft.y;
		e.area.w = s.w;
		e.area.h = s.h;
		c->eventHandler(c, &e);

		pageTopLeft.y += s.h;
		fprintf(out, "showpage\n");
	}
	c->psoutput = 0;
}

void guiXCanvasSetGC(guiCanvasPtr c)
{
	if (c->psoutput) {
		fprintf(c->psoutput, "%f setgray\n",
			(c->drawColor.r + c->drawColor.g +
				c->drawColor.b) / 765.0);
		c->pslastfill = 0;
		fprintf(c->psoutput, "/%s findfont %d scalefont setfont\n",
			guiXPSFonts[c->font.family]
				[c->font.italic][c->font.bold], c->font.size);
		return;
	} 

}

void guiFontFindClosest(guiFontPtr f)
{
	int i, j;
	int found = 0;
	for (i = 0; (i < guiXFontCacheSize); i++) {
		if (guiXFontCache[i].used) {
			if (guiImplFontEq(&guiXFontCache[i].spec, f) ||
				guiImplFontEq(&guiXFontCache[i].actual, f)) {
				/* Move it to the first slot; slide
					the others to the right */
				if (i != 0) {
					guiXFontCacheEntry e = 
						guiXFontCache[i];
					for (j = (i - 1); (j >= 0); j--) { 	
						guiXFontCache[j + 1] = 	
							guiXFontCache[j];
					}
					guiXFontCache[0] = e;
				}				
				found = 1;
				break;
			}
		}
	}
	if (!found) {
		char fontname[256];
		int trysize;
		guiXFontCacheEntry *e;
		/* Give up on the oldest font in the cache */
		if (guiXFontCache[guiXFontCacheSize - 1].used) {
			XFreeFont(guiXDisplay, 
				guiXFontCache[guiXFontCacheSize - 1].data);
		}
		/* Move the other entries to the right */
		for (j = (guiXFontCacheSize - 2); (j >= 0); j--) { 	
			guiXFontCache[j + 1] = 	
				guiXFontCache[j];
		}	
		/* Claim the first slot */
		trysize = f->size;
		found = 0;
		while(trysize > 1) {
			e = &guiXFontCache[0];
			sprintf(fontname, 
				"%s-normal--%d-*-*-*-*-*-iso8859-1",
				guiXFonts[f->family]
					[f->italic][f->bold],
				trysize);
			e->data = XLoadQueryFont(
				guiXDisplay, fontname);
			if (e->data) {
				/* Got something */
				e->actual = *f;
				e->actual.size = trysize;
				e->actual.ascent = e->data->ascent;
				e->actual.descent = e->data->descent;
				found = 1;
				break;
			}	
			trysize--;
		}	
		if (!found) {
			guiFatalError("Unable to find the font %s at size %d or below. "
				"You will need to install this freely available "
				"font in its regular, bold, italic, and bold italic "
				"versions. This font is part of the standard X "
				"consortium distribution of the X Window System. "
				"Please consult your system administrator.", 
				guiXFonts[f->family][0][0], f->size);
		}
	}
	guiXFontCache[0].spec = *f;
	f->size = guiXFontCache[0].actual.size;
	f->ascent = guiXFontCache[0].actual.ascent;
	f->descent = guiXFontCache[0].actual.descent;
	guiXFontCache[0].used = 1;
}

void guiXCanvasPSDraw(guiCanvasPtr c)
{
	if (c->pslastfill) {
		fprintf(c->psoutput, "%f setgray\n",
			(c->drawColor.r + c->drawColor.g +
				c->drawColor.b) / 765.0);
		c->pslastfill = 0;
	}
}

void guiXCanvasPSFill(guiCanvasPtr c)
{
	if (!c->pslastfill) {
		fprintf(c->psoutput, "%f setgray\n",
			(c->fillColor.r + c->fillColor.g +
				c->fillColor.b) / 765.0);
		c->pslastfill = 1;
	}
}

guiSize guiFontTextSize(guiFontPtr fp, char *text, int len)
{
	guiSize s;
	XCharStruct xc;
	int direction;
	int ascent, descent;
	guiFont f = *fp;
	guiFontFindClosest(&f);
	XTextExtents(guiXFontCache[0].data, text, len, &direction,
		&ascent, &descent, &xc);
	s.w = xc.width;
	s.h = xc.ascent + xc.descent;
	return s;
}

guiBitmapPtr guiBitmapLoad(FILE *in, int dither)
{
	unsigned long w;
	unsigned long h;
	guiImplSizeResult r;
	r = guiImplSizeGif(in, &w, &h);
	rewind(in);		
	if (r == guiImplSizeSuccess) {
		return guiXBitmapLoadGif(in, dither);
	}
	r = guiImplSizeJpeg(in, &w, &h);
	rewind(in);		
	if (r == guiImplSizeSuccess) {
		return guiXBitmapLoadJpeg(in, dither);
	}
	r = guiImplSizePng(in, &w, &h);
	rewind(in);		
	if (r == guiImplSizeSuccess) {
		return guiXBitmapLoadPng(in, 0, dither);
	}
	/* What type is this? */
	return 0;
}

guiBitmapPtr guiBitmapBuiltin(unsigned char *resource, int dither)
{
	unsigned char pngsig[] = { 137, 80, 78, 71, 13, 10, 26, 10 };
	if (strncmp((const char *) resource, (const char *) pngsig, 8))  {
		fprintf(stderr, "Sorry, only PNG images can be bound into a "\
			"program. Use the utilities provided with libgui "\
			"to convert non-PNG images.\n");
		return 0;
	}
	return guiXBitmapLoadPng(0, resource, dither);
}

#if BITMAP_SAVE_IMPLEMENTED
int guiBitmapSave(guiBitmapPtr bitmap, FILE *out, guiBitmapSourceType s)
{
	switch(s) {
		case guiBitmapJpeg:
		return guiXBitmapSaveJpeg(bitmap, out);
		case guiBitmapPng:
		return guiXBitmapSavePng(bitmap, out);
		case guiBitmapGif:
		return guiXBitmapSaveGif(bitmap, out);
	}		
	/* What type is this? */
	return 0;
}
#endif

guiBitmapPtr guiXBitmapCreate(int w, int h)
{
	guiBitmapPtr b; 
	unsigned char *imageData;
	b = (guiBitmapPtr) malloc(sizeof(guiBitmap));
	if (!b) {
		return 0;
	}
	b->sourceType = guiBitmapPng;
	/* Set the size */
	b->s.w = w;
	b->s.h = h;
	if (guiXDepth == 24) {
		imageData = (unsigned char *) calloc((w * 4 + 2) * h, 1);
	} else if (guiXDepth == 12) {
		imageData = (unsigned char *) calloc((w * 2 + 2) * h, 1);
	} else {
		imageData = (unsigned char *) calloc((w * guiXDepth / 8 + 2) 
			* h, 1);
	}
	if (!imageData) {
		free(b);
		return 0;
	}
	b->handle = (guiNativeHandle) XCreateImage(
		guiXDisplay, guiXVisual, 
			guiXDepth, ZPixmap,
			0, (char *) imageData, w, h, 8, 0);
	if (!b->handle) {
		free(imageData);
		free(b);
		return 0;
	}
	return b;
}

/* Returns a scaled version of a portion of a bitmap.
        The original bitmap is not changed. */
guiBitmapPtr guiBitmapPortion(guiBitmapPtr src,
	int srcx, int srcy,
	int srcw, int srch,
	int dstw, int dsth)
{
	guiBitmapPtr b; 
	int x, y;
	b = guiXBitmapCreate(dstw, dsth);
	if (!b) {
		return 0;
	}
	/* Inherit the source type. */
	b->sourceType = src->sourceType;
	/* Set the size */
	b->s.w = dstw;
	b->s.h = dsth;	
	/* Dumb requests */
	if ((srcx >= src->s.w) || (srcy >= src->s.h)) {	
		srcx = 0;
		srcy = 0;
		srcw = 0;
		srch = 0;
	}	
	/* Clip the request */
	if (srcx + srcw > src->s.w) {	
		srcw = src->s.w - srcx;
	}	
	if (srcy + srch > src->s.h) {	
		srch = src->s.h - srcy;
	}	
	/* This is not efficient. Too bad. X visuals are simply
		too complicated to cope with any other way. */
	for (y = 0; (y < dsth); y++) {
		for (x = 0; (x < dstw); x++) {
			int xs, ys;
			xs = (x * srcw / dstw) + srcx;
			ys = (y * srcw / dstw) + srcy;
			XPutPixel((XImage *) b->handle,
				x, y, 
				XGetPixel(
					(XImage *) src->handle,
					xs, ys));
		}
	}			
	return b;
}

void guiBitmapDestroy(guiBitmapPtr bitmap)
{
	XDestroyImage((XImage *) bitmap->handle);	
	free(bitmap);
}

int guiXToGray(unsigned long xp)
{
	int r, g, b;
	switch(guiXColorType) {
		case guiXTrueColor:
		r = ((xp & guiXRedMask) >> guiXShiftR) <<
			(8 - guiXBitsR);
		g = ((xp & guiXGreenMask) >> guiXShiftG) <<
			(8 - guiXBitsG);
		b = ((xp & guiXBlueMask) >> guiXShiftB) <<
			(8 - guiXBitsB);
		return (r >> 2) + (g >> 2) + (b >> 1);
		case guiXCubeColor:
		case guiXGrayScale:
		return guiXToGrayTable[(int) (xp & 255)];
	}	
	return 0;
}

#define guiXTimersInitial 16

extern unsigned char guiXDingusData[];
extern unsigned char guiXThumb000Data[];
extern unsigned char guiXThumb010Data[];
extern unsigned char guiXThumb020Data[];
extern unsigned char guiXThumb100Data[];
extern unsigned char guiXThumb110Data[];
extern unsigned char guiXThumb120Data[];
extern unsigned char guiXSarrow00Data[];
extern unsigned char guiXSarrow01Data[];
extern unsigned char guiXSarrow02Data[];
extern unsigned char guiXSarrow03Data[];
extern unsigned char guiXSarrow10Data[];
extern unsigned char guiXSarrow11Data[];
extern unsigned char guiXSarrow12Data[];
extern unsigned char guiXSarrow13Data[];


int guiSetup()
{
	int i;
	int j;
	Colormap colormap;
	unsigned long colorsAllocated[256];
	int colorsAllocatedTotal = 0;
	XGCValues gcvalues;
	guiColor bg;
	XColor color;
	XColor qcolors[256];
	guiXModalFirst = 0;
	guiXTimers = (guiXTimerPtr) malloc(sizeof(guiXTimer) *
		guiXTimersInitial);
	if (!guiXTimers) {
		return 0;
	}
	guiXTimersAllocated = guiXTimersInitial;
	guiXTimersTotal = 0;
	guiXFirstGrab = 0;
	for (i = 0; (i < guiXFontCacheSize); i++) {
		guiXFontCache[i].used = 0;
	}	
	guiXDisplay = XOpenDisplay("");
	if (!guiXDisplay) {
		free(guiXTimers);
		return 0;
	}
	guiXScreen = DefaultScreen(guiXDisplay);
	colormap = DefaultColormap(guiXDisplay, guiXScreen);
	guiXDepth = DefaultDepth(guiXDisplay, guiXScreen);
	guiXContext = XUniqueContext();
	guiXVisual = DefaultVisual(guiXDisplay, guiXScreen);
	switch (guiXVisual->class) {
		int failed;
		case PseudoColor:
		guiXColorType = guiXCubeColor;
		failed = 0;
		for (j = 5; (j >= 2); j--) {
			failed = 0;
			if ((1 << guiXDepth) < (j * j * j)) {
				failed = 1;
				continue;
			}
			for (i=0; (i < (j * j * j)); i++) {
				int r, g, b;
				r = (i % j);
				g = (i / j) % j;
				b = (i / (j * j));
				color.red = r * 65535L / (j-1);
				color.green = g * 65535L / (j-1);
				color.blue = b * 65535L / (j-1);
				if (!XAllocColor(guiXDisplay, colormap, &color)) {
					failed = 1;
					XFreeColors(guiXDisplay, colormap,
					colorsAllocated, colorsAllocatedTotal, 0);
					colorsAllocatedTotal = 0;
					break;
				} else {
					colorsAllocated[colorsAllocatedTotal++] = color.pixel;
				}
			}
			if (!failed) {
				/* Okay, point the entries in the
					8x8x8 virtual cube at whatever
					entries are closest in the
					real cube we were able to get */
				int rd, gd, bd;
				for (rd = 0; (rd < 8); rd++) {
					for (gd = 0; (gd < 8); gd++) {	
						for (bd = 0; (bd < 8); bd++) {
							int i;
							guiColor c;
							i = (rd * j / 8) +
								(gd * j / 8) * j +
								(bd * j / 8) * j * j;
							guiXCubeToIndexTable[rd][gd][bd] = colorsAllocated[i];
							c.r = (rd * j / 8) * 255 / (j - 1);
							c.g = (gd * j / 8) * 255 / (j - 1);
							c.b = (bd * j / 8) * 255 / (j - 1);
							guiXCubeToColorTable[rd][gd][bd] = c;	
						}
					}
				}
				if (j == 5) {
					/* We got it all, so try to allocate a
						linear gray ramp also! */
					int i;	
					for (i = 0; (i < 8); i++) {
						XColor color;
						int r, g, b;
						r = g = b = i * 255 / 7;
						color.red = r * 65535L / 255;
						color.green = g * 65535L / 255;
						color.blue = b * 65535L / 255;
						if (!XAllocColor(guiXDisplay, colormap, &color)) {
							XFreeColors(guiXDisplay, colormap,
								guiXGrayToIndexTable, i, 0);
							goto complete;
						}
						guiXCubeToIndexTable[i][i][i] = color.pixel;
						guiXCubeToColorTable[i][i][i].r = r;
						guiXCubeToColorTable[i][i][i].g = g;
						guiXCubeToColorTable[i][i][i].b = b;
					} 
				}
complete:
				break;
			}
		}
		if (failed) {
			/* Okay, black and white, then. Grr. */
			guiColor white;
			guiColor black;
			int rd, gd, bd;
			white.r = 255;
			white.g = 255;
			white.b = 255;
			black.r = 0;
			black.g = 0;
			black.b = 0;
			for (rd = 0; (rd < 8); rd++) {
				for (gd = 0; (gd < 8); gd++) {	
					for (bd = 0; (bd < 8); bd++) {
						if (rd + gd + bd > 12) {
							guiXCubeToIndexTable[rd][gd][bd] = WhitePixel(guiXDisplay, guiXScreen);
							guiXCubeToColorTable[rd][gd][bd] = white;
						} else {
							guiXCubeToIndexTable[rd][gd][bd] = BlackPixel(guiXDisplay, guiXScreen);
							guiXCubeToColorTable[rd][gd][bd] = black;
						}
					}
				}
			}
		}
		/* Now we get REALLY clever and suck the existing
			shared read-only colors into an array, decide
			if they are useful to us, and try reallocating
			them. */
		for (i = 0; (i < 256); i++) {
			qcolors[i].pixel = i;
		}
		XQueryColors(guiXDisplay, colormap, qcolors, 256);
		for (i = 0; (i < 256); i++) {
			long distance1, distance2;
			int rd, bd, gd;
			guiColor c;
			guiColor try;
			guiColor ideal;
			try.r = qcolors[i].red >> 8;
			try.g = qcolors[i].green >> 8;
			try.b = qcolors[i].blue >> 8;
			rd = try.r >> 5;
			gd = try.g >> 5;
			bd = try.b >> 5;
			ideal.r = guiMin((rd << 5) + (rd << 2) + (rd >> 1), 255);
			ideal.g = guiMin((gd << 5) + (gd << 2) + (gd >> 1), 255);
			ideal.b = guiMin((bd << 5) + (bd << 2) + (bd >> 1), 255);
			c = guiXCubeToColorTable[rd][gd][bd];
			distance1 = abs(try.r - ideal.r) +
				abs(try.g - ideal.g) +	
				abs(try.b - ideal.b);
			distance2 = abs(c.r - ideal.r) +
				abs(c.g - ideal.g) +	
				abs(c.b - ideal.b);
			if (distance1 < distance2) 
			{
				/* We want the little stinker */
				if (XAllocColor(guiXDisplay, 
					colormap, &qcolors[i])) 
				{
					guiXCubeToIndexTable[rd][gd][bd] =
						qcolors[i].pixel;
					guiXCubeToColorTable[rd][gd][bd] =
						try;
				}
			}
		}
		break;
		case GrayScale:
		case StaticGray:
		guiXColorType = guiXGrayScale;
		for (i = 0; (i < 255); i++) { 
			if (i > 127) {
				guiXGrayToColorTable[i].r = 255;
				guiXGrayToColorTable[i].g = 255;
				guiXGrayToColorTable[i].b = 255;
				guiXGrayToIndexTable[i] = 
					WhitePixel(guiXDisplay, guiXScreen);
			} else {
				guiXGrayToColorTable[i].r = 0;
				guiXGrayToColorTable[i].g = 0;
				guiXGrayToColorTable[i].b = 0;
				guiXGrayToIndexTable[i] = 
					BlackPixel(guiXDisplay, guiXScreen);
			}
		}
		break;
		guiXColorType = guiXGrayScale;
		break;
		case TrueColor:
		case DirectColor:
		guiXColorType = guiXTrueColor;
		guiXFindShiftAndBits(guiXRedMask = guiXVisual->red_mask, 
			&guiXShiftR, &guiXBitsR);
		guiXFindShiftAndBits(guiXGreenMask = guiXVisual->green_mask, 
			&guiXShiftG, &guiXBitsG);
		guiXFindShiftAndBits(guiXBlueMask = guiXVisual->blue_mask, 
			&guiXShiftB, &guiXBitsB);
		break;	
	}

	/* Create a window to be window group leader and
		argument to gc creation; never mapped */

	guiXLeader = XCreateSimpleWindow(guiXDisplay,
		DefaultRootWindow(guiXDisplay),
		0, 0, 64, 64, 0, 
		WhitePixel(guiXDisplay, guiXScreen),
		BlackPixel(guiXDisplay, guiXScreen));
	XStoreName(guiXDisplay, guiXLeader, "guiXLeader");
	guiXLastCanvas = 0;
	guiXFocusCanvas = 0;
	/* Allocate GCs */
	gcvalues.foreground = WhitePixel(guiXDisplay, guiXScreen);
	gcvalues.background = BlackPixel(guiXDisplay, guiXScreen); 
	guiXDrawGC = XCreateGC(guiXDisplay, guiXLeader,
		GCForeground | GCBackground, &gcvalues);
	guiXFillGC = XCreateGC(guiXDisplay, guiXLeader,
		GCForeground | GCBackground, &gcvalues);
	gcvalues.foreground = guiXColorGet(&guiScrollbarBackgroundColor);
	gcvalues.background = guiXColorGet(&guiScrollbarBodyColor);
	guiXScrollbarBackgroundGC = XCreateGC(guiXDisplay, guiXLeader,
		GCForeground | GCBackground, &gcvalues);
	gcvalues.foreground = guiXColorGet(&guiScrollbarBodyColor);
	gcvalues.background = guiXColorGet(&guiScrollbarBackgroundColor);
	guiXScrollbarBodyGC = XCreateGC(guiXDisplay, guiXLeader,
		GCForeground | GCBackground, &gcvalues);
	gcvalues.foreground = guiXColorGet(&guiBorderHighColor);
	bg = guiDefaultBackgroundColor;
	gcvalues.background = guiXColorGet(&bg);
	guiXBorderHighGC = XCreateGC(guiXDisplay, guiXLeader,
		GCForeground | GCBackground, &gcvalues);
	gcvalues.foreground = guiXColorGet(&guiBorderLowColor);
	gcvalues.background = guiXColorGet(&bg);
	guiXBorderLowGC = XCreateGC(guiXDisplay, guiXLeader,
		GCForeground | GCBackground, &gcvalues);
	guiXPoints = (XPoint *) malloc(sizeof(XPoint) * 16);
	if (!guiXPoints) {
		return 0;
	}
	guiXPointsAllocated = 16;
	guiXArrowBitmaps[0][0] = guiBitmapBuiltin(guiXSarrow00Data, 1);
	guiXArrowBitmaps[0][1] = guiBitmapBuiltin(guiXSarrow01Data, 1);
	guiXArrowBitmaps[0][2] = guiBitmapBuiltin(guiXSarrow02Data, 1);
	guiXArrowBitmaps[0][3] = guiBitmapBuiltin(guiXSarrow03Data, 1);
	guiXArrowBitmaps[1][0] = guiBitmapBuiltin(guiXSarrow10Data, 1);
	guiXArrowBitmaps[1][1] = guiBitmapBuiltin(guiXSarrow11Data, 1);
	guiXArrowBitmaps[1][2] = guiBitmapBuiltin(guiXSarrow12Data, 1);
	guiXArrowBitmaps[1][3] = guiBitmapBuiltin(guiXSarrow13Data, 1);
	for (i = 0; (i < 2); i++) {
		int j;
		for (j = 0; (j < 4); j++) {
			guiXArrows[i][j] = guiCanvasCreateOffscreen(
				guiXArrowBitmaps[i][j]->s.w,
				guiXArrowBitmaps[i][j]->s.h);
			guiCanvasDrawBitmap(guiXArrows[i][j],
				0, 0, guiXArrowBitmaps[i][j]);	
		}
	}
	guiXThumbBitmaps[0][0][0] = guiBitmapBuiltin(guiXThumb000Data, 1);
	guiXThumbBitmaps[0][1][0] = guiBitmapBuiltin(guiXThumb010Data, 1);
	guiXThumbBitmaps[0][2][0] = guiBitmapBuiltin(guiXThumb020Data, 1);
	guiXThumbBitmaps[1][0][0] = guiBitmapBuiltin(guiXThumb100Data, 1);
	guiXThumbBitmaps[1][1][0] = guiBitmapBuiltin(guiXThumb110Data, 1);
	guiXThumbBitmaps[1][2][0] = guiBitmapBuiltin(guiXThumb120Data, 1);
	for (i = 0; (i < 2); i++) {
		int j;
		for (j = 0; (j < 3); j++) {
			int k;
			for (k = 0; (k < 1); k++) {
				guiXThumbs[i][j][k] = guiCanvasCreateOffscreen(
					guiXThumbBitmaps[i][j][k]->s.w,
					guiXThumbBitmaps[i][j][k]->s.h);
				guiCanvasDrawBitmap(guiXThumbs[i][j][k],
					0, 0, guiXThumbBitmaps[i][j][k]);	
			}
		}
	}
	guiXDingusBitmap = guiBitmapBuiltin(guiXDingusData, 1);
	if (!guiXDingusBitmap) {
		fprintf(stderr, "couldn't decode a standard bitmap\n");
		exit(1);
	}
	guiXDingus = guiCanvasCreateOffscreen(
		guiXDingusBitmap->s.w,
		guiXDingusBitmap->s.h);
	guiCanvasDrawBitmap(guiXDingus,
		0, 0, guiXDingusBitmap);
	return 1;
}

void guiShutdown()
{
	guiXDone = 1;
}

void guiXShutdown()
{
	/* The X server will destroy everything when we do this... */
	XCloseDisplay(guiXDisplay);
}

int guiXFindShiftAndBits(unsigned long mask, int *shiftP, int *bitsP)
{
	int shift = 0;
	int bits = 0;	
	while (!(mask & 1)) {
		shift++;
		mask >>= 1;
		if (!mask) {
			fprintf(stderr, "Can't decode image mask -- don't understand your display\n");
			return 0;
		}
	}	
	while (mask & 1) {
		bits++;
		mask >>= 1;
	}
	*shiftP = shift;
	*bitsP = bits;
	return 1;
}

guiSize guiGetScreenSize()
{
	guiSize s;
	s.w = DisplayWidth(guiXDisplay, guiXScreen);
	s.h = DisplayHeight(guiXDisplay, guiXScreen);
	return s;
}

void guiFatalError(char *fmt, ...)
{
	va_list v;
	va_start(v, fmt);
	fprintf(stderr, "FATAL ERROR: ");
	vfprintf(stderr, fmt, v);
	exit(1);
}

void guiError(char *fmt, ...)
{
	va_list v;
	va_start(v, fmt);
	fprintf(stderr, "ERROR: ");
	vfprintf(stderr, fmt, v);
}

int guiXNow = 0;

int guiXFetchEvent();

static struct timeval startStruct;

void guiMainLoop() {
	struct timezone dummy;
	gettimeofday(&startStruct, &dummy);
	while(!guiXDone) {
		if (!guiXFetchEvent()) {
			guiXDone = 1;
		}
	}
	guiXShutdown();
}

int guiXProcessEvent(XEvent *e);

void guiCanvasDoModelessDialog(guiCanvasPtr c)
{
	guiSize s = guiCanvasGetPhysicalSize(c);
	guiSize ss = guiGetScreenSize(c);
	guiPoint p = guiGetMouseLocation();
	p.x -= s.w / 2;
	p.y -= s.h / 2; 
	if (p.x + s.w > ss.w) {
		p.x = ss.w - s.w;
	}	
	if (p.y + s.h > ss.h) {
		p.y = ss.h - s.h;
	}
	if (p.x < 0) {
		p.x = 0;
	}
	if (p.y < 0) {
		p.y = 0;
	}
	guiCanvasMove(c, p.x, p.y);
	guiCanvasShow(c, 1);
}

void guiCanvasDoDialog(guiCanvasPtr c)
{
	guiCanvasDoModelessDialog(c);
	guiCanvasModal(c);
}

void guiCanvasModal(guiCanvasPtr c)
{
	guiXModalItemPtr p = (guiXModalItemPtr) malloc(sizeof(guiXModalItem));
	int ourDepth;
	p->c = c;
	p->next = guiXModalFirst;
	guiXModalFirst = p;
	guiXModalDepth++;
	ourDepth = guiXModalDepth;
	while(!guiXDone) {
		if (!guiXFetchEvent()) {
			guiXDone = 1;
		}
		if (guiXModalDepth != ourDepth) {
			break;
		}
	}
}

int guiXFetchEvent() {
	fd_set readfds;
	struct timeval tv;
	struct timeval nowStruct;
	struct timezone dummy;
	int delay;
	int now;
	int timerPresent = 0;
	int i;
	int input;
	int elapsed;
	XEvent event;
	int found = 0;
	if (guiXDone) {
		return 0;
	}
	/* 1. Discard all but the last MotionNotify event;
		process the last if there were any. */
	while (XCheckMaskEvent(guiXDisplay, ButtonMotionMask, &event)) {
		found = 1;
	}
	if (found) {
		if (!guiXProcessEvent(&event)) {
			return 0;
		}
	}
	/* 2. Process other events normally. */
	while (XCheckMaskEvent(guiXDisplay, ~0 - ButtonMotionMask, &event)) {
		if (!guiXProcessEvent(&event)) {
			return 0;
		}
	}
	for (i = 0; (i < guiXTimersTotal); i++) {
		if (guiXTimers[i].canvas) {
			if (guiXModalFirst && 
				(guiCanvasGetRootParent(guiXTimers[i].canvas) 
					!= guiXModalFirst->c)) 
			{
				continue;
			}
			if ((!timerPresent) || (guiXTimers[i].delay < delay)) {
				delay = guiXTimers[i].delay;
				timerPresent = 1;
			}
		}
	}
	if (timerPresent) {
		tv.tv_sec = delay / 1000;
		tv.tv_usec = (delay % 1000) * 1000;
	}
	FD_ZERO(&readfds);
	FD_SET(ConnectionNumber(guiXDisplay), &readfds);
	if (timerPresent) {
		input = select(ConnectionNumber(guiXDisplay) + 1,
			&readfds, 0, 0, &tv);			
	} else {
		input = select(ConnectionNumber(guiXDisplay) + 1,
			&readfds, 0, 0, 0);
	}
	gettimeofday(&nowStruct, &dummy);
	nowStruct.tv_sec -= startStruct.tv_sec;
	now = nowStruct.tv_usec / 1000 + nowStruct.tv_sec * 1000;	
	if (now < guiXNow) {
		/* Midnight crossover. Compensate. */
		guiXNow -= 86400000;
	}
	elapsed = now - guiXNow;
	guiXNow = now;
	for (i = 0; (i < guiXTimersTotal); i++) {
		if (guiXTimers[i].canvas) {
			if (guiXModalFirst && 
				(guiCanvasGetRootParent(guiXTimers[i].canvas) 
					!= guiXModalFirst->c)) 
			{
				continue;
			}			
			guiXTimers[i].delay -= elapsed;
			if (guiXTimers[i].delay <= 0) {
				guiEvent e;
				if (guiXTimers[i].id == guiXScrollbarTimer) {
					guiCanvasPtr c = guiXTimers[i].canvas;
					guiXTimers[i].canvas = 0;
					guiXCanvasScrollbarAct();
					guiXCanvasSetTimerBody(
						c,
						guiXScrollbarTimer, 100);
				} else {
					guiCanvasPtr c = guiXTimers[i].canvas;
					e.t = guiEventTimer;
					e.timerid = guiXTimers[i].id;
					guiXTimers[i].canvas = 0;
					c -> eventHandler(c, &e);
				}
			}
		}
	}
	return 1;
}

int guiXProcessEvent(XEvent *ep)
{
	guiCanvasPtr c;
	if (XFindContext(guiXDisplay, ep->xany.window, guiXContext,
		(caddr_t*) &c) != 0) {
		return 1;
	}
	if (c->dying) {
		if (ep->type == DestroyNotify) {
			canvasDestroyed(c);
			return 1;
		} else {
			/* X server is trying to drive me insane */	
			return 1;
		}
	}
	switch (ep->type) {
		case MappingNotify:
			XRefreshKeyboardMapping((XMappingEvent*)ep);
			break;	
		case Expose:
		case GraphicsExpose:
		{
			XExposeEvent *ee = (XExposeEvent*)ep;
			/* Add to invalid rectangle using virtual coordinates */
			guiRectangle r;
			r.x = ee->x + c->visibleTop.x - c->border;
			r.y = ee->y + c->visibleTop.y - c->border;
			r.w = ee->width;
			r.h = ee->height; 
			if (c->barV) {
				if (ee->x + ee->width > (c->physicalSize.w -
					guiVerticalScrollbarWidth - c->border)) 
				{
					c->repaintV = 1;
				}	
			}
			if (c->barH) {
				if (ee->y + ee->height > (c->physicalSize.h -
					guiHorizontalScrollbarHeight - c->border)) 
				{
					c->repaintH = 1;
				}	
			}			
			if (c->border) {
				if ((ee->x < c->border) || ((ee->x + ee->width)
					>= (c->physicalSize.w - c->border)) ||
					(ee->y < c->border) || 
					((ee->y + ee->height)
					>= (c->physicalSize.h - c->border))) 
				{
					c->repaintB = 1;
				}
			}
			if (!ee->count) {
				guiXCanvasInvalidateBody(c, &r, 1);
			} else {
				guiXCanvasInvalidateBody(c, &r, 0);
			}
			break;
		}
		case ButtonPress:
		{
			int skip = 0;
			XButtonEvent *be = (XButtonEvent *)ep;
			guiEvent e;
			guiXMouseDownCanvas = c;
			/* In case the ButtonRelease event gets lost */
			if (guiXScrollbarButtonCanvas) {
				guiXCanvasScrollbarMouseUp();
				break;
			}				
			if (guiXModalFirst && 
				(guiCanvasGetRootParent(c)
					!= guiXModalFirst->c)) 
			{
				break;
			}			
			if (c->barV) {
				guiSize s = guiCanvasGetVisibleSize(c);
				if ((be->x > (s.w + c->border)) && (be->x < c->physicalSize.w)
					 && (be->y > c->border) && 
					(be->y < c->physicalSize.h)) 
				{
					guiXCanvasScrollbarVerticalMouseDown(
						c, be->x, be->y);
					break;
				}
			}	
			if (c->barH) {
				guiSize s = guiCanvasGetVisibleSize(c);
				if ((be->y > (s.h + c->border)) && (be->y < c->physicalSize.h)
					 && (be->x > 0) && 
					(be->x < c->physicalSize.w)) 
				{
					guiXCanvasScrollbarHorizontalMouseDown(
						c, be->x, be->y);
					break;
				}
			}				
			e.t = guiEventMouseDown;
			e.l.x = be->x + c->visibleTop.x - c->border;
			e.l.y = be->y + c->visibleTop.y - c->border;
			switch(be->button) {
				case Button1:	
				e.button = guiButtonLeft;
				break;
				case Button2:	
				e.button = guiButtonMiddle;
				break;
				case Button3:	
				e.button = guiButtonRight;
				break;
				default:
				skip = 1;
				break;
			}
			guiXButtonsDown |= e.button;
			if (skip) {
				break;
			}
			if (c->eventMask & guiEventMouseDown) {
				c->eventHandler(c, &e);
			}
			break;
		}
		case ButtonRelease:
		{
			XButtonEvent *be = (XButtonEvent *)ep;
			guiEvent e;
			int skip = 0;
			if (guiXMouseDownCanvas && 
				(((Window) guiXMouseDownCanvas->handle) 
				!= be->window)) 
			{
				/* The window that got the original
					ButtonPress is also entitled
					to know about it when the
					user lets go. */
				Window dummy;
				XEvent copy = *ep;
				XButtonEvent *be = (XButtonEvent *)
					&copy;
				be->window = (Window)
					guiXMouseDownCanvas->handle;
				if (guiCanvasIsVisible(guiXMouseDownCanvas)) {
					XTranslateCoordinates(guiXDisplay, 
						(Window) c->handle,
						(Window) guiXMouseDownCanvas->handle,
						be->x,
						be->y,
						&be->x, 
						&be->y, 
						&dummy);
				}
				if (!guiXMouseDownCanvas->dying) {
					XSendEvent(guiXDisplay,
						(Window) guiXMouseDownCanvas->handle,
						True, ButtonReleaseMask, &copy);
				}
			}
			if (guiXScrollbarButtonCanvas) {
				guiXCanvasScrollbarMouseUp();
				break;
			}				
			e.t = guiEventMouseUp;
			e.l.x = be->x + c->visibleTop.x - c->border;
			e.l.y = be->y + c->visibleTop.y - c->border;
			switch(be->button) {
				case Button1:	
				e.button = guiButtonLeft;
				break;
				case Button2:	
				e.button = guiButtonMiddle;
				break;
				case Button3:	
				e.button = guiButtonRight;
				break;
				default:
				skip = 1;
				break;
			}			
			guiXButtonsDown &= ~e.button;
			if (skip) {
				break;
			}
			if (c->eventMask & guiEventMouseUp) {
				c->eventHandler(c, &e);
			}
			break;		
		}
		case MotionNotify:
		{
			XMotionEvent *me = (XMotionEvent *)ep;
			int i;
			int drag = 0;
			int m = Button1Mask;
			guiEvent e;
			e.t = guiEventMouseMove;
			e.l.x = me->x + c->visibleTop.x - c->border;
			e.l.y = me->y + c->visibleTop.y - c->border;
			if (c == guiXScrollbarButtonCanvas) {
				guiXCanvasScrollbarMouseDrag(me->x, me->y);
				break;
			}
			if (guiXModalFirst && 
				(guiCanvasGetRootParent(c) 
					!= guiXModalFirst->c)) 
			{
				break;
			}			
			if (c->eventMask & guiEventMouseDrag) {
				for (i = 1; (i < 4); i++) {
					if (me->state & m) {
						e.button = i;
						e.t = guiEventMouseDrag;
						c->eventHandler(c, &e);
						drag = 1;
					}
					m <<= 1;
				
				}
			}
			if (drag) {
				break;
			}
			if (c->eventMask & guiEventMouseMove) {
				e.t = guiEventMouseMove;
				c->eventHandler(c, &e);
			}
			break;
		}
		case DestroyNotify:
		{
			canvasDestroyed(c);
			break;	
		}
		case ConfigureNotify:
		{
			XConfigureEvent *ce = (XConfigureEvent *)ep;
			guiEvent e;
			guiSize s;
			c->l.x = ce->x;
			c->l.y = ce->y;
			if ((ce->width != c->physicalSize.w) ||
				(ce->height != c->physicalSize.h)) {
				e.t = guiEventResize;
				e.s.w = ce->width;
				e.s.h = ce->height;
				c->physicalSize.w = e.s.w;
				c->physicalSize.h = e.s.h;
				if (c->virtualized) {
					guiXCanvasCheckScrollbars(c);
				}
				if (c->eventMask & guiEventResize) {
					c->eventHandler(c, &e);
				}
				s = guiCanvasGetVisibleSize(c);
				if ((s.w > 0) && (s.h > 0)) {
					if (c->clearing) {
						XClearArea(guiXDisplay, 
							(Window) c->handle,
							c->border, c->border, 
							s.w, s.h, 1);
					} else {
						guiCanvasInvalidate(c, 0);
					}
				}
			}
			break;	
		}
		case KeyPress:
		{
			XKeyEvent *ke = (XKeyEvent *)ep;
			guiEvent e;
			char ch;
			KeySym ks;
			static XComposeStatus cs;
			int length;
			e.t = guiEventKeyDown;
			e.l.x = ke->x + c->visibleTop.x - c->border;
			e.l.y = ke->y + c->visibleTop.y - c->border;
			length = XLookupString(ke, &ch, 1, &ks, &cs);
			if (IsModifierKey(ks)) {
				break;
			}
			e.shift = (ke->state & ShiftMask)
				? 1 : 0;
			e.control = (ke->state & ControlMask)
				? 1 : 0;
			e.alt = (ke->state & Mod1Mask)
				? 1 : 0;
			/* Struggle mightily to catch the alt key */
			if (!e.alt) {
				e.alt = (ke->state & Mod2Mask)
					? 1 : 0;
			}
			if (!e.alt) {
				e.alt = (ke->state & Mod3Mask)
					? 1 : 0;
			}
			if (!e.alt) {
				e.alt = (ke->state & Mod4Mask)
					? 1 : 0;
			}
			if (!e.alt) {
				e.alt = (ke->state & Mod5Mask)
					? 1 : 0;
			}
			if (!length) {
				ch = 0;
			}
			switch (ks) {
				case XK_Left:
				e.key = guiKeyLeft;	
				break;
				case XK_Right:
				e.key = guiKeyRight;
				break;
				case XK_Down:
				e.key = guiKeyDown;
				break;
				case XK_Up:
				e.key = guiKeyUp;
				break;
				case XK_Insert:
				e.key = guiKeyInsert;
				break;
				case XK_Delete:
				e.key = guiKeyDelete;
				break;
				case XK_Home:
				e.key = guiKeyHome;
				break;
				case XK_End:
				e.key = guiKeyEnd;
				break;
				case XK_Prior:
				e.key = guiKeyPageUp;
				break;
				case XK_Next:
				e.key = guiKeyPageDown;
				break;
				case XK_F1:
				e.key = guiKeyF1;
				break;
				case XK_F2:
				e.key = guiKeyF2;
				break;
				case XK_F3:
				e.key = guiKeyF3;
				break;
				case XK_F4:
				e.key = guiKeyF4;
				break;
				case XK_F5:
				e.key = guiKeyF5;
				break;
				case XK_F6:
				e.key = guiKeyF6;
				break;
				case XK_F7:
				e.key = guiKeyF7;
				break;
				case XK_F8:
				e.key = guiKeyF8;
				break;
				case XK_F9:
				e.key = guiKeyF9;
				break;
				case XK_F10:
				e.key = guiKeyF10;
				break;
				case XK_F11:
				e.key = guiKeyF11;
				break;
				case XK_F12:
				e.key = guiKeyF12;
				break;
				case XK_BackSpace:
				e.key = guiKeyBackspace;
				break;
				case XK_Escape:
				e.key = guiKeyEscape;
				break;
				case XK_Tab:
				e.key = guiKeyTab;
				break;				
				case XK_space:
				e.key = guiKeySpace;
				break;				
				default:
				e.key = ch;
				break;
			}
			if (!guiXCanvasResolveKey(&c, &e)) {
				/* Default keyboard support for scrolling 
					and tabbing. */
				if (guiXModalFirst && 
					(guiCanvasGetRootParent(c)
						!= guiXModalFirst->c)) 
				{
					break;
				}			
				switch(e.key) {
					case guiKeyLeft:
					guiCanvasTakeFocus(c);
					guiCanvasScrollHorizontal(c, 
						- guiScrollbarStepPixels);
					break;
					guiCanvasTakeFocus(c);
					case guiKeyRight:
					guiCanvasScrollHorizontal(c, 
						guiScrollbarStepPixels);
					break;
					guiCanvasTakeFocus(c);
					case guiKeyUp:
					guiCanvasScrollVertical(c, 
						- guiScrollbarStepPixels);
					break;
					case guiKeyDown:
					guiCanvasTakeFocus(c);
					guiCanvasScrollVertical(c, 
						guiScrollbarStepPixels);
					break;
					case guiKeyPageUp:
					guiCanvasTakeFocus(c);
					guiCanvasScrollVertical(c, 
						- guiCanvasGetVisibleSize(c).h);
					break;
					case guiKeyPageDown:
					guiCanvasTakeFocus(c);
					guiCanvasScrollVertical(c, 
						guiCanvasGetVisibleSize(c).h);
					break;
					case guiKeyHome:
					guiCanvasTakeFocus(c);
					guiCanvasScrollToVertical(c, 0);
					break;
					case guiKeyEnd:
					guiCanvasTakeFocus(c);
					guiCanvasScrollToVertical(c, 
						c->virtualSize.h - 
						guiCanvasGetVisibleSize(c).h);
					break;
					case guiKeyTab:
					if (!c->tabbing) {
						break;
					}
					/* Dialog tabbing support. */
					/* No currently focused child, 
						so go to the first child 
						(or grandchild...) that
						wants the focus */
					if (!c->childFocused) {
						guiXCanvasFirstFocus(c);
						break;
					}
					/* Forward tab */
					if (!e.shift) {
						guiCanvasPtr s, p;
						s = c->childFocused;
						p = s;
						while (1) {
							while (!p->next) {
								p = p->parent;
								if (p == c) {
									p = 0;
									break;
								}
							}	
							if (p) {
								p = p->next;
							} else {
								p = c->first;
							}
							if (p == s) {
								break;
							}
tryForFocus:
							if (p->wantFocus) {
								guiCanvasTakeFocus(p);	
								return 1;
							} else {
								if (p->first) {
									p = p->first;
									goto tryForFocus;
								}
							}
						} while (p != s);
						/* There's nothing new under the sun */
						break;
					} else { /* shift-tab */
						guiCanvasPtr p, prev;
						p = c->first;
						if (!p) {
							return 1;
						}
						prev = 0;
						while (p != c->childFocused) {
							if (p->wantFocus) {
								prev = p;
							} else if (p->first) {
								p = p->first;
								continue;
							}
							while (!p->next) {
								p = p->parent;
								if (p == c) {
									return 0;
								}
							}	
							if (p) {
									p = p->next;
							}
						}
						if (prev) {
							guiCanvasTakeFocus(prev);	
							return 1;
						}
						/* Nothing before it, so look
							for the last item
							that wants the focus */
						prev = 0;
						p = c->childFocused;
						while (p) {
							if (p->wantFocus) {
								prev = p;
							} else if (p->first) {
								p = p->first;
								continue;
							}
							while (!p->next) {
								p = p->parent;
								if (p == c) {
									break;
								}
							}	
							if (p) {
								p = p->next;
							}
						}
						if (prev) {
							guiCanvasTakeFocus(prev);	
							break;
						}
					}	
					/* Consume the key */
					break;
					default:
					break;
				}
			}
			break;
		}
		case EnterNotify:
		{
			guiEvent e;
			e.t = guiEventMouseEnter;
			if (guiXModalFirst && 
				(guiCanvasGetRootParent(c)
					!= guiXModalFirst->c)) 
			{
				break;
			}			
			if (c->eventMask & guiEventMouseEnter) {
				c->eventHandler(c, &e);
			}
			break;
		}
		case LeaveNotify:
		{
			guiEvent e;
			e.t = guiEventMouseExit;
			if (guiXModalFirst && 
				(guiCanvasGetRootParent(c)
					!= guiXModalFirst->c)) 
			{
				break;	
			}			
			if (c->eventMask & guiEventMouseExit) {
				c->eventHandler(c, &e);
			}
			break;
		}
		case FocusIn:
		{
			XFocusChangeEvent *ce = (XFocusChangeEvent *)ep;
			if (guiXModalFirst && 
				(guiCanvasGetRootParent(c)
					!= guiXModalFirst->c)) 
			{
				break;
			}			
			if ((ce->mode == NotifyNormal) &&
				(ce->detail == NotifyNonlinear)) 
			{
				if (!c->parent) {
					if (c->childFocused) {
						guiCanvasTakeFocus(
							c->childFocused);
					} else {
						guiXCanvasFirstFocus(c);
					}
				}
			}
			break;
		}
		default:
		break;
	}
	return 1;
}

void guiXCanvasRepaintVerticalScrollbar(guiCanvasPtr c)
{
	int s;
	int x, y, w, h;
	int p1;
	int a;
	guiSize vs;
	vs = guiCanvasGetVisibleSize(c);
	p1 = c->visibleTop.y * (c->physicalSize.h - c->border * 2 -
		guiHorizontalScrollbarHeight * 3) / c->virtualSize.h +
		c->border;
	s = vs.h * (c->physicalSize.h - c->border * 2 -
		guiHorizontalScrollbarHeight * 3) / c->virtualSize.h;
	x = c->physicalSize.w - guiVerticalScrollbarWidth - c->border;
	y = p1;
	w = guiVerticalScrollbarWidth;
	h = s;
	/* Use the raw X drawing calls to avoid being clipped
		out of the scrollbar area */
	XFillRectangle(guiXDisplay, (Drawable) c->handle, 
		guiXScrollbarBackgroundGC,
		x, guiHorizontalScrollbarHeight, w, y);
	XFillRectangle(guiXDisplay, (Drawable) c->handle, 
		guiXScrollbarBackgroundGC,		
		x, y + h + guiHorizontalScrollbarHeight,
		w, c->physicalSize.h - (y + h) - c->border -
			guiHorizontalScrollbarHeight * 3);
	XCopyArea(guiXDisplay, 
		(Drawable) guiXThumbs[0][0][0]->handle,
		(Drawable) c->handle, 
		guiXScrollbarBodyGC,
		0, 0,
		guiXThumbs[0][0][0]->physicalSize.w, 
		guiXThumbs[0][0][0]->physicalSize.h,
		x, y - guiXThumbs[0][0][0]->physicalSize.h + 
		guiHorizontalScrollbarHeight);
	XCopyArea(guiXDisplay, 
		(Drawable) guiXThumbs[0][2][0]->handle,
		(Drawable) c->handle, 
		guiXScrollbarBodyGC,
		0, 0,
		guiXThumbs[0][2][0]->physicalSize.w, 
		guiXThumbs[0][2][0]->physicalSize.h,
		x, y + h + guiHorizontalScrollbarHeight);
	a = h;	
	while (a) {
		int step = guiMin(a, guiXThumbs[0][1][0]->physicalSize.h);
		XCopyArea(guiXDisplay, 
			(Drawable) guiXThumbs[0][1][0]->handle,
			(Drawable) c->handle, 
			guiXScrollbarBodyGC,
			0, 0,
			guiXThumbs[0][1][0]->physicalSize.w, step,
			x, y + h - a + guiHorizontalScrollbarHeight);
		a -= step;
	}
	XCopyArea(guiXDisplay, 
		(Drawable) guiXArrows[(guiXScrollbarButtonCanvas && 
			(guiXScrollbarButtonType == guiXDownArrow)) 
			? 1 : 0][2] -> handle,	
		(Drawable) c->handle, 
		guiXScrollbarBodyGC,
		0, 0,
		guiVerticalScrollbarWidth, guiHorizontalScrollbarHeight,
		x, c->physicalSize.h - guiHorizontalScrollbarHeight * 2 -
		c->border);
	XCopyArea(guiXDisplay,
		(Drawable) guiXArrows[ (guiXScrollbarButtonCanvas && 
			(guiXScrollbarButtonType == guiXUpArrow)) 
			? 1 : 0][0] -> handle,
		(Drawable) c->handle, 		
		guiXScrollbarBodyGC,
		0, 0,
		guiVerticalScrollbarWidth, guiHorizontalScrollbarHeight,
		x, c->border);
}

void guiXCanvasRepaintHorizontalScrollbar(guiCanvasPtr c)
{
	int s;
	int x, y, w, h;
	int p1;
	guiSize vs;
	int a;
	vs = guiCanvasGetVisibleSize(c);
	p1 = c->visibleTop.x * (c->physicalSize.w - 
		guiVerticalScrollbarWidth * 3 - c->border * 2) 
		/ c->virtualSize.w + c->border;
	s = vs.w * (c->physicalSize.w -
		guiVerticalScrollbarWidth * 3 - c->border * 2) 
		/ c->virtualSize.w;
	x = p1;
	y = c->physicalSize.h - guiHorizontalScrollbarHeight - c->border;
	w = s;
	h = guiHorizontalScrollbarHeight;
	/* Use the raw X drawing calls to avoid being clipped
		out of the scrollbar area */
	XFillRectangle(guiXDisplay, (Drawable) c->handle, 
		guiXScrollbarBackgroundGC,
		guiVerticalScrollbarWidth, y, x, h);
	XFillRectangle(guiXDisplay, (Drawable) c->handle, 
		guiXScrollbarBackgroundGC,		
		x + w + guiVerticalScrollbarWidth, y, 
		c->physicalSize.w - (x + w) - 
			guiVerticalScrollbarWidth * 3 - c->border, h);
	XCopyArea(guiXDisplay, 
		(Drawable) guiXThumbs[1][0][0]->handle,
		(Drawable) c->handle, 
		guiXScrollbarBodyGC,
		0, 0,
		guiXThumbs[1][0][0]->physicalSize.w, 
		guiXThumbs[1][0][0]->physicalSize.h,
		x - guiXThumbs[1][0][0]->physicalSize.w + 
		guiVerticalScrollbarWidth, y);
	XCopyArea(guiXDisplay, 
		(Drawable) guiXThumbs[1][2][0]->handle,
		(Drawable) c->handle, 
		guiXScrollbarBodyGC,
		0, 0,
		guiXThumbs[1][2][0]->physicalSize.w, 
		guiXThumbs[1][2][0]->physicalSize.h,
		x + w + guiVerticalScrollbarWidth, y);
	a = w;	
	while (a) {
		int step = guiMin(a, guiXThumbs[1][1][0]->physicalSize.w);
		XCopyArea(guiXDisplay, 
			(Drawable) guiXThumbs[1][1][0]->handle,
			(Drawable) c->handle, 			
			guiXScrollbarBodyGC,
			0, 0,
			step, guiXThumbs[1][1][0]->physicalSize.h,
			x + w - a + guiVerticalScrollbarWidth, y);
		a -= step;
	}
	XCopyArea(guiXDisplay, 
		(Drawable) guiXArrows[(guiXScrollbarButtonCanvas && 
			(guiXScrollbarButtonType == guiXRightArrow)) 
			? 1 : 0][1] -> handle,
		(Drawable) c->handle, 
		guiXScrollbarBodyGC,
		0, 0,
		guiVerticalScrollbarWidth, guiHorizontalScrollbarHeight,
		c->physicalSize.w - guiVerticalScrollbarWidth * 2 - c->border, 
		y);
	XCopyArea(guiXDisplay, 
		(Drawable) guiXArrows[(guiXScrollbarButtonCanvas && 
			(guiXScrollbarButtonType == guiXLeftArrow)) 
			? 1 : 0][3] -> handle,
		(Drawable) c->handle, 
		guiXScrollbarBodyGC,
		0, 0,
		guiVerticalScrollbarWidth, guiHorizontalScrollbarHeight,
		c->border, y);
}

void guiXCanvasRepaintBorder(guiCanvasPtr c)
{
	if (c->dying) {
		return;
	}
	if (c->border <= 0) {
		return;
	}
	XClearArea(guiXDisplay,
		(Drawable) c->handle,
		0, 0,
		c->physicalSize.w, c->border,
		0);
	if (c->physicalSize.h - c->border * 2 > 0) {
		XClearArea(guiXDisplay,
			(Drawable) c->handle,
			0, c->border,
			c->border, c->physicalSize.h - c->border * 2,
			0);
		XClearArea(guiXDisplay,
			(Drawable) c->handle,
			c->physicalSize.w - c->border, c->border,
			c->border, c->physicalSize.h - c->border * 2,
			0);
	}	
	XClearArea(guiXDisplay,
		(Drawable) c->handle,
		0, c->physicalSize.h - c->border,
		c->physicalSize.w, c->border,
		0);
	if (c->border >= 2) {
		switch(c->borderStyle) {
			case guiBorderIn:
			guiXDrawFrame(c,
				0,
				0, 0, 
				c->physicalSize.w, c->physicalSize.h);
			break;
			case guiBorderOut:
			guiXDrawFrame(c,
				1,
				0, 0, 
				c->physicalSize.w, c->physicalSize.h);
			break;
			default:
			break;
		}
	}
}

void guiCanvasScrollToVertical(guiCanvasPtr c, int pos)
{
	guiSize s;
	int dy;
	guiRectangle r;
	if (c->dying) {
		return;
	}
	if (!c->virtualized) {
		return;
	}
	if (guiCanvasGetVisibleSize(c).h > c->virtualSize.h) {
		return;
	}	
	dy = c->visibleTop.y; 
	c->visibleTop.y = pos;
	s = guiCanvasGetVisibleSize(c);
	if (c->visibleTop.y + s.h > c->virtualSize.h) {
		c->visibleTop.y = c->virtualSize.h - s.h;
	}
	if (c->visibleTop.y < 0) {
		c->visibleTop.y = 0;
	}
	if (!c->noScrollV) {
		c->repaintV = 1;
	}
	dy -= c->visibleTop.y;
	if (!dy) {
		return;
	}
	if (dy > 0) {
		if ((s.w > 0) && ((s.h - dy) > 0)) { 
			XCopyArea(guiXDisplay, (Window) c->handle, 
				(Window) c->handle,
				guiXScrollbarBodyGC, c->border, c->border, 
				s.w, s.h - dy, c->border, c->border + dy);	
		}
	} else {
		if ((s.w > 0) && ((s.h + dy) > 0)) { 
			XCopyArea(guiXDisplay, (Window) c->handle, 
				(Window) c->handle,
				guiXScrollbarBodyGC, c->border, c->border - dy, 
				s.w, s.h + dy, c->border, c->border);	
		}
	}
	if (dy > 0) {
		r.y = 0 + c->visibleTop.y;
	} else { 
		r.y = s.h + dy + c->visibleTop.y;
	}
	r.h = abs(dy);
	r.x = 0 + c->visibleTop.x;
	r.w = s.w;
	if ((r.w > 0) && (r.h > 0)) {
		XClearArea(guiXDisplay, (Window) c->handle,
			r.x - c->visibleTop.x + c->border,	
			r.y - c->visibleTop.y + c->border,	
			r.w, r.h, 0);
	}
	guiCanvasInvalidate(c, &r);
}

void guiCanvasScrollToHorizontal(guiCanvasPtr c, int pos)
{
	guiSize s;
	int dx;
	guiRectangle r;
	if (!c->virtualized) {
		return;
	}
	if (guiCanvasGetVisibleSize(c).w > c->virtualSize.w) {
		return;
	}	
	dx = c->visibleTop.x; 
	c->visibleTop.x = pos;
	s = guiCanvasGetVisibleSize(c);
	if (c->visibleTop.x + s.w > c->virtualSize.w) {
		c->visibleTop.x = c->virtualSize.w - s.w;
	}
	if (c->visibleTop.x < 0) {
		c->visibleTop.x = 0;
	}
	if (!c->noScrollH) {
		c->repaintH = 1;
	}
	dx -= c->visibleTop.x;
	if (!dx) {
		return;
	}
	if (dx > 0) {
		if (((s.w - dx) > 0) && (s.h > 0)) { 
			XCopyArea(guiXDisplay, (Window) c->handle, 
				(Window) c->handle, guiXScrollbarBodyGC, 
				c->border, c->border, 
				s.w - dx, s.h, c->border + dx, c->border);	
		}
	} else {
		if (((s.w + dx) > 0) && (s.h > 0)) { 
			XCopyArea(guiXDisplay, (Window) c->handle, 
				(Window) c->handle, guiXScrollbarBodyGC, 
				c->border - dx, c->border, s.w + dx, s.h, 
				c->border, c->border);	
		}
	}
	if (dx > 0) {
		r.x = 0 + c->visibleTop.x;
	} else { 
		r.x = s.w + dx + c->visibleTop.x;
	}
	r.w = abs(dx);
	r.y = 0 + c->visibleTop.y;
	r.h = s.h;
	if ((r.w > 0) && (r.h > 0)) {
		XClearArea(guiXDisplay, (Window) c->handle,
			r.x - c->visibleTop.x + c->border,	
			r.y - c->visibleTop.y + c->border,	
			r.w, r.h, 0);
	}
	guiCanvasInvalidate(c, &r);
}

void guiCanvasScrollTo(guiCanvasPtr c, int x, int y)
{
	guiSize s;
	int x1, y1, x2, y2, w, h;
	int dx, dy;
	guiRectangle r;
	if (!c->virtualized) {
		return;
	}
	dx = c->visibleTop.x;
	dy = c->visibleTop.y;
	if ((guiCanvasGetVisibleSize(c).w > c->virtualSize.w) &&
		(guiCanvasGetVisibleSize(c).h > c->virtualSize.h)) 
	{
		return;
	}	
	s = guiCanvasGetVisibleSize(c);
	if (c->visibleTop.y != y) {
		if (guiCanvasGetVisibleSize(c).h < c->virtualSize.h) {
			c->visibleTop.y = y;
			if (!c->noScrollV) {
				c->repaintV = 1;
			}
			if (c->visibleTop.y + s.h > c->virtualSize.h) {
				c->visibleTop.y = c->virtualSize.h - s.h;
			}
			if (c->visibleTop.y < 0) {
				c->visibleTop.y = 0;
			}
		}
	}
	if (c->visibleTop.x != x) {
		if (guiCanvasGetVisibleSize(c).w < c->virtualSize.w) {
			c->visibleTop.x = x;
			if (!c->noScrollH) {
				c->repaintH = 1;
			}
			if (c->visibleTop.x + s.w > c->virtualSize.w) {
				c->visibleTop.x = c->virtualSize.w - s.w;
			}
			if (c->visibleTop.x < 0) {
				c->visibleTop.x = 0;
			}
		}
	}
	dx -= c->visibleTop.x;
	dy -= c->visibleTop.y;
	/* 1. Translate the coordinates of the existing visible area. */
	x1 = dx;
	y1 = dy;
	x2 = x1 + s.w;
	y2 = y1 + s.h;
	/* 2. Clip the parts no longer visible. */
	x1 = guiBound(x1, 0, s.w);	
	y1 = guiBound(y1, 0, s.h);	
	x2 = guiBound(x2, 0, s.w);	
	y2 = guiBound(y2, 0, s.h);	
	w = x2 - x1;
	h = y2 - y1;
	/* 3. Copy the visible part from its old location. */
	if (w && h) {
		XCopyArea(guiXDisplay, (Window) c->handle, 
			(Window) c->handle, guiXScrollbarBodyGC, 
			x1 - dx + c->border, y1 - dy + c->border, 
			w, h, 
			x1 + c->border, y1 + c->border);	
	}	
	/* 4. Invalidate everything but the region copied to. */
	if (x1 > 0) {
		r.x = c->visibleTop.x;
		r.y = c->visibleTop.y;
		r.w = x1;
		r.h = s.h;
		if ((x1 > 0) && (s.h > 0)) {
			XClearArea(guiXDisplay, (Window) c->handle,
				c->border, c->border, x1, s.h, 0); 
		}
		guiXCanvasInvalidateBody(c, &r, 1);
	}
	if (x2 < s.w) {
		r.x = c->visibleTop.x + x2;
		r.y = c->visibleTop.y;
		r.w = s.w - x2;
		r.h = s.h;
		if (((s.w - x2) > 0) && (s.h > 0)) {
			XClearArea(guiXDisplay, (Window) c->handle,
				x2 + c->border, c->border, s.w - x2, s.h, 0); 
		}
		guiXCanvasInvalidateBody(c, &r, 1);
	}	
	if (y1 > 0) {
		r.x = c->visibleTop.x;
		r.y = c->visibleTop.y;
		r.w = s.w;
		r.h = y1;
		if ((s.w > 0) && (y1 > 0)) {
			XClearArea(guiXDisplay, (Window) c->handle,
				c->border, c->border, s.w, y1, 0); 
		}
		guiXCanvasInvalidateBody(c, &r, 1);
	}
	if (y2 < s.h) {
		r.y = c->visibleTop.y + y2;
		r.x = c->visibleTop.x;
		r.w = s.w;
		r.h = s.h - y2;
		if ((s.w > 0) && ((s.h - y2) > 0)) {
			XClearArea(guiXDisplay, (Window) c->handle,
				x2 + c->border, c->border, s.w, s.h - y2, 0); 
		}
		guiXCanvasInvalidateBody(c, &r, 1);
	}
}

void guiCanvasScrollVertical(guiCanvasPtr c, int delta)
{
	guiCanvasScrollToVertical(c, c->visibleTop.y + delta);
}

void guiCanvasScrollHorizontal(guiCanvasPtr c, int delta)
{
	guiCanvasScrollToHorizontal(c, c->visibleTop.x + delta);
}

void guiCanvasScroll(guiCanvasPtr c, int dx, int dy)
{
	guiCanvasScrollTo(c, c->visibleTop.x + dx, c->visibleTop.y + dy);
}

void guiCanvasSetBackgroundColor(guiCanvasPtr c, guiColor b)
{
	XSetWindowAttributes a;
	guiColor bt;
	bt = b;
	a.background_pixel = guiXColorGet(&bt);
	c->backgroundColor = bt;
	XChangeWindowAttributes(guiXDisplay, (Window) c->handle,
		CWBackPixel, &a);
}

guiSize guiCanvasGetVisibleSize(guiCanvasPtr c)
{
	guiSize s;
	s.w = c->physicalSize.w - 
		(c->barV ? guiVerticalScrollbarWidth : 0) -
		c->border * 2;
	s.h = c->physicalSize.h - 
		(c->barH ? guiHorizontalScrollbarHeight : 0) -
		c->border * 2;
	return s;
}

void guiXCanvasScrollbarVerticalMouseDown(guiCanvasPtr c, int x, int y)
{
	guiXScrollbarButtonCanvas = c;
	guiCanvasTakeFocus(c);
	if (y < (guiHorizontalScrollbarHeight + c->border)) {
		guiXScrollbarButtonType = guiXUpArrow;
		guiXCanvasSetTimerBody(c, guiXScrollbarTimer, 500);
		guiXCanvasScrollbarAct();
	} else if (y > (c->physicalSize.h - 
			guiHorizontalScrollbarHeight * 2 - c->border)) {	
		guiXScrollbarButtonType = guiXDownArrow;
		guiXCanvasSetTimerBody(c, guiXScrollbarTimer, 500);
		guiXCanvasScrollbarAct();
	} else {
		int s;
		int p1;
		guiSize vs;
		vs = guiCanvasGetVisibleSize(c);
		p1 = c->visibleTop.y * (c->physicalSize.h - 
			guiHorizontalScrollbarHeight * 3 - c->border * 2) / 
			c->virtualSize.h + c->border;
		s = vs.h * (c->physicalSize.h -
			guiHorizontalScrollbarHeight * 3 - c->border * 2) / 
			c->virtualSize.h;
		if (y < (p1 + guiHorizontalScrollbarHeight)) {
			guiXScrollbarButtonType = guiXPageUp;
			guiXCanvasSetTimerBody(c, guiXScrollbarTimer, 1000);
			guiXCanvasScrollbarAct();
		} else if (y >= (p1 + s + guiHorizontalScrollbarHeight)) {
			guiXScrollbarButtonType = guiXPageDown;
			guiXCanvasSetTimerBody(c, guiXScrollbarTimer, 1000);
			guiXCanvasScrollbarAct();
		} else {
			guiXScrollbarButtonType = guiXVerticalThumb;
			guiXScrollbarDragOffset = y - p1;
		}
	}
}

void guiXCanvasScrollbarHorizontalMouseDown(guiCanvasPtr c, int x, int y)
{
	guiXScrollbarButtonCanvas = c;
	guiCanvasTakeFocus(c);
	if (x < (guiVerticalScrollbarWidth + c->border)) {
		guiXScrollbarButtonType = guiXLeftArrow;
		guiXCanvasSetTimerBody(c, guiXScrollbarTimer, 1000);
		guiXCanvasScrollbarAct();
	} else if (x > (c->physicalSize.w - 
			guiVerticalScrollbarWidth * 2 - c->border)) {	
		guiXScrollbarButtonType = guiXRightArrow;
		guiXCanvasSetTimerBody(c, guiXScrollbarTimer, 1000);
		guiXCanvasScrollbarAct();
	} else {
		int s;
		int p1;
		guiSize vs;
		vs = guiCanvasGetVisibleSize(c);
		p1 = c->visibleTop.x * (c->physicalSize.w - 
			guiVerticalScrollbarWidth * 3 - c->border * 2) / 
			c->virtualSize.w + c->border;
		s = vs.w * (c->physicalSize.w -
			guiVerticalScrollbarWidth * 3 - c->border * 2) / 
			c->virtualSize.w;
		if (x < (p1 + guiVerticalScrollbarWidth)) {
			guiXScrollbarButtonType = guiXPageLeft;
			guiXCanvasSetTimerBody(c, guiXScrollbarTimer, 1000);
			guiXCanvasScrollbarAct();
		} else if (x >= (p1 + s + guiVerticalScrollbarWidth)) {
			guiXScrollbarButtonType = guiXPageRight;
			guiXCanvasSetTimerBody(c, guiXScrollbarTimer, 1000);
			guiXCanvasScrollbarAct();
		} else {
			guiXScrollbarButtonType = guiXHorizontalThumb;
			guiXScrollbarDragOffset = x - p1;
		}
	}
}

void guiXCanvasScrollbarAct()
{
	guiCanvasPtr c = guiXScrollbarButtonCanvas;
	switch(guiXScrollbarButtonType) {
		case guiXLeftArrow:
		guiCanvasScrollHorizontal(c, - guiScrollbarStepPixels);
		break;
		case guiXRightArrow:
		guiCanvasScrollHorizontal(c, guiScrollbarStepPixels);
		break;
		case guiXUpArrow:
		guiCanvasScrollVertical(c, - guiScrollbarStepPixels);
		break;
		case guiXDownArrow:
		guiCanvasScrollVertical(c, guiScrollbarStepPixels);
		break;
		case guiXPageUp:
		guiCanvasScrollVertical(c, - guiCanvasGetVisibleSize(c).h);
		break;
		case guiXPageDown:
		guiCanvasScrollVertical(c, guiCanvasGetVisibleSize(c).h);
		break;
		case guiXPageLeft:
		guiCanvasScrollHorizontal(c, - guiCanvasGetVisibleSize(c).w);
		break;
		case guiXPageRight:
		guiCanvasScrollHorizontal(c, guiCanvasGetVisibleSize(c).w);
		break;
		default:
		break;
	}
}

void guiXCanvasScrollbarMouseUp()
{
	guiCanvasPtr c = guiXScrollbarButtonCanvas; 
	guiCanvasClearTimer(guiXScrollbarButtonCanvas, guiXScrollbarTimer);
	guiXScrollbarButtonCanvas = 0;
	switch (guiXScrollbarButtonType) {
		case guiXUpArrow:
		case guiXDownArrow:
		case guiXPageUp:
		case guiXPageDown:
		case guiXVerticalThumb:
		guiXCanvasRepaintVerticalScrollbar(c);
		break;
		case guiXLeftArrow:
		case guiXRightArrow:
		case guiXPageLeft:
		case guiXPageRight:
		case guiXHorizontalThumb:
		guiXCanvasRepaintHorizontalScrollbar(c);
		break;
	}
}

void guiXCanvasScrollbarMouseDrag(int x, int y)
{
	guiCanvasPtr c = guiXScrollbarButtonCanvas;
	switch (guiXScrollbarButtonType) {
		case guiXUpArrow:
		case guiXDownArrow:
		case guiXLeftArrow:
		case guiXRightArrow:
		case guiXPageUp:
		case guiXPageDown:
		case guiXPageLeft:
		case guiXPageRight:
		break;
		case guiXVerticalThumb:
		guiCanvasScrollToVertical(c, 
			(y - guiXScrollbarDragOffset) 
			* c->virtualSize.h /
			(guiCanvasGetVisibleSize(c).h - 
				guiHorizontalScrollbarHeight * 2));
				
		break;
		case guiXHorizontalThumb:
		guiCanvasScrollToHorizontal(c, 
			(x - guiXScrollbarDragOffset) 
			* c->virtualSize.w /
			(guiCanvasGetVisibleSize(c).w -
				guiVerticalScrollbarWidth * 2));
		break;
	}
}

void guiCanvasSetClientData(guiCanvasPtr c, void *clientData)
{
	c->clientData = clientData;
}

void *guiCanvasGetClientData(guiCanvasPtr c)
{
	return c->clientData;
}

char *guiGetClipboard(int *pLength)
{
	int len;
	char *raw;
	char *result;
	raw = XFetchBuffer(guiXDisplay, &len, 0);	
	if (!raw) {
		result = (char *) malloc(1);	
		*result = '\0';
		if (pLength) {
			*pLength = 0;
		}
		return result;
	}
	result = (char *) malloc(len + 1);
	memcpy(result, raw, len);
	result[len] = '\0';	
	if (pLength) {
		*pLength = len;
	}
	return result;
}

void guiSetClipboard(char *data, int length)
{
	XStoreBuffer(guiXDisplay, data, length, 0);
}

int guiKeyIsPrintable(guiKey k)
{
	if (k > 31) {
		return 1;
	} else {
		return 0;
	}
}

void guiCanvasTakeFocus(guiCanvasPtr c)
{
	guiEvent e;
	guiCanvasPtr p;
	p = c->parent;
	if (!c->wantFocus) {
		return;
	}	
	while(1) {
		if (!p->parent) {
			p->childFocused = c;
			break;
		}		
		p = p->parent;
	}
	c->focused = 1;
	if (guiXFocusCanvas && (guiXFocusCanvas != c)) {
		guiXFocusCanvas->focused = 0;
		e.t = guiEventFocusOut;
		if (guiXFocusCanvas->eventMask & guiEventFocusOut) {
			guiXFocusCanvas->eventHandler(guiXFocusCanvas, &e);
		}
	}
	guiXFocusCanvas = c;
	e.t = guiEventFocusIn;
	if (c->eventMask & guiEventFocusIn) {
		c->eventHandler(c, &e);
	}
}

int guiCanvasPointIsVisible(guiCanvasPtr c, int x, int y)
{
	guiSize s;
	if (!c->virtualized) {
		return 1;
	}
	s = guiCanvasGetVisibleSize(c);
	if ((x >= c->visibleTop.x) && 
		(y >= c->visibleTop.y) &&
		(x < (c->visibleTop.x + s.w)) && 
		(y < (c->visibleTop.y + s.h)))
	{
		return 1;
	}
	return 0;	
}

static void guiXDrawFrame(guiCanvasPtr c, int state,
	int x, int y, int w, int h)
{
	XSegment segments[4], *s;
	GC gc;
	s = segments;
	s->x1 = x;
	s->y1 = y;
	s->x2 = x + w - 1;
	s->y2 = y;
	s++;
	s->x1 = x + 1;
	s->y1 = y + 1;
	s->x2 = x + w - 2;
	s->y2 = y + 1;
	s++;
	s->x1 = x;
	s->y1 = y + 1;
	s->x2 = x;
	s->y2 = y + h - 1;
	s++;
	s->x1 = x + 1;
	s->y1 = y + 2;
	s->x2 = x + 1;
	s->y2 = y + h - 2;
	if (state) {
		gc = guiXBorderHighGC;
	} else {
		gc = guiXBorderLowGC;
	}
	XDrawSegments(guiXDisplay, (Window) c->handle, gc,
		segments, 4);	
	s = segments;
	s->x1 = x + w - 1;
	s->y1 = y + 1;
	s->x2 = x + w - 1;
	s->y2 = y + h - 1;
	s++;
	s->x1 = x + w - 2;
	s->y1 = y + 2;
	s->x2 = x + w - 2;
	s->y2 = y + h - 1;
	s++;
	s->x1 = x + 1;
	s->y1 = y + h - 1;
	s->x2 = x + w - 2;
	s->y2 = y + h - 1;
	s++;
	s->x1 = x + 2;
	s->y1 = y + h - 2;
	s->x2 = x + w - 3;
	s->y2 = y + h - 2;
	if (state) {
		gc = guiXBorderLowGC;
	} else {
		gc = guiXBorderHighGC;
	}
	XDrawSegments(guiXDisplay, (Window) c->handle, gc,
		segments, 4);	
}

void guiCanvasAllowVerticalScrollbar(guiCanvasPtr c, int flag)
{
	c->noScrollV = !flag;
	guiXCanvasCheckScrollbars(c);
	guiCanvasInvalidate(c, 0);
}

void guiCanvasAllowHorizontalScrollbar(guiCanvasPtr c, int flag)
{
	c->noScrollH = !flag;
	guiXCanvasCheckScrollbars(c);
	guiCanvasInvalidate(c, 0);
}

void guiCanvasSetClearing(guiCanvasPtr c, int flag)
{
	c->clearing = flag;
}

guiCanvasPtr guiXCanvasGetTopParent(guiCanvasPtr c)
{
	while (c->parent) {
		c = c->parent;
	}
	return c;
}

guiCanvasPtr guiXCanvasGetRootBenefactor(guiCanvasPtr c)
{
	while (c->parent || c->onBehalfOf) {
		if (c->parent) {
			c = c->parent;
		} else if (c->onBehalfOf) {
			c = c->onBehalfOf;
		}
	}
	return c;
}

/* Passes a keyboard event to the right canvas,
	taking grabs into account. */
                                                        
int guiXCanvasResolveKey(guiCanvasPtr *cp, guiEventPtr e)
{
	guiCanvasPtr c = *cp;
	guiCanvasPtr p = guiXCanvasGetRootBenefactor(c);
	guiXKeyGrabPtr g = guiXFirstGrab;
	Window dummy;
	/* Make sure shift flag is correct for alpha keys */
	if ((e->key > 0) && (isalpha(e->key))) {
		if (isupper(e->key)) {
			e->shift = 1;
		} else {
			e->shift = 0;
		}
	}
	/* 1. Look for a grab with the same root benefactor canvas. */
	while (g) {
		/* Unmapped grabs lose */
		if (!guiCanvasIsVisible(g->c)) {
			g = g->next;
			continue;
		}	
		if (guiXCanvasGetRootBenefactor(g->c) == p) {
			if (g->k == e->key) {
				if ((g->shift == e->shift) && 
					(g->control == e->control) &&
					(g->alt == e->alt))
				{
					if (guiCanvasIsVisible(g->c)) {
						XTranslateCoordinates(guiXDisplay, 
							(Window) c->handle,
							(Window) g->c->handle,
							e->l.x,
							e->l.y,
							&e->l.x, 
							&e->l.y, 
							&dummy);
					}
					*cp = g->c;
					if (!(guiXModalFirst && 
						(guiCanvasGetRootParent(g->c)
							!= guiXModalFirst->c))) 
					{
						return g->c->eventHandler(g->c, e);
					}
				}
			}
		}
		g = g->next;
	}
	/* 2. Give other grabs a chance. */
	g = guiXFirstGrab;
	while (g) {
		/* Unmapped grabs lose */
		if (!guiCanvasIsVisible(g->c)) {
			g = g->next;
			continue;
		}	
		if (g->k == e->key) {
			if ((g->shift == e->shift) && 
				(g->control == e->control) &&
				(g->alt == e->alt))
			{
				if (guiCanvasIsVisible(g->c)) {
					XTranslateCoordinates(guiXDisplay, 
						(Window) c->handle,
						(Window) g->c->handle,
						e->l.x,
						e->l.y,
						&e->l.x, 
						&e->l.y, 
						&dummy);
				}
				*cp = g->c;
				if (!(guiXModalFirst &&
					(guiCanvasGetRootParent(g->c)
						!= guiXModalFirst->c))) 
				{
					return g->c->eventHandler(g->c, e);
				}
			}
		}
		g = g->next;
	}				
	*cp = c;
	if (guiXFocusCanvas) {
		if (!(guiXModalFirst &&
			(guiCanvasGetRootParent(guiXFocusCanvas)
				!= guiXModalFirst->c))) 
		{
			if (guiCanvasIsVisible(guiXFocusCanvas)) {
				XTranslateCoordinates(guiXDisplay, 
					(Window) c->handle,
					(Window) guiXFocusCanvas->handle,
					e->l.x,
					e->l.y,
					&e->l.x, 
					&e->l.y, 
					&dummy);
			}
			return guiXFocusCanvas->eventHandler(
				guiXFocusCanvas, e);
		} else {
			return 0;
		}	
	}
	if (!(guiXModalFirst &&
		(guiCanvasGetRootParent(c)
			!= guiXModalFirst->c))) 
	{
		return c->eventHandler(c, e);
	} else {
		return 0;
	}
}

/* Grabs a key. */

void guiCanvasGrabKey(guiCanvasPtr c, guiKey k, 
	int shift, int control, int alt)
{
	guiXKeyGrabPtr g = (guiXKeyGrabPtr) malloc(sizeof(guiXKeyGrab));
	if (!g) {
		return;
	}
	/* Make sure shift flag is correct for alpha keys */
	if ((k > 0) && (isalpha(k))) {
		if (isupper(k)) {
			shift = 1;
		} else {
			shift = 0;
		}
	}	
	g->c = c;	
	g->k = k;
	g->shift = shift;
	g->control = control;
	g->alt = alt;
	if (guiXFirstGrab) {
		g->next = guiXFirstGrab;
	} else {
		g->next = 0;
	}
	guiXFirstGrab = g;
}

/* Releases a keystroke grab. */

void guiCanvasReleaseKey(guiCanvasPtr c, guiKey k, 
	int shift, int control, int alt)
{
	guiXKeyGrabPtr g = guiXFirstGrab, p;
	/* Make sure shift flag is correct for alpha keys */
	if ((k > 0) && (isalpha(k))) {
		if (isupper(k)) {
			shift = 1;
		} else {
			shift = 0;
		}
	}	
	p = 0;
	while (g) {
		if ((c == g->c) && (g->k == k)) 
		{
			if ((g->shift == shift) && 
				(g->control == control) &&
				(g->alt == alt))
			{
				if (p) {
					p->next = g->next;	
				} else {
					guiXFirstGrab = g->next;
				}
				free(g);
				return;
			}
		}			
		p = g;
		g = g->next;
	}
}

void guiCanvasSetDim(guiCanvasPtr c, int flag)
{
	guiColor color;
	c->dim = flag;
	guiCanvasSetDrawColor(c, c->drawColor);
	guiCanvasSetFillColor(c, c->drawColor);
	guiCanvasSetTextBackgroundColor(c, c->textBackgroundColor);
	if (flag) {
		color = guiXColorAverage(c->drawColor, c->backgroundColor);
		XSetForeground(guiXDisplay, (GC) c->drawContext,
			guiXColorGet(&color));
		color = guiXColorAverage(c->textBackgroundColor, 
			c->backgroundColor);
		XSetBackground(guiXDisplay, (GC) c->drawContext,
			guiXColorGet(&color));
		color = guiXColorAverage(c->fillColor, c->backgroundColor);
		XSetForeground(guiXDisplay, (GC) c->fillContext,
			guiXColorGet(&color));
	} else {
		XSetForeground(guiXDisplay, (GC) c->drawContext,
			guiXColorGet(&c->drawColor));
		XSetBackground(guiXDisplay, (GC) c->drawContext,
			guiXColorGet(&c->textBackgroundColor));
		XSetForeground(guiXDisplay, (GC) c->fillContext,
			guiXColorGet(&c->fillColor));
	}
}


guiColor guiXColorAverage(guiColor a, guiColor b)
{
	guiColor c;
	c.r = (a.r + b.r) / 2;
	c.g = (a.g + b.g) / 2;
	c.b = (a.b + b.b) / 2;
	return c;
}

void guiCanvasWantFocus(guiCanvasPtr c, int flag)
{
	c->wantFocus = flag;
}

int guiCanvasQueryFocus(guiCanvasPtr c)
{
	return c->focused;
}

void guiCanvasSetTabbing(guiCanvasPtr c, int flag)
{
	if (c->tabbing != flag) {
		c->tabbing = flag;
		if (c->tabbing) {
		        guiCanvasGrabKey(c, guiKeyTab, 0, 0, 0);
		        guiCanvasGrabKey(c, guiKeyTab, 1, 0, 0);
		} else {
		        guiCanvasReleaseKey(c, guiKeyTab, 0, 0, 0);
		        guiCanvasReleaseKey(c, guiKeyTab, 1, 0, 0);
		}
	}
}

void guiXCanvasFirstFocus(guiCanvasPtr c)
{
	guiCanvasPtr p = c->first;
	while (p) {
		if (p->wantFocus) {
			guiCanvasTakeFocus(p);
			return;
		} else if (p->first) {
			p = p->first;
			continue;
		}
		while (!p->next) {
			p = p->parent;
			if (p == c) {
				return;
			}
		}	
		p = p->next;
	}
}

void guiCanvasScrollToRectangle(guiCanvasPtr c, guiRectanglePtr r)
{
	int x, y;
	int change = 0;
	guiSize s = guiCanvasGetVisibleSize(c);
	x = c->visibleTop.x;
	y = c->visibleTop.y;
	if (r->x < c->visibleTop.x) {
		x = r->x;
		change = 1;
	}
	if (r->w <= s.w) {
		if ((r->x + r->w) > (c->visibleTop.x + s.w))
		{
			x += (r->x + r->w) - (c->visibleTop.x + s.w);
			change = 1;
		}		
	}	
	if (r->y < c->visibleTop.y) {
		y = r->y;
		change = 1;
	}
	if (r->h <= s.h) {
		if ((r->y + r->h) > (c->visibleTop.y + s.h))
		{
			y += (r->y + r->h) - (c->visibleTop.y + s.h);
			change = 1;
		}		
	}	
	if (change) {
		guiCanvasScrollTo(c, x, y);
	}
}

typedef struct {
	int result;
	guiCanvasPtr c;
} messageData, *messageDataPtr;

static void messageActionHandler(char *text, void *clientData);


int guiMessage(guiCanvasPtr parent, char *title, char *text, int mtype)
{
	guiCanvasPtr c;
	guiSize s, ms;
	int bw;
	guiButtonPtr b;
	guiLabelPtr l;
	messageData md;
	int total = 0, which = 0;
	ms = guiLabelEstimate(text);
	if (ms.w > (guiGetScreenSize().w / 3)) {
		ms.w = guiGetScreenSize().w / 3;
		ms.h = guiLabelEstimateHeight(text, ms.w);
	}
	s = ms;
	s.h += guiButtonEstimate("OK", guiButtonSimple).h;	
	s.h += 4;
	bw = 0;
	if ((mtype == guiMessageOK) || (mtype == guiMessageOKCancel)) 
	{
		bw += guiButtonEstimate("OK", guiButtonSimple).w;	
		total++;
	}
	if ((mtype == guiMessageYesNo) || (mtype == guiMessageYesNoCancel)) 
	{
		bw += guiButtonEstimate("Yes", guiButtonSimple).w;	
		bw += guiButtonEstimate("No", guiButtonSimple).w;	
		total += 2;
	}
	if ((mtype == guiMessageCancel) || (mtype == guiMessageYesNoCancel))
	{
		bw += guiButtonEstimate("Cancel", guiButtonSimple).w;	
		total ++;
	}
	if (bw > s.w) {
		s.w = bw;
	}
	c = guiCanvasCreate(0, 0, 0, s.w + 4, s.h + 4);
	guiCanvasSetBorder(c, 2, guiBorderBackground);
	guiCanvasSetType(c, guiCanvasDialog, parent);
	l = guiLabelCreateWithSize(c, 0, 0, ms.w, ms.h, text);
	guiLabelShow(l, 1);
	guiCanvasSetTabbing(c, 1);
	if ((mtype == guiMessageOK) || (mtype == guiMessageOKCancel)) 
	{
		b = guiButtonCreate(c, 
			which * s.w / total + s.w / total / 2 -
			guiButtonEstimate("OK", guiButtonSimple).w / 2, 
			ms.h + 4, 
			"OK", guiButtonSimple);
		guiButtonSetClientData(b, (void *) &md);
		guiButtonSetActionHandler(b, messageActionHandler);
		guiButtonShow(b, 1);	
		which++;
	}	
	if ((mtype == guiMessageYesNo) || (mtype == guiMessageYesNoCancel)) 
	{
		b = guiButtonCreate(c, 
			which * s.w / total + s.w / total / 2 -
			guiButtonEstimate("Yes", guiButtonSimple).w / 2, 
			ms.h + 4, 
			"Yes", guiButtonSimple);
		guiButtonSetClientData(b, (void *) &md);
		guiButtonSetActionHandler(b, messageActionHandler);
		guiButtonShow(b, 1);	
		which++;
		b = guiButtonCreate(c, 
			which * s.w / total + s.w / total / 2 -
			guiButtonEstimate("No", guiButtonSimple).w / 2, 
			ms.h + 4, 
			"No", guiButtonSimple);
		guiButtonSetActionHandler(b, messageActionHandler);
		guiButtonSetClientData(b, (void *) &md);
		guiButtonShow(b, 1);	
		which++;
	}
	if ((mtype == guiMessageOKCancel) || (mtype == guiMessageYesNoCancel))
	{
		b = guiButtonCreate(c, 
			which * s.w / total + s.w / total / 2 -
			guiButtonEstimate("Cancel", guiButtonSimple).w / 2, 
			ms.h + 4, 
			"Cancel", guiButtonSimple);
		guiButtonSetActionHandler(b, messageActionHandler);
		guiButtonSetClientData(b, (void *) &md);
		guiButtonShow(b, 1);	
		which++;
	}
	guiCanvasSetName(c, title);
	md.c = c;
	guiCanvasDoDialog(c);
	guiCanvasDestroy(c);
	return md.result;
}

static void messageActionHandler(char *text, void *clientData)
{
	messageDataPtr mdp = (messageDataPtr) clientData;
	if (!strcmp(text, "Yes")) {
		mdp->result = guiMessageYes;
	} else if (!strcmp(text, "No")) {
		mdp->result = guiMessageNo;
	} else if (!strcmp(text, "OK")) {
		mdp->result = guiMessageOK;
	} else if (!strcmp(text, "Cancel")) {
		mdp->result = guiMessageCancel;
	}
	/* This will cause guiCanvasModal and then guiMessage to return. */
	guiCanvasShow(mdp->c, 0);
}

guiPoint guiGetMouseLocation()
{
	guiPoint p;
	Window dummyw;
	int dummy;
	unsigned int udummy;
	XQueryPointer(guiXDisplay, guiXLeader, &dummyw, &dummyw, 
		&p.x, &p.y, &dummy, &dummy, &udummy);
	return p;
}

void guiCanvasSetName(guiCanvasPtr c, char *name)
{
	XStoreName(guiXDisplay, (Window) c->handle, name);
}

void guiCanvasSetIcon(guiCanvasPtr c, guiBitmapPtr b)
{
	XWMHints xwmh;
	guiCanvasPtr copy = guiCanvasCreateOffscreen(b->s.w, b->s.h);
	guiCanvasDrawBitmap(copy, 0, 0, b);
	xwmh.icon_pixmap = (Pixmap) copy->handle;
	xwmh.flags = IconPixmapHint;
	if (c->icon) {
		guiCanvasDestroy(c->icon);
	}		
	c->icon = copy; 
	XSetWMHints(guiXDisplay, (Window) c->handle, &xwmh);
}

static void canvasDestroyed(guiCanvasPtr c)
{
	int i;
	guiEvent e;
	guiXKeyGrabPtr p;
	guiXKeyGrabPtr g;
	guiXKeyGrabPtr target;
	guiXModalItemPtr m, prev;
	e.t = guiEventDestroyed;
	if (guiXMouseDownCanvas == c) {
		guiXMouseDownCanvas = 0;
	}
	for (i = 0; (i < guiXTimersTotal); i++) {
		if (guiXTimers[i].canvas == c) {
			guiXTimers[i].canvas = 0;
		}
	}
	if (guiXLastCanvas == c) {
		guiXLastCanvas = 0;
	}
	if (guiXFocusCanvas == c) {
		guiXFocusCanvas = 0;
	}
	if (guiXScrollbarButtonCanvas == c) {
		guiXScrollbarButtonCanvas = 0;	
	}
	XFreeGC(guiXDisplay, (GC) c->drawContext);
	XFreeGC(guiXDisplay, (GC) c->fillContext);
	p = 0;
	g = guiXFirstGrab;
	while (g) {
		if (c == g->c)
		{
			if (p) {
				p->next = g->next;	
			} else {
				guiXFirstGrab = g->next;
			}
			target = g;
			g = g->next;
			free(target);
		} else {			
			p = g;
			g = g->next;
		}
	}			
	if (c->parent) {
		guiCanvasPtr p = c->parent->first;
		guiCanvasPtr prev = 0;
		while (p != c) {
			prev = p;	
			p = p->next;
		}
		if (prev) {
			prev->next = c->next;
		} else {
			c->parent->first = c->next;
		}
	}
	if (c->parent) {
		guiCanvasPtr p = c->parent;
		while (p) {
			if (p->childFocused == c) {
				p->childFocused = 0;
				break;
			}
			p = p->parent;
		}
	}
	m = guiXModalFirst;
	prev = 0;
	while (m) {
		if (m->c == c) {
			if (prev) {
				prev->next = m->next;
			} else {
				guiXModalFirst = m->next;						}
			free(m);
			guiXModalDepth--;
		}
		prev = m;
		m = m->next;		
	}
	if (c->icon) {                                         
		guiCanvasDestroy(c->icon);                     
	}                 
	c->eventHandler(c, &e);
	free(c);
}

guiBitmapPtr guiBitmapRaw(unsigned char *resource, int w, int h, int dither)
{
	int x, y;
	guiBitmapPtr b = guiXBitmapCreate(w, h); 
	if (!b) {
		return 0;
	}
	for (y = 0; (y < b->s.h); y++) {
		guiColor c;
		c.r = 0;
		c.g = 0;
		c.b = 0;
		if (dither) {
			for (x = 0; (x < b->s.w); x++) {
				guiColor try;
				c.r += *resource++;
				c.g += *resource++;
				c.b += *resource++;
				try = c;
				XPutPixel((XImage *) b->handle,
					x, y, guiXColorGet(&try));
				c.r -= try.r;
				c.g -= try.g;
				c.b -= try.b;
			}
		} else {
			for (x = 0; (x < b->s.w); x++) {
				guiColor try;
				try.r = *resource++;
				try.g = *resource++;
				try.b = *resource++;
				XPutPixel((XImage *) b->handle,
					x, y, guiXColorGet(&try));
			}	
		}
	}
	return b;
}

guiBitmapPtr guiBitmapCopy(guiBitmapPtr src)
{
	guiBitmapPtr dst = guiXBitmapCreate(src->s.w, src->s.h);
	int x, y;
	if (!dst) {
		return 0;
	}
	for (y = 0; (y < src->s.h); y++) {
		for (x = 0; (x < src->s.w); x++) {
			XPutPixel((XImage *) dst->handle,
				x, y, XGetPixel((XImage *) src->handle, x, y));
		}

	}	
	return dst;
}


