#include "guib.h"

guiFont guiButtonFont = 
{
        guiTimesRoman,
        12,
        0,
        1,
        0,
        0
};

guiColor guiButtonColor =
{
	0, 0, 0
};

guiColor guiButtonCheckColor =
{
	0, 255, 255 
};

static void guiButtonDestroyed(guiButtonPtr b);

static void guiButtonRepaint(guiButtonPtr b);

static int eventHandler(guiCanvasPtr c, guiEventPtr e);

static void guiXButtonDown(guiButtonPtr b);

static void guiXButtonUp(guiButtonPtr b);

static guiButtonPtr createBody(guiCanvasPtr parent,
	int x, int y, char *label, guiBitmapPtr bitmap, guiButtonStyle style);

guiButtonPtr guiButtonCreate(guiCanvasPtr parent,
	int x, int y, char *label, guiButtonStyle style)
{
	return createBody(parent, x, y, label, 0, style);
}

guiButtonPtr guiButtonCreateBitmap(guiCanvasPtr parent,
	int x, int y, guiBitmapPtr bitmap, guiButtonStyle style)
{
	return createBody(parent, x, y, 0, bitmap, style);
}

guiButtonPtr createBody(guiCanvasPtr parent,
	int x, int y, char *label, guiBitmapPtr bitmap, guiButtonStyle style)
{
	guiButtonPtr b = (guiButtonPtr) malloc(sizeof(guiButton));
	guiSize s;
	guiFontFindClosest(&guiButtonFont);

	b->style = style;
	b->active = 1;
	b->state = 0;
	b->down = 0;
	b->actionHandler = 0;
	b->clientData = 0;
	if (bitmap) {
		b->bitmap = guiBitmapCopy(bitmap);
		if (!b->bitmap) {
			free(b);
			return 0;
		}
		s = guiButtonEstimateBitmap(bitmap, style);
	} else {
		b->bitmap = 0;
	}	
	if (label) {
		/* Copy the label */
		b->label = (char *) malloc(strlen(label) + 1);
		strcpy(b->label, label);
		if (!b->label) {
			free(b);
			return 0;
		}
		s = guiButtonEstimate(label, style);
	} else {
		b->label = 0;
	}
	b->frame = (guiNativeHandle) guiCanvasCreate(parent, x, y, s.w, s.h);
	guiCanvasSetBorder((guiCanvasPtr) b->frame, 2, guiBorderBackground);
	b->c = (guiNativeHandle) guiCanvasCreate(
		(guiCanvasPtr) b->frame, 0, 0, s.w - 4, s.h - 4);
	guiCanvasSetBorder((guiCanvasPtr) b->c, 2, guiBorderOut);
	guiCanvasSetClientData((guiCanvasPtr) b->c, (void *) b);
	guiCanvasSetEventHandler((guiCanvasPtr) b->c, eventHandler);
	guiCanvasSetEventMask((guiCanvasPtr) b->c,
		guiEventRepaint | guiEventMouseDown | guiEventMouseUp |
		guiEventKeyDown |
		guiEventFocusIn |
		guiEventFocusOut);
	guiCanvasWantFocus((guiCanvasPtr) b->c, 1);
	guiCanvasShow((guiCanvasPtr) b->c, 1);
	b->leader = b;
	b->winner = 0;
	return b;
}
        
guiSize guiButtonEstimate(char *label, guiButtonStyle style)
{                                          
	guiSize s, sm;
	guiFontFindClosest(&guiButtonFont);
	s = guiFontTextSize(&guiButtonFont, label, strlen(label));
	if (style != guiButtonSimple) {
		sm = guiFontTextSize(&guiButtonFont, " M", 2);
		s.w += sm.w;
	}
	s.h = guiButtonFont.ascent + guiButtonFont.descent;
	s.w += 8;
	s.h += 8;
	return s;
}

guiSize guiButtonEstimateBitmap(guiBitmapPtr bitmap, guiButtonStyle style)
{                                          
	guiSize s;
	guiFontFindClosest(&guiButtonFont);
	s = bitmap->s;
	s.w += 8;
	s.h += 8;
	return s;
}

void guiButtonDestroy(guiButtonPtr b)
{
	guiCanvasDestroy((guiCanvasPtr) b->c);
}

static void guiButtonDestroyed(guiButtonPtr b)
{
	if (b->label) {
		free(b->label);
	}
	if (b->bitmap) {
		guiBitmapDestroy(b->bitmap);
	}	
	free(b);
}

static int eventHandler(guiCanvasPtr c, guiEventPtr e)
{
	guiButtonPtr b = (guiButtonPtr) guiCanvasGetClientData(c);	
	switch (e->t) {
		case guiEventDestroyed:
		guiButtonDestroyed(b);
		break;
		case guiEventRepaint:
		guiButtonRepaint(b);
		break;
		case guiEventKeyDown:
		if ((e->key == guiKeySpace) || (e->key == guiKeyEnter)) {
			guiXButtonDown(b);
			guiCanvasSetTimer((guiCanvasPtr) b->c, 1, 50);
		}
		break;
		case guiEventTimer:
		guiXButtonUp(b);
		break;
		case guiEventMouseDown:
		guiXButtonDown(b);
		break;
		case guiEventMouseUp:
		guiXButtonUp(b);
		break;	
		case guiEventFocusIn:
		guiCanvasSetBorder((guiCanvasPtr) b->frame, 2, 
			guiBorderIn);
		break;
		case guiEventFocusOut:
		guiCanvasSetBorder((guiCanvasPtr) b->frame, 2, 
			guiBorderBackground);
		break;
		default:
		break;
	}
	return 0;		
}

void guiButtonRepaint(guiButtonPtr b)
{
	guiSize s, ss, sm;
	guiCanvasPtr c = (guiCanvasPtr) b->c;
	if (!b->active) {
		guiCanvasSetDim(c, 1);
	}

	if (b->bitmap) {
		guiCanvasDrawBitmap(c, 0, 0, b->bitmap);
	} else {
		guiCanvasSetFont(c, &guiButtonFont);
		guiCanvasSetDrawColor(c, guiButtonColor);
		guiCanvasSetFillColor(c, guiButtonCheckColor);
		guiCanvasDrawText(c, 0, guiButtonFont.ascent, 
			b->label, strlen(b->label));
		s = guiFontTextSize(&guiButtonFont, b->label, strlen(b->label));	
		ss = guiFontTextSize(&guiButtonFont, " ", 1);
		sm = guiFontTextSize(&guiButtonFont, "M", 1);
		switch(b->style) {	
			case guiButtonSimple:
			break;
			case guiButtonCheck:
			case guiButtonRadio:
			{
				int x, y, w, h;
				x = s.w + ss.w;
				w = sm.w - 3;
				h = sm.h - 1;
				y = guiButtonFont.ascent - h;
				guiCanvasDrawRectangle(c, x - 1, y - 1,
					sm.w - 2, sm.h);
				guiCanvasFillRectangle(c, x, y, w, h);
				if (b->state) {	
					guiCanvasDrawLine(c, x, y, 
						x + w - 1, y + h - 1);
					guiCanvasDrawLine(c, x + w - 1, y,
						x, y + h - 1);	
				}
			}
			break;
		}	
	}
	if (!b->active) {
		guiCanvasSetDim(c, 0);
	}
}

void guiButtonSetLeader(guiButtonPtr b, guiButtonPtr l)
{
	b->leader = l;
}

void guiButtonSetActionHandler(guiButtonPtr b, 
	void (*actionHandler) (char *text, void *clientData))
{
	b->actionHandler = actionHandler;
}

void guiButtonSetClientData(guiButtonPtr b, void *clientData)
{
	b->clientData = clientData;
}

void *guiButtonGetClientData(guiButtonPtr b)
{
	return b->clientData;
}

void guiButtonSetActive(guiButtonPtr b, int flag)
{
	b->active = flag;
	guiCanvasWantFocus((guiCanvasPtr) b->c, flag);
	guiCanvasInvalidate((guiCanvasPtr) b->c, 0);
}

int guiButtonGetActive(guiButtonPtr b)
{
	return b->active;
}

void guiButtonSetState(guiButtonPtr b, int flag)
{
	if (b->bitmap) {
		b->down = flag;
	}
	if (flag) {
		if (b->style == guiButtonRadio) {
			if (b->leader && b->leader->winner) {
				guiButtonSetState(b->leader->winner, 0);
			}
		}
		if (b->leader) {
			b->leader->winner = b;
		}
		if (b->bitmap) {
			guiCanvasSetBorder((guiCanvasPtr) b->c, 
				2, guiBorderIn);
			guiCanvasInvalidate((guiCanvasPtr) b->c, 0);
		}
	} else {
		if (b->leader && (b->leader->winner == b)) {
			b->leader->winner = 0;
		}
		if (b->bitmap) {
			guiCanvasSetBorder((guiCanvasPtr) b->c, 
				2, guiBorderOut);
			guiCanvasInvalidate((guiCanvasPtr) b->c, 0);
		}
	}	
	b->state = flag;
	if (b->label) {
		guiCanvasInvalidate((guiCanvasPtr) b->c, 0);
	}
}
                                                  
int guiButtonGetState(guiButtonPtr b)
{
	return b->state;
}

void guiButtonMove(guiButtonPtr b, int x, int y)
{
	guiCanvasMove((guiCanvasPtr) b->frame, x, y);
}

guiSize guiButtonGetPhysicalSize(guiButtonPtr b)
{
	return ((guiCanvasPtr) b->frame)->physicalSize;
}

guiPoint guiButtonGetPosition(guiButtonPtr b)
{
	return ((guiCanvasPtr) b->frame)->l;
}

void guiButtonShow(guiButtonPtr b, int flag)
{
	guiCanvasShow((guiCanvasPtr) b->frame, flag);
}
                
static void guiXButtonDown(guiButtonPtr b)
{   
	guiCanvasPtr c = (guiCanvasPtr) b->c;
	if (!b->active) {
		return;
	}
	b->down = 1;
	guiCanvasSetBorder(c, 2, guiBorderIn);
	switch (b->style) {
		case guiButtonSimple:
		if (b->label) {
			guiCanvasInvalidate(c, 0);
		}
		break;
		case guiButtonRadio:
		/* Calls invalidate */
		guiButtonSetState(b, 1);
		break;
		case guiButtonCheck:
		/* Calls invalidate */
		guiButtonSetState(b, !b->state);
		break;
	}
	if (b->actionHandler) {
		b->actionHandler(b->label, b->clientData);
	}
}

static void guiXButtonUp(guiButtonPtr b)
{
	guiCanvasPtr c = (guiCanvasPtr) b->c;
	if (!b->active) {
		return;
	}
	if (b->label || (b->style == guiButtonSimple)) {
		b->down = 0;
		guiCanvasSetBorder(c, 2, guiBorderOut);
		guiCanvasInvalidate(c, 0);
	}
}


