#include "guic.h"
#include "guila.h"
#include "guib.h"

typedef struct {
	guiLabelPtr label;
	guiButtonPtr colors[25][5];
	guiButtonPtr ok;
	guiButtonPtr cancel;
	guiButtonPtr help;
	guiCanvasPtr canvas;
	guiCanvasPtr frame;
	guiColorSelectOptionsPtr options;
	guiColorSelectResult result;
} colorSelectData, *colorSelectDataPtr;

static int eventHandler(guiCanvasPtr c, guiEventPtr e);
static void okHandler(char *text, void *clientData);
static void cancelHandler(char *text, void *clientData);
static void helpHandler(char *text, void *clientData);

#define cellSize 10
#define cellPixels (cellSize * cellSize)
#define cellBytes (cellPixels * 3)

guiColorSelectResult guiColorSelect(guiColorSelectOptionsPtr options)
{
	guiSize s, ls, bs, ts;
	int bx, by;
	int x, y;
	int top;
	colorSelectData fd;
	guiBitmapPtr cell;
	unsigned char *data;
	data = (unsigned char *) calloc(cellBytes, 1);
	cell = guiBitmapRaw(data, cellSize, cellSize, 0);
	bs = guiButtonEstimateBitmap(cell, guiButtonRadio);	
	guiBitmapDestroy(cell);
	s.w = bs.w * 25;
	s.h = bs.h * 5;
	if (options->label) {
		ls.h = guiLabelEstimateHeight(options->label, s.w);
	}
	s.h += ls.h + 4;
	ts = guiButtonEstimate("OK", guiButtonSimple);
	s.h += ts.h + 4;
	s.w += 4;
	fd.label = 0;
	fd.frame = guiCanvasCreate(0, 0, 0, s.w, s.h);
	guiCanvasSetBorder(fd.frame, 2, guiBorderBackground);
	s = guiCanvasGetVisibleSize(fd.frame);
	if (options->label) {
		int height;
		height = guiLabelEstimateHeight(options->label, s.w);
		fd.label = guiLabelCreateWithSize(fd.frame, 0, 0, s.w, height, 
			options->label);
		guiLabelShow(fd.label, 1);
		top = height + 4;
	} else {
		top = 0;
	}
	by = s.h - guiButtonEstimate("OK", guiButtonSimple).h;
	bx = 0;
	fd.ok = guiButtonCreate(fd.frame, 0, by, "OK", guiButtonSimple);
	guiButtonSetClientData(fd.ok, &fd);
	guiButtonSetActionHandler(fd.ok, okHandler);
	guiButtonShow(fd.ok, 1);
	bx += guiButtonGetPhysicalSize(fd.ok).w;
	bx += 16;
	fd.cancel = guiButtonCreate(fd.frame, 
		bx, by, "Cancel", guiButtonSimple);
	bx += guiButtonGetPhysicalSize(fd.cancel).w;
	bx += 16;
	guiButtonSetClientData(fd.cancel, &fd);
	guiButtonSetActionHandler(fd.cancel, cancelHandler);
	guiButtonShow(fd.cancel, 1);
	if (options->helpHandler) {
		fd.help = guiButtonCreate(fd.frame, bx, by, "HELP!", 
			guiButtonSimple);
		bx += guiButtonGetPhysicalSize(fd.help).w;
		bx += 16;
		guiButtonSetClientData(fd.help, &fd);
		guiButtonSetActionHandler(fd.help, helpHandler);
		guiButtonShow(fd.help, 1);
	}
	fd.options = options;
	for (y = 0; (y < 5); y++) {
		for (x = 0; (x < 25); x++) {
			int r, g, b;
			int p;
			unsigned char *dp = data;
			r = x % 5;
			g = x / 5;
			b = y;
			for (p = 0; (p < cellPixels); p++) {	
				*dp++ = guiMin(r * 64, 255);
				*dp++ = guiMin(g * 64, 255);
				*dp++ = guiMin(b * 64, 255);
			}
			cell = guiBitmapRaw(data, cellSize, cellSize, 0);
			fd.colors[x][y] = guiButtonCreateBitmap(
				fd.frame,
				x * bs.w,
				y * bs.h + top,	
				cell,
				guiButtonRadio);
			guiButtonShow(fd.colors[x][y], 1);
			guiBitmapDestroy(cell);
			guiButtonSetLeader(fd.colors[x][y], fd.colors[0][0]);
		}
	}	
	x = (guiBound(options->current.r, 0, 255) + 1) / 64;
	x += ((guiBound(options->current.g, 0, 255) + 1) / 64) * 5;
	y = (guiBound(options->current.b, 0, 255) + 1) / 64;
	guiButtonSetState(fd.colors[x][y], 1);	
	guiCanvasSetTabbing(fd.frame, 1);
	guiCanvasSetType(fd.frame, guiCanvasDialog, options->from);
	guiCanvasSetEventMask(fd.frame, guiEventKeyDown);
	guiCanvasSetEventHandler(fd.frame, eventHandler);
	guiCanvasSetClientData(fd.frame, (void *) &fd);
	guiCanvasGrabKey(fd.frame, guiKeyEnter, 0, 0, 0);
	guiCanvasGrabKey(fd.frame, guiKeyEscape, 0, 0, 0);
	if (options->title) {
		guiCanvasSetName(fd.frame, options->title);
	}
	guiCanvasDoDialog(fd.frame);
	guiCanvasDestroy(fd.frame);
	return fd.result;
}

static int eventHandler(guiCanvasPtr c, guiEventPtr e)
{
	colorSelectDataPtr fd = (colorSelectDataPtr) guiCanvasGetClientData(c);
	switch (e->t) {
		case guiEventKeyDown:
		if (e->key == guiKeyEscape) {
			cancelHandler("", (void *) fd);
		} else if (e->key == guiKeyEnter) {
			okHandler("", (void *) fd);
		}
		break;		
		default:
		break;
	}
	return 0;
}

static void okHandler(char *text, void *clientData)
{
	int x, y;
	colorSelectDataPtr fd = (colorSelectDataPtr) clientData;
	for (y = 0; (y < 5); y++) {
		for (x = 0; (x < 25); x++) {
			int r, g, b;
			if (guiButtonGetState(fd->colors[x][y])) {
				r = x % 5;
				g = x / 5;
				b = y;	
				fd->options->chosen.r = guiMin(r * 64, 255);
				fd->options->chosen.g = guiMin(g * 64, 255);
				fd->options->chosen.b = guiMin(b * 64, 255);
			}
		}
	}
	fd->result = guiColorSelectOK;
	guiCanvasShow(fd->frame, 0);
}

static void cancelHandler(char *text, void *clientData)
{
	colorSelectDataPtr fd = (colorSelectDataPtr) clientData;
	fd->result = guiColorSelectCanceled;
	guiCanvasShow(fd->frame, 0);
}

static void helpHandler(char *text, void *clientData)
{
	colorSelectDataPtr fd = (colorSelectDataPtr) clientData;
	fd->options->helpHandler("HELP!", fd->options->helpClientData);
}

