#include "guif.h"
#include "guila.h"
#include "guil.h"
#include "guib.h"
#include "guit.h"
#include "guimatch.h"

#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>

enum {
	recentDirectories,
	recentFiles,
	recentFilter
};

typedef struct {
	guiLabelPtr label;
	guiLabelPtr flabel;
	guiLabelPtr plabel;
	guiLabelPtr dirslabel;
	guiLabelPtr fileslabel; 	
	guiButtonPtr ok;
	guiButtonPtr cancel;
	guiButtonPtr filterb;
	guiButtonPtr help;
	guiTextPtr filename;
	guiTextPtr filter;
	guiListPtr dlist;
	guiListPtr flist;
	char *activeFilter;
	guiCanvasPtr frame;
	guiFileOptionsPtr options;
	guiFileResult result;
	int recent;
} fileData, *fileDataPtr;

static int eventHandler(guiCanvasPtr c, guiEventPtr e);
static void okHandler(char *text, void *clientData);
static void filterHandler(char *text, void *clientData);
static void cancelHandler(char *text, void *clientData);
static void helpHandler(char *text, void *clientData);
static void dlistHandler(char *text, void *itemData, 
	void *clientData, int final);
static void flistHandler(char *text, void *itemData, 
	void *clientData, int final);
static void fileHandler(guiTextPtr t, void *clientData);
static void filterUpdateHandler(guiTextPtr t, void *clientData);

static void updateDir(fileDataPtr fd);
static char *getJustDir(char *filename);
static char *getJustFile(char *filename);

guiFileResult guiFileSelect(guiFileOptionsPtr options)
{
	guiSize s, ls;
	int bx, by;
	int h;
	int listTop;
	int listBottom;
	int i;
	char str[10];
	fileData fd;
	s = guiGetScreenSize();
	s.w /= 3;
	s.h /= 3;
	if (options->filter) {
		fd.activeFilter = (char *) malloc(strlen(options->filter) + 1);
		strcpy(fd.activeFilter, options->filter);
	} else {
		fd.activeFilter = (char *) malloc(2);
		strcpy(fd.activeFilter, "*");	
	}
	fd.label = 0;
	fd.frame = guiCanvasCreate(0, 0, 0, s.w, s.h);
	guiCanvasSetBorder(fd.frame, 2, guiBorderBackground);
	s = guiCanvasGetVisibleSize(fd.frame);
	if (options->label) {
		int height;
		height = guiLabelEstimateHeight(options->label, s.w);
		fd.label = guiLabelCreateWithSize(fd.frame, 0, 0, s.w, height, 
			options->label);
		guiLabelShow(fd.label, 1);
		listTop = height + 4;
	} else {
		listTop = 0;
	}
	by = s.h - guiButtonEstimate("OK", guiButtonSimple).h;
	bx = 0;
	fd.ok = guiButtonCreate(fd.frame, 0, by, "OK", guiButtonSimple);
	guiButtonSetClientData(fd.ok, &fd);
	guiButtonSetActionHandler(fd.ok, okHandler);
	guiButtonShow(fd.ok, 1);
	bx += guiButtonGetPhysicalSize(fd.ok).w;
	bx += 16;
	fd.filterb = 
		guiButtonCreate(fd.frame, bx, by, "Filter", guiButtonSimple);
	guiButtonSetClientData(fd.filterb, &fd);
	guiButtonSetActionHandler(fd.filterb, filterHandler);
	guiButtonShow(fd.filterb, 1);
	bx += guiButtonGetPhysicalSize(fd.filterb).w;
	bx += 16;
	fd.cancel = guiButtonCreate(fd.frame, bx, by, "Cancel", guiButtonSimple);
	guiButtonSetClientData(fd.cancel, &fd);
	guiButtonSetActionHandler(fd.cancel, cancelHandler);
	guiButtonShow(fd.cancel, 1);
	if (options->helpHandler) {
		bx += guiButtonGetPhysicalSize(fd.help).w;
		bx += 16;
		fd.help = guiButtonCreate(fd.frame, bx, by, "HELP!", 
			guiButtonSimple);
		guiButtonSetClientData(fd.help, &fd);
		guiButtonSetActionHandler(fd.help, helpHandler);
		guiButtonShow(fd.help, 1);
	}
	ls = guiLabelEstimate("Filename:");
	h = guiTextEstimateSingle(1).h;
	if (h > ls.h) {
		ls.h = h;
	}
	fd.flabel = guiLabelCreateWithSize(fd.frame, 
			0, listTop, ls.w, ls.h, 
			"Filename:");
	fd.filename = guiTextCreateSingle(fd.frame,
		ls.w + 16, listTop, s.w - (ls.w + 16), ls.h, guiNone);
	guiTextSetUpdateHandler(fd.filename, fileHandler, (void *) &fd);
	guiLabelShow(fd.flabel, 1);
	guiTextShow(fd.filename, 1);
	listTop += h;
	listTop += 4;
	ls = guiLabelEstimate("Filter:");
	h = guiTextEstimateSingle(1).h;
	if (h > ls.h) {
		ls.h = h;
	}
	listBottom = by - ls.h - 8;
	fd.plabel = guiLabelCreateWithSize(fd.frame, 
			0, listBottom + 4, ls.w, ls.h, 
			"Filter:");
	fd.filter = guiTextCreateSingle(fd.frame,
		ls.w + 16, listBottom + 4, s.w - (ls.w + 16), ls.h, guiNone);
	guiTextSetUpdateHandler(fd.filter, filterUpdateHandler, (void *) &fd);
	guiTextSet(fd.filter, fd.activeFilter, strlen(fd.activeFilter));
	guiLabelShow(fd.plabel, 1);
	guiTextShow(fd.filter, 1);
	fd.dlist = guiListCreate(fd.frame, 
		0, listTop + 8 + ls.h, 
		s.w / 2, (listBottom - listTop) - 8 - ls.h);
	guiListSetSorted(fd.dlist, 1);
	for (i = 0; (i < 100); i++) 
	{
		sprintf(str, "%d", i);
		guiListAddItem(fd.dlist, str, 0);
	}
	guiListSetClientData(fd.dlist, (void *) &fd);
	guiListSetActionHandler(fd.dlist, dlistHandler);
	guiListShow(fd.dlist, 1);
	ls = guiLabelEstimate("Directories");
	fd.dirslabel = guiLabelCreateWithSize(fd.frame, 
		0, listTop + 4, ls.w, ls.h,
		"Directories"); 
	guiLabelShow(fd.dirslabel, 1);	
	ls = guiLabelEstimate("Files");
	fd.fileslabel = guiLabelCreateWithSize(fd.frame, 
		s.w / 2, listTop + 4, ls.w, ls.h,
		"Files"); 
	guiLabelShow(fd.fileslabel, 1);
	fd.flist = guiListCreate(fd.frame, 
		s.w / 2, listTop + 8 + ls.h, 
		s.w / 2, (listBottom - listTop) - 8 - ls.h);
	guiListSetSorted(fd.flist, 1);
	for (i = 0; (i < 100); i++) 
	{
		sprintf(str, "%d", i);
		guiListAddItem(fd.flist, str, 0);
	}
	guiListSetClientData(fd.flist, (void *) &fd);
	guiListSetActionHandler(fd.flist, flistHandler);
	guiListShow(fd.flist, 1);	
	fd.options = options;
	guiCanvasSetTabbing(fd.frame, 1);
	guiCanvasSetType(fd.frame, guiCanvasDialog, options->from);
	guiCanvasSetEventMask(fd.frame, guiEventKeyDown);
	guiCanvasSetEventHandler(fd.frame, eventHandler);
	guiCanvasSetClientData(fd.frame, (void *) &fd);
	guiCanvasGrabKey(fd.frame, guiKeyEnter, 0, 0, 0);
	guiCanvasGrabKey(fd.frame, guiKeyEscape, 0, 0, 0);
	if (options->initialFilename) {
		char *justDir;
		char *justFile;
		justDir = getJustDir(options->initialFilename);
		justFile = getJustFile(options->initialFilename);
		guiTextSet(fd.filename, 
			justFile,
			strlen(justFile));
		if (strcmp(justFile, "")) {
			guiListFindItemByText(fd.flist, justFile);
		}
		free(justFile);
		chdir(justDir);
		updateDir(&fd);
		free(justDir);
	} else {
		updateDir(&fd);
	}
	fd.recent = recentFiles;
	if (options->title) {
		guiCanvasSetName(fd.frame, options->title);
	}
	guiCanvasDoDialog(fd.frame);
	guiCanvasDestroy(fd.frame);
	free(fd.activeFilter);
	return fd.result;
}

static int eventHandler(guiCanvasPtr c, guiEventPtr e)
{
	fileDataPtr fd = (fileDataPtr) guiCanvasGetClientData(c);
	switch (e->t) {
		case guiEventKeyDown:
		if (e->key == guiKeyEscape) {
			cancelHandler("", (void *) fd);
		} else if (e->key == guiKeyEnter) {
			if (fd->recent == recentFiles) {
				okHandler("", (void *) fd);
			} else if (fd->recent == recentDirectories) {
				int item = guiListGetSelectedItem(fd->dlist);
				if (item != guiNone) {
					char *text = guiListGetItemText(
						fd->dlist, item);
					dlistHandler(text, 0, (void *) fd, 1);
					free(text);
				}
			} else {
				filterHandler("", (void *) fd);
			}
		}
		break;		
		default:
		break;
	}
	return 0;
}

static void okHandler(char *text, void *clientData)
{
	fileDataPtr fd = (fileDataPtr) clientData;
	int len;
	int clast;
	char *filename;
	char fullname[2048];
	struct stat stat_buf;
	fd->result = guiFileExisting;
	len = guiTextGetLength(fd->filename);
	if (!len) {
		guiMessage(fd->frame,
			"File Selection Incomplete",
			"Please select a file before clicking "
			"OK or pressing ENTER.",
			guiMessageOK);
		return;
	}
	filename = (char *) malloc(len + 1);
	guiTextGetBuffer(fd->filename, filename, len + 1, 0);
	if (stat(filename, &stat_buf) != 0) {
		if (fd->options->mustExist) {
			guiMessage(fd->frame,
				"No Such File",
				"That file does not exist or cannot be "
				"accessed by the current user.",
				guiMessageOK);
			free(filename);
			return;
		} else {
			fd->result = guiFileNew;
		}
	} else {
		if (S_ISDIR(stat_buf.st_mode)) {
			chdir(filename);
			guiTextSet(fd->filename, "", 0);
			updateDir(fd);
			return;
		}
	}
	getcwd(fullname, 1024);
	clast = strlen(fullname);
	if (clast) {
		if (fullname[clast-1] != '/') {
			strcat(fullname, "/");
		}
	}
	strcat(fullname, filename);
	free(filename);
	filename = (char *) malloc(strlen(fullname) + 1);
	strcpy(filename, fullname);	
	fd->options->outputFilename = filename;
	guiCanvasShow(fd->frame, 0);
}

static void filterHandler(char *text, void *clientData)
{
	fileDataPtr fd = (fileDataPtr) clientData;
	int len = guiTextGetLength(fd->filter);
	if (!len) {
		strcpy(fd->activeFilter, "*");
	} else {
		free(fd->activeFilter);
		fd->activeFilter = (char *) malloc(len + 1);
		guiTextGetBuffer(fd->filter, fd->activeFilter, len + 1, 0);
	}	
	updateDir(fd);
}

static void cancelHandler(char *text, void *clientData)
{
	fileDataPtr fd = (fileDataPtr) clientData;
	fd->result = guiFileCanceled;
	guiCanvasShow(fd->frame, 0);
}

static void helpHandler(char *text, void *clientData)
{
	fileDataPtr fd = (fileDataPtr) clientData;
	fd->options->helpHandler("HELP!", fd->options->helpClientData);
}

static void dlistHandler(char *text, void *itemData, 
	void *clientData, int final)
{
	fileDataPtr fd = (fileDataPtr) clientData;
	fd->recent = recentDirectories;
	if (final) {
		guiTextSet(fd->filename, "", 0);
		chdir(text);
		updateDir(fd);
	}
}

static void flistHandler(char *text, void *itemData, 
	void *clientData, int final)
{
	fileDataPtr fd = (fileDataPtr) clientData;
	guiTextSet(fd->filename, text, strlen(text));
	fd->recent = recentFiles;
	if (final) {
		okHandler("", clientData);		 	
	} 
}

static void fileHandler(guiTextPtr t, void *clientData)
{
	fileDataPtr fd = (fileDataPtr) clientData;
	fd->recent = recentFiles;
}

static void filterUpdateHandler(guiTextPtr t, void *clientData)
{
	fileDataPtr fd = (fileDataPtr) clientData;
	fd->recent = recentFilter;
}

static void updateDir(fileDataPtr fd)
{
	struct dirent *entry;
	char *path, *file;
	DIR *dir = opendir(".");
	if (!dir) {
		/* What should I do about this? */
		return;
	}
	guiListClear(fd->flist);
	guiListClear(fd->dlist);
	guiListShow(fd->flist, 0);
	guiListShow(fd->dlist, 0);
	while ((entry = readdir(dir))) 
	{
		struct stat stat_buf;
		stat(entry->d_name, &stat_buf);
		if (S_ISDIR(stat_buf.st_mode)) {
			guiListAddItem(fd->dlist, entry->d_name, 0);
		} else {		
			if (guiMatch(entry->d_name, fd->activeFilter)) {
				guiListAddItem(fd->flist, entry->d_name, 0);
			}
		}
	}
	closedir(dir);
	guiListShow(fd->flist, 1);
	guiListShow(fd->dlist, 1);
	path = guiTextGet(fd->filename);
	file = getJustFile(path);
	guiTextSet(fd->filename, file, strlen(file));
	free(file);
	free(path);	
}

static char *getJustDir(char *filename)
{
	struct stat stat_buf;
	char *dirname;
	char *lastslash;
	dirname = (char *) malloc(strlen(filename) + 1);
	if (stat(filename, &stat_buf) != 0) {
		if (S_ISDIR(stat_buf.st_mode)) {	
			strcpy(dirname, filename);
			return dirname;
		}
	}
	/* It's not a directory, so back off to the last slash */
	lastslash = strrchr(filename, '/');
	if (lastslash) {
		strncpy(dirname, filename, lastslash - filename);
		dirname[lastslash - filename] = '\0';
		return dirname;
	}		
	/* Probably just a local filename */
	strcpy(dirname, ".");
	return dirname;
}

static char *getJustFile(char *filename)
{
	struct stat stat_buf;
	char *fname;
	char *lastslash;
	fname = (char *) malloc(strlen(filename) + 1);
	if (stat(filename, &stat_buf) != 0) {
		if (S_ISDIR(stat_buf.st_mode)) {	
			strcpy(fname, "");
			return fname;
		}
	}
	/* It's not a directory, so take everything after the last slash */
	lastslash = strrchr(filename, '/');
	if (lastslash) {
		strcpy(fname, lastslash + 1);
		return fname;
	}		
	/* Probably just a local filename */
	strcpy(fname, filename);
	return fname;
}

