#include "guil.h"

static int eventHandler(guiCanvasPtr c, guiEventPtr e);
static void drawRow(guiListPtr l, int i);
static void updateVirtualSize(guiListPtr l);
static void guiListDestroyed(guiListPtr l);
static void extend(guiListPtr l);
static void listAct(guiListPtr l, int final);
static void updateVirtualSizeForRow(guiListPtr l, int i, int *wp, int *hp);

guiFont guiListFont = {
        guiTimesRoman,
        12,
        0,
        0,
        0,
        0
};

guiColor guiListColor = {
        0, 0, 0
};

guiSize guiListEstimate(int rows, int cols)
{
	guiSize s;
	guiFontFindClosest(&guiListFont);
	s.w = guiFontTextSize(&guiListFont, "M", 1).w * cols + 4;
	s.h = (guiListFont.ascent + guiListFont.descent) * rows + 4;
	return s;
}

guiListPtr guiListCreate(guiCanvasPtr parent, int x, int y, int w, int h)
{
	guiListPtr l;
	guiFontFindClosest(&guiListFont);
	l = (guiListPtr) malloc(sizeof(guiList));
	if (!l) {
		return 0;
	}
	l->c = (guiNativeHandle) guiCanvasCreate(parent, x, y, w, h);
	l->text = (char **) malloc(sizeof(char *) * 32);
	if (!l->text) {
		guiCanvasDestroy((guiCanvasPtr) l->c);
		return 0;
	}
	l->itemData = (void **) malloc(sizeof(void *) * 32);
	if (!l->itemData) {
		guiCanvasDestroy((guiCanvasPtr) l->c);
		return 0;
	}
	l->rowsAllocated = 32;
	l->rowsTotal = 0;
	l->selected = guiNone;
	l->clickWaiting = 0;
	l->actionHandler = 0;		
	l->clientData = 0;
	l->sorted = 0;
	guiCanvasSetBorder((guiCanvasPtr) l->c, 2, guiBorderBackground);
	/* Nothing to it yet */
	guiCanvasSetVirtualSize((guiCanvasPtr) l->c, 0, 0);
	guiCanvasSetEventMask((guiCanvasPtr) l->c,
		guiEventRepaint | guiEventMouseDown | guiEventKeyDown |
		guiEventFocusIn | guiEventFocusOut);
	guiCanvasSetEventHandler((guiCanvasPtr) l->c, eventHandler);
	guiCanvasSetClientData((guiCanvasPtr) l->c, (void *) l);
	guiCanvasSetFont((guiCanvasPtr) l->c, &guiListFont);
	guiCanvasWantFocus((guiCanvasPtr) l->c, 1);
	return l;
}

void guiListDestroy(guiListPtr l)
{
	guiCanvasDestroy((guiCanvasPtr) l->c);
}

void guiListSetSorted(guiListPtr l, int flag)
{
	l->sorted = flag;
}

void guiListAddItem(guiListPtr l, char *text, void *itemData)
{
	if (!l->sorted) {
		guiListInsertItem(l, l->rowsTotal, text, itemData);
	} else {
		int i;
		int index = guiNone;
		for (i = 0; (i < l->rowsTotal); i++) {
			if (strcmp(l->text[i], text) > 0) {
				index = i;
				break;
			}
		}	
		if (index == guiNone) {
			index = l->rowsTotal;
		}
		guiListInsertItem(l, index, text, itemData);
	}
}

void guiListInsertItem(guiListPtr l, int index, char *text, void *itemData)
{
	int i;
	char *p;
	guiSize s;
	if (index < 0) {
		return;
	}
	if (index > l->rowsTotal) {
		return;
	}
	extend(l);
	p = (char *) malloc(strlen(text) + 1);
	if (!p) {
		return;
	}
	for (i = l->rowsTotal; (i > index); i--) {
		l->text[i] = l->text[i-1];
		l->itemData[i] = l->itemData[i-1];
	}
	l->text[index] = p;
	strcpy(l->text[index], text);
	l->itemData[index] = itemData;	
	s = guiCanvasGetVirtualSize((guiCanvasPtr) l->c);
	updateVirtualSizeForRow(l, l->rowsTotal, &s.w, &s.h);
	l->rowsTotal++;
	guiCanvasSetVirtualSize((guiCanvasPtr) l->c, s.w, s.h);
}

static void extend(guiListPtr l)
{
	int a;
	if (l->rowsTotal >= l->rowsAllocated) {
		char **t;
		void **d;
		a = l->rowsAllocated;
		while (a <= l->rowsTotal) {
			a *= 2;
			if (a < l->rowsAllocated) {
				/* Wraparound -- 16 bit windows -- ugh */
				return;
			}
		}
		t = (char **) realloc(l->text,
			a * sizeof(char *));
		if (!t) {
			return;
		}
		d = (void **) realloc(l->itemData,
			a * sizeof(void *));
		if (!d) {
			return;
		}
		l->rowsAllocated = a;
		l->text = t;
		l->itemData = d;
	}
}

void guiListRemoveItem(guiListPtr l, int index)
{
	int i;
	if (index <= 0) {
		return;
	}
	if (index >= l->rowsTotal) {
		return;
	}
	free(l->text[index]);
	for (i = index; (i < (l->rowsTotal - 1)); i++) {
		l->text[i] = l->text[i+1];
		l->itemData[i] = l->itemData[i+1];
	}
	if (index == l->selected) {
		l->selected = guiNone;
	} else if (index < l->selected) {
		l->selected--;
	} 
	updateVirtualSize(l);
}

int guiListFindItemByText(guiListPtr l, char *text)
{
	int i;
	for (i = 0; (i < l->rowsTotal); i++) {
		int r = strcmp(l->text[i], text);
		if (!r) {
			guiListSelectItem(l, i);
			return i;
		} else if (l->sorted) {
			if (r > 0) {
				if (i) {
					guiListSelectItem(l, i - 1);
				} else {
					guiListSelectItem(l, i);
				}
			}
		}
	}
	return guiNone;
}

int guiListFindItemByClientData(guiListPtr l, void *data)
{
	int i;
	for (i = 0; (i < l->rowsTotal); i++) {
		if (l->itemData[i] == data) {
			return i;
		}
	}
	return guiNone;
}

void guiListClear(guiListPtr l)
{
	l->rowsTotal = 0;
	l->selected = guiNone;
	updateVirtualSize(l);
}

void guiListSelectItem(guiListPtr l, int index)
{
	guiRectangle r;
	int check;
	if (index < 0) {
		return;
	}
	if (index >= l->rowsTotal) {
		return;
	}
	if (l->selected != guiNone) {
		int old = l->selected;
		l->selected = guiNone;
		drawRow(l, old);
	}	
	l->selected = index;
	r.x = 0;
	if (l->selected != guiNone) {
		r.y = l->selected * (guiListFont.ascent + guiListFont.descent);
	} else {
		r.y = 0;
	}
	check = (l->selected == guiNone) ? 0 : l->selected;
	if (l->selected != 0) {
		r.w = guiFontTextSize(&guiListFont,
			l->text[check], strlen(l->text[check])).w;
	} else {
		r.w = 0;
	}
	r.h = guiListFont.ascent + guiListFont.descent;
	guiCanvasScrollToRectangle((guiCanvasPtr) l->c, &r);
	if (l->selected != guiNone) {
		drawRow(l, l->selected);
	}
}

int guiListGetSelectedItem(guiListPtr l)
{
	return l->selected;
}

char *guiListGetItemText(guiListPtr l, int index)
{
	char *text;
	if (index < 0) {
		return 0;
	}
	if (index >= l->rowsTotal) {
		return 0;
	}
	text = (char *) malloc(strlen(l->text[index]) + 1);
	strcpy(text, l->text[index]);
	return text;	
}

void *guiListGetItemClientData(guiListPtr l, int index)
{
	if (index < 0) {
		return 0;
	}
	if (index >= l->rowsTotal) {
		return 0;
	}
	return l->itemData[index];
}

void guiListShow(guiListPtr l, int flag)
{
	guiCanvasShow((guiCanvasPtr) l->c, flag);
}

void guiListMove(guiListPtr l, int x, int y)
{
	guiCanvasMove((guiCanvasPtr) l->c, x, y);
}

void guiListSetPhysicalSize(guiListPtr l, int w, int h)
{
	guiCanvasSetPhysicalSize((guiCanvasPtr) l->c, w, h);
}

guiSize guiListGetPhysicalSize(guiListPtr l)
{
	return guiCanvasGetPhysicalSize((guiCanvasPtr) l->c);
}

void guiListSetActionHandler(guiListPtr l,                                      
	void (*actionHandler) (char *item, void *itemData, 
	void *clientData, int final))
{
	l->actionHandler = actionHandler;
}

static int eventHandler(guiCanvasPtr c, guiEventPtr e)
{
	guiListPtr l = (guiListPtr) guiCanvasGetClientData(c);
	guiSize s;
	int i;
	switch (e->t) {
		case guiEventDestroyed:
		guiListDestroyed(l);
		break;
		case guiEventRepaint:
		s = guiCanvasGetVirtualSize(c);
		for (i = 0; (i < l->rowsTotal); i++) {	
			int y = (guiListFont.ascent + guiListFont.descent) * i;
			if (((y + guiListFont.ascent + 
				guiListFont.descent) >= e->area.y) 
				&& (y < e->area.y + e->area.h)) 
			{
				drawRow(l, i);
			}
		}			
		break;
                case guiEventKeyDown:
		switch(e->key) {
			case guiKeySpace:
			case guiKeyEnter:
			if (l->selected != guiNone) {
				listAct(l, 1);
			}
			break;
			case guiKeyUp:
			if (!l->rowsTotal) {
				return 1;
			}
			if (l->selected == guiNone) {
				l->selected = 0;
				guiListSelectItem(l, 0);
				listAct(l, 0);
				return 1;
			}
			if (l->selected > 0) {
				guiListSelectItem(l, l->selected - 1);
				listAct(l, 0);
			}
			return 1;
			case guiKeyDown:
			if (!l->rowsTotal) {
				return 1;
			}
			if (l->selected == guiNone) {
				l->selected = 0;
				guiListSelectItem(l, 0);
				listAct(l, 0);
				return 1;
			}
			if (l->selected < (l->rowsTotal - 1)) {
				guiListSelectItem(l, l->selected + 1);
				listAct(l, 0);
			}
			return 1;
			case guiKeyPageUp:
			if (!l->rowsTotal) {
				return 1;
			}
			if (l->selected == guiNone) {
				l->selected = 0;
				guiListSelectItem(l, 0);
				listAct(l, 0);
				return 1;
			}
			if (l->selected > 0) {
				guiListSelectItem(l, 
					guiMax(l->selected - 
						guiCanvasGetVisibleSize(c).h / 
						(guiListFont.ascent + 
							guiListFont.descent), 
						0));
				listAct(l, 0);
			}
			return 1;
			case guiKeyPageDown:
			if (!l->rowsTotal) {
				return 1;
			}
			if (l->selected == guiNone) {
				l->selected = 0;
				guiListSelectItem(l, 0);
				listAct(l, 0);
				return 1;
			}	
			if (l->selected < (l->rowsTotal - 1)) {
				guiListSelectItem(l, 
					guiMin(l->selected + 
						guiCanvasGetVisibleSize(c).h / 
						(guiListFont.ascent + 
							guiListFont.descent), 
						l->rowsTotal - 1));
				listAct(l, 0);
			}			
			return 1;
			case guiKeyHome:
			if (l->rowsTotal > 0) {
				guiListSelectItem(l, 0);
				listAct(l, 0);
			}
			return 1;
			case guiKeyEnd:
			if (l->rowsTotal > 0) {
				guiListSelectItem(l, l->rowsTotal - 1);
				listAct(l, 0);
			}
			return 1;
			default:
			break;
		}
                break;
		case guiEventFocusIn:
		guiCanvasSetBorder((guiCanvasPtr) l->c, 2, guiBorderIn);
		break;
		case guiEventFocusOut:
		guiCanvasSetBorder((guiCanvasPtr) l->c, 2, guiBorderBackground);
		break;
                case guiEventTimer:
		l->clickWaiting = 0;	
                break;
                case guiEventMouseDown:
		if (!guiCanvasQueryFocus((guiCanvasPtr) l->c)) {
			guiCanvasTakeFocus((guiCanvasPtr) l->c);
		}
		i = e->l.y / (guiListFont.ascent + guiListFont.descent);
		if (i < l->rowsTotal) {
			guiListSelectItem(l, i);
			if (l->clickWaiting) {
				l->clickWaiting = 0;	
				guiCanvasClearTimer(c, 0);
				listAct(l, 1);
			} else {
				l->clickWaiting = 1;
				guiCanvasSetTimer(c, 0, 200);
				listAct(l, 0);
			}
		}	
                break;
		default:
		break;
	}
	return 0;
}

static void drawRow(guiListPtr l, int i)
{
	guiCanvasPtr c = (guiCanvasPtr) l->c;
	int x, y, w, len;
	int bw;
	x = 0;
	y = (guiListFont.ascent + guiListFont.descent) * i;
	if (l->selected == i) {
		guiCanvasSetDrawColor(c, 
			guiDefaultBackgroundColor);
		guiCanvasSetTextBackgroundColor(c, 
			guiListColor);
		guiCanvasSetFillColor(c, 
			guiListColor);
	} else {	
		guiCanvasSetDrawColor(c, 
			guiListColor);
		guiCanvasSetTextBackgroundColor(c, 
			guiDefaultBackgroundColor);
		guiCanvasSetFillColor(c, 
			guiDefaultBackgroundColor);
	}
	len = strlen(l->text[i]);
	guiCanvasDrawOpaqueText(c, x, y + guiListFont.ascent,
		l->text[i], len);
	w = guiFontTextSize(&guiListFont, l->text[i], len).w;
	bw = guiCanvasGetVirtualSize(c).w - w;
	guiCanvasFillRectangle(c, 
		w, 
		y,
		bw,
		guiListFont.ascent + guiListFont.descent);
}

static void updateVirtualSize(guiListPtr l)
{
	int w, h;
	int i;
	w = guiCanvasGetVisibleSize((guiCanvasPtr) l->c).w;
	h = 0;
	for (i = 0; (i < l->rowsTotal); i++)
	{
		updateVirtualSizeForRow(l, i, &w, &h);
	}
	if (!h) {
		h = 1;
	}
	guiCanvasSetVirtualSize((guiCanvasPtr) l->c, w, h);
}

static void updateVirtualSizeForRow(guiListPtr l, int i, int *wp, int *hp)
{
	guiSize s = guiFontTextSize(&guiListFont, l->text[i], 
			strlen(l->text[i]));
	if (s.w > *wp) {
		*wp = s.w;
	}
	*hp += (guiListFont.ascent + guiListFont.descent);
}

static void guiListDestroyed(guiListPtr l)
{
	int i;
	if (l->text) {
		for (i = 0; (i < l->rowsTotal); i++) {
			free(l->text[i]);
		}
		free(l->text);
	}
	if (l->itemData) {
		free(l->itemData);
	}
}

void guiListSetClientData(guiListPtr l, void *clientData)
{
	l->clientData = clientData;
}

void *guiListGetClientData(guiListPtr l)
{
	return l->clientData;
}

int guiListGetRowsTotal(guiListPtr l)
{
	return l->rowsTotal;
}

void listAct(guiListPtr l, int final)
{
	if (l->selected == guiNone) {
		return;
	}
	if (l->actionHandler) {
		l->actionHandler(l->text[l->selected],
			l->itemData[l->selected],
			l->clientData, final);
	}
}

