#include "guila.h"

guiFont guiLabelFont = {
        guiTimesRoman,
        12,
        0,
        0,
        0,
        0
};

guiColor guiLabelColor = {
        0, 0, 0
};

static int labelRender(guiLabelPtr l, char *text, int width);

static int eventHandler(guiCanvasPtr c, guiEventPtr e);

guiSize guiLabelEstimate(char *text)
{
	guiSize s;
	guiFontFindClosest(&guiLabelFont);
	s = guiFontTextSize(&guiLabelFont, text, strlen(text));
	s.h = guiLabelFont.ascent + guiLabelFont.descent;
	return s;
}

int guiLabelEstimateHeight(char *text, int width)
{
	guiFontFindClosest(&guiLabelFont);
	return labelRender(0, text, width);
}

static int labelRender(guiLabelPtr l, char *text, int width)
{
	int height = 0;
	char *tp = text;
	while(*tp) 
	{
		char *last = 0;
		int w;
		char *sp = tp;
		w = 0;
		while (1) {
			if (w > width) {
				if (!last) {
					if (l) {
						guiCanvasDrawText(
							(guiCanvasPtr) 
							l->c,
							0, height + guiLabelFont.ascent,
							tp, sp - tp);	
					}
					tp = sp;		
				} else {
					if (l) {
						guiCanvasDrawText(
							(guiCanvasPtr) 
							l->c,
							0, height + guiLabelFont.ascent,
							tp, last - tp); 
					}
					tp = last + 1;
				}
				break;
			}
			if (((*sp) == '\n') || ((*sp) == '\0')) {
				if (l) {
					guiCanvasDrawText(
						(guiCanvasPtr) 
						l->c,
						0, height + guiLabelFont.ascent,
						tp, sp - tp);	
				}
				if ((*sp) == '\n') {
					tp = sp + 1;
				} else {
					tp = sp;
				}
				break;
			}		
			if ((*sp) == ' ') {
				last = sp;
			}		
			w += guiFontTextSize(
				&guiLabelFont, sp, 1).w;
			sp++;
		}
		height += (guiLabelFont.ascent + guiLabelFont.descent);		
	}
	return height;
}

guiLabelPtr guiLabelCreate(guiCanvasPtr parent, 
	int x, int y, char *text)
{
	guiSize s = guiLabelEstimate(text);
	return guiLabelCreateWithSize(parent, x, y, s.w, s.h, text);
}

guiLabelPtr guiLabelCreateWithSize(guiCanvasPtr parent, 
	int x, int y, int w, int h, char *text)
{
	guiLabelPtr l = (guiLabelPtr) malloc(sizeof(guiLabel));
	if (!l) {
		return 0;
	}
        l->text = (char *) malloc(strlen(text) + 1);
        if (!l->text) {
                free(l);
                return 0;
        }
        strcpy(l->text, text);
	l->c = (guiNativeHandle) guiCanvasCreate(parent, x, y, w, h);
        guiCanvasSetClientData((guiCanvasPtr) l->c, (void *) l);
        guiCanvasSetEventHandler((guiCanvasPtr) l->c, eventHandler);
        guiCanvasSetEventMask((guiCanvasPtr) l->c, guiEventRepaint);
	guiCanvasSetFont((guiCanvasPtr) l->c, &guiLabelFont);
	guiCanvasSetDrawColor((guiCanvasPtr) l->c, guiLabelColor);
        return l;
}

void guiLabelDestroy(guiLabelPtr l)
{
	guiCanvasDestroy((guiCanvasPtr) l->c); 
}

void guiLabelSetText(guiLabelPtr l, char *newtext)
{
	char *t;
	t = malloc(strlen(newtext) + 1);
	if (!t) {
		return;
	}
	free(l->text);
	strcpy(t, newtext);	
	l->text = t;
	guiCanvasInvalidate((guiCanvasPtr) l->c, 0);
}

void guiLabelSetPhysicalSize(guiLabelPtr l, int w, int h)
{
	guiCanvasSetPhysicalSize((guiCanvasPtr) l->c, w, h);
}

void guiLabelMove(guiLabelPtr l, int x, int y)
{
	guiCanvasMove((guiCanvasPtr) l->c, x, y);
}

void guiLabelShow(guiLabelPtr l, int flag)
{
	guiCanvasShow((guiCanvasPtr) l->c, flag);
}

guiSize guiLabelGetPhysicalSize(guiLabelPtr l)
{
	return guiCanvasGetPhysicalSize((guiCanvasPtr) l->c);
}

static void guiLabelDestroyed(guiLabelPtr p)
{
	free(p->text);
}

static int eventHandler(guiCanvasPtr c, guiEventPtr e)
{
        guiLabelPtr l = (guiLabelPtr) guiCanvasGetClientData(c);
        switch (e->t) {
                case guiEventDestroyed:
                guiLabelDestroyed(l);
                break;
                case guiEventRepaint:
		labelRender(l, l->text, guiCanvasGetVirtualSize(c).w);
                break;
		default:
		break;
	}
	return 0;
}


