#include "guim.h"

static void guiMenuPopUp(guiMenuDataPtr m);
static void guiMenuPopDown(guiMenuDataPtr m);
static int buttonEventHandler(guiCanvasPtr c, guiEventPtr e);
static int bodyEventHandler(guiCanvasPtr c, guiEventPtr e);
static int barEventHandler(guiCanvasPtr c, guiEventPtr e);
static void guiMenuDestroyed(guiMenuDataPtr m);

guiFont guiMenuFont = {
	guiTimesRoman,
	12,
	1,
	1,
	0,
	0
};

guiColor guiMenuBackgroundColor = {
	128, 128, 255
};
	
guiColor guiMenuHighlightColor = {
	192, 255, 255
};

guiColor guiMenuCheckColor = {
	255, 128, 128
};

guiColor guiMenuLabelColor = {
	0, 0, 0 
};

guiMenuPtr guiMenuCreate(guiCanvasPtr parent, int x, int y)
{
	guiMenuPtr m = (guiMenuPtr) malloc(sizeof(guiMenu));
	if (!m) {
		return 0;
	}
	guiFontFindClosest(&guiMenuFont);
	m->first = 0;
	m->c = (guiNativeHandle) guiCanvasCreate(
		parent,
		x, y, 1, 
		guiMenuFont.ascent + guiMenuFont.descent + 4);
	guiCanvasSetClientData((guiCanvasPtr) m->c, m);
	guiCanvasSetEventMask((guiCanvasPtr) m->c, guiEventKeyDown);
	guiCanvasSetEventHandler((guiCanvasPtr) m->c, barEventHandler);
	return m;
}

guiMenuDataPtr guiMenuAdd(guiMenuPtr bar, char *label)
{
	guiMenuDataPtr m = (guiMenuDataPtr) malloc(sizeof(guiMenuData));
	guiMenuDataPtr p;
	guiSize s;
	char *at;
	int x;
	if (!m) {
		return 0;
	}
	m->bar = bar;
	/* Set the menu font to the closest available real font... */
	guiFontFindClosest(&guiMenuFont);
	/* Copy the label */
	m->label = (char *) malloc(strlen(label) + 1);
	if (!m->label) {
		free(m);
		return 0;
	}
	strcpy(m->label, label);
	/* Get the true label */
	at = strchr(m->label, '&');
	if (at) {
		m->hotkey = tolower(*(at + 1));
		m->hotkeyPos = at - m->label;
		/* Move the rest left one place */
		while(*at) {
			*at = *(at + 1);
			at++;
		}
	} else {
		m->hotkey = guiKeyNone;
	}		
	/* So size information will be accurate */
	s = guiFontTextSize(&guiMenuFont, m->label, strlen(m->label));
	m->first = 0;
	x = 0;
	p = bar->first;
	while (p) {
		x += guiCanvasGetPhysicalSize((guiCanvasPtr) p->button).w;	
		x += guiMenuGapEstimate();
		p = p->next;
	}
	guiCanvasSetPhysicalSize((guiCanvasPtr) bar->c, 
		x + guiMenuButtonEstimate(label).w + 
		guiMenuGapEstimate() + 4, 
		guiMenuFont.ascent + guiMenuFont.descent + 4); 
	m->button = (guiNativeHandle) guiCanvasCreate(
		(guiCanvasPtr) bar->c,
		x, 0, s.w + 4, 
		guiMenuFont.ascent + guiMenuFont.descent + 4);
	if (!m->button) {
		free(m);
		return 0;
	}
	guiCanvasSetBorder((guiCanvasPtr) m->button, 2, guiBorderBackground);
	guiCanvasSetClientData((guiCanvasPtr) m->button, (void *) m);
	guiCanvasSetEventMask((guiCanvasPtr) m->button,
		guiEventRepaint | guiEventMouseDown | guiEventMouseUp);	
	m->body = 0;
	guiCanvasSetEventHandler((guiCanvasPtr) m->button, buttonEventHandler);
	if (m->hotkey != guiKeyNone) {
		guiCanvasGrabKey((guiCanvasPtr) m->button,
			m->hotkey, 0, 0, 1); 
	}
	m->next = 0;
	p = bar->first;
	if (p) {
		while (p->next) {
			p = p->next;
		}
		p->next = m;
	} else {
		bar->first = m;	
	}
	m->selected = 0;
	guiCanvasShow((guiCanvasPtr) m->button, 1);
	return m;
}

guiSize guiMenuButtonEstimate(char *text)
{
	guiSize s, s2;
	char *at;
	guiFontFindClosest(&guiMenuFont);
	s = guiFontTextSize(&guiMenuFont, text, strlen(text)); 
	at = strchr(text, '&');
	if (at) {
		s2 = guiFontTextSize(&guiMenuFont, "&", 1);
		s.w -= s2.w;
	}	
	s.h = guiMenuFont.ascent + guiMenuFont.descent;
	s.w += 4;
	s.h += 4;
	return s;
}

int guiMenuGapEstimate()
{
	return 8;
}

static int buttonEventHandler(guiCanvasPtr c, guiEventPtr e)
{
	guiMenuDataPtr m = (guiMenuDataPtr) guiCanvasGetClientData(c);
	switch(e->t)
	{
		case guiEventRepaint:
		guiCanvasSetDrawColor(c,
			guiMenuLabelColor);
		guiCanvasSetTextBackgroundColor(c,
			guiDefaultBackgroundColor);
		guiCanvasSetFont(c, &guiMenuFont);
		guiCanvasDrawText(c, 
			0,
			guiMenuFont.ascent,
			m->label, strlen(m->label)); 
		if (m->hotkey != guiKeyNone) {
			guiSize s1 = guiFontTextSize(
				&guiMenuFont, m->label, m->hotkeyPos);	
			guiSize s2 = guiFontTextSize(
				&guiMenuFont, m->label, m->hotkeyPos + 1);	
			guiCanvasDrawLine(c, 
				s1.w, 
				guiMenuFont.ascent, 
				s2.w,
				guiMenuFont.ascent);
		}
		break;
		case guiEventDestroyed:
		guiMenuDestroyed(m);
		break;
		case guiEventMouseDown:
		guiMenuPopUp(m);
		break;
		case guiEventMouseUp:
		guiMenuPopDown(m);
		break;
		case guiEventKeyDown:
		if ((e->key == m->hotkey) && (e->alt == 1)) {
			guiMenuPopUp(m);
		} else {
			return 1;
		}
		default:
		break;
	}
	return 0;
}

void guiMenuDestroy(guiMenuPtr m)
{
	/* Destroy the bar. This will result in
		destroy events for the buttons, which
		will result in calls to guiMenuDestroyed(). */
	guiCanvasDestroy((guiCanvasPtr) m->c);
}

static void guiMenuDestroyed(guiMenuDataPtr m)
{
	guiMenuDataPtr p;
	guiMenuDataPtr prev;
	guiMenuItemPtr pi;
	prev = 0;
	/* Let the bar forget about this menu */
	p = m->bar->first;
	while (p) {
		if (p == m) {
			if (prev) {
				prev->next = p->next;
			} else {
				m->bar->first = p->next;
			}
		}
		prev = p;
		p = p->next;	
	}
	if (m->body) {
		guiCanvasSetClientData((guiCanvasPtr) m->body, 0);
		guiCanvasDestroy((guiCanvasPtr) m->body);
		m->body = 0;
	}
	if (m->hotkey != guiKeyNone) {
		guiCanvasReleaseKey((guiCanvasPtr) m->button,
			m->hotkey, 0, 0, 1); 
	}
	m->button = 0;
	free(m->label);
	pi = m->first;
	while (pi) {
		guiMenuItemPtr n = pi->next;
		free(pi->label);
		free(pi);
		pi = n;
	}	
	free(m);
}

guiMenuItemPtr guiMenuAddItem(guiMenuDataPtr m, char *text, 
	void *clientData, void (*actionHandler) 
		(char *text, void *clientData))
{
	guiMenuItemPtr p = (guiMenuItemPtr) malloc(sizeof(guiMenuItem));
	guiMenuItemPtr c;
	char *at;
	if (!p) {
		return 0;
	}
	p->label = (char *) malloc(strlen(text) + 1);
	if (!p->label) {
		free(p);
		return 0;
	}
	strcpy(p->label, text);
	/* Get the true label */
	at = strchr(p->label, '&');
	if (at) {
		p->hotkey = tolower(*(at + 1));
		p->hotkeyPos = at - p->label;
		/* Move the rest left one place */
		while(*at) {
			*at = *(at + 1);
			at++;
		}
	} else {
		p->hotkey = guiKeyNone;
	}		
	p->clientData = clientData;
	p->actionHandler = actionHandler;
	p->selected = 0;
	p->next = 0;
	p->checked = 0;
	if (!m->first) {
		m->first = p;
	} else {
		c = m->first;	
		while (c->next) {
			c = c->next;
		}
		c->next = p;
	}
	p->active = 1;
	return p;
}

void guiMenuItemSetActive(guiMenuItemPtr p, int flag)
{
	p->active = flag;
}

int guiMenuItemGetActive(guiMenuItemPtr p)
{
	return p->active;
}

void guiMenuItemSetCheck(guiMenuItemPtr p, int flag)
{
	p->checked = flag;
}

int guiMenuItemGetCheck(guiMenuItemPtr p)
{
	return p->checked;
}

static void guiMenuPopUp(guiMenuDataPtr m)
{
	guiPoint pt;
	int w = 4, h = 2;
	guiMenuItemPtr p = m->first;
	guiMenuDataPtr dp;
	guiSize cs = guiFontTextSize(&guiMenuFont, "M", 1);
	dp = m->bar->first;
	while (dp) {
		if (dp->selected) {
			guiMenuPopDown(dp);
			dp->selected = 0;
		}
		dp = dp->next;
	}
	m->selected = 1;	
	while (p) {
		int iw;
		h += guiMenuFont.ascent;
		h += guiMenuFont.descent;
		iw = guiFontTextSize(&guiMenuFont, p->label, 
			strlen(p->label)).w + 4;
		iw += cs.h * 2;
		if ((p == m->first) || (iw > w)) {
			w = iw;
		}
		p->selected = 0;
		p = p->next;	
	}			
	p = m->first;
	while (p) {
		if (p->active) {
			p->selected = 1;
			break;
		}
		p = p->next;
	}	
	h += 2;
	if (!m->body) {
		pt = guiCanvasGetScreenPosition((guiCanvasPtr) m->button);
		m->body = (guiNativeHandle) 
			guiCanvasCreate(0,
				pt.x, 
				pt.y + guiMenuFont.ascent +
					guiMenuFont.descent +
					4,
				w, h);	
		if (!m->body) {
			/* Sigh, no memory */
			return;
		}
		guiCanvasSetBorder((guiCanvasPtr) m->body, 2, guiBorderOut);
		guiCanvasSetType((guiCanvasPtr) m->body, 
			guiCanvasUndecorated, 
			m->button ? 
			guiCanvasGetParent(guiCanvasGetParent(
				(guiCanvasPtr) m->button)) :
			0);
		guiCanvasSetClientData((guiCanvasPtr) m->body, m);
		guiCanvasSetEventMask((guiCanvasPtr) m->body,
			guiEventMouseUp | 
			guiEventMouseDrag | 
			guiEventMouseDown |
			guiEventRepaint);
		guiCanvasSetEventHandler((guiCanvasPtr) m->body,
			bodyEventHandler);
		guiCanvasSetBackgroundColor((guiCanvasPtr) m->body,
			guiMenuBackgroundColor);
	} else {
		guiSize s = guiCanvasGetPhysicalSize((guiCanvasPtr) m->body);
		if ((w != s.w) || (h != s.h)) {
			guiCanvasSetSize((guiCanvasPtr) m->body,
				w, h);
		}
		pt = guiCanvasGetScreenPosition((guiCanvasPtr) m->button);
		guiCanvasMove((guiCanvasPtr) m->body,
			pt.x,
			pt.y + guiCanvasGetPhysicalSize(
					(guiCanvasPtr) m->button).h);
	}
	guiCanvasSetBorder((guiCanvasPtr) m->button, 2, guiBorderOut);
	guiCanvasShow((guiCanvasPtr) m->body, 1);
	/* Grab keys */
	guiCanvasGrabKey((guiCanvasPtr) m->body, guiKeyEnter, 0, 0, 0); 
	guiCanvasGrabKey((guiCanvasPtr) m->body, guiKeyEscape, 0, 0, 0); 
	guiCanvasGrabKey((guiCanvasPtr) m->body, guiKeyUp, 0, 0, 0); 
	guiCanvasGrabKey((guiCanvasPtr) m->body, guiKeyDown, 0, 0, 0); 
	guiCanvasGrabKey((guiCanvasPtr) m->bar->c, guiKeyLeft, 0, 0, 0); 
	guiCanvasGrabKey((guiCanvasPtr) m->bar->c, guiKeyRight, 0, 0, 0); 
	p = m->first;
	while (p) {
		if (p->hotkey != guiKeyNone) {
			guiCanvasGrabKey(
				(guiCanvasPtr) m->body, 
				tolower(p->hotkey), 0, 0, 0); 
		}
		p = p->next;	
	}
}

static void guiMenuPopDown(guiMenuDataPtr m)
{
	if (m->body) {
		guiMenuItemPtr p;
		guiCanvasShow((guiCanvasPtr) m->body, 0);
		/* Release keys */
		guiCanvasReleaseKey((guiCanvasPtr) m->body, 
			guiKeyEnter, 0, 0, 0); 
		guiCanvasReleaseKey((guiCanvasPtr) m->body, 
			guiKeyEscape, 0, 0, 0); 
		guiCanvasReleaseKey((guiCanvasPtr) m->body, 
			guiKeyUp, 0, 0, 0); 
		guiCanvasReleaseKey((guiCanvasPtr) m->body, 
			guiKeyDown, 0, 0, 0); 
		guiCanvasReleaseKey((guiCanvasPtr) m->body, 
			guiKeyEscape, 0, 0, 0); 
		guiCanvasReleaseKey((guiCanvasPtr) m->bar->c,
			guiKeyLeft, 0, 0, 0); 
		guiCanvasReleaseKey((guiCanvasPtr) m->bar->c,
			guiKeyRight, 0, 0, 0); 
		p = m->first;
		while (p) {
			if (p->hotkey != guiKeyNone) {
				guiCanvasReleaseKey(
					(guiCanvasPtr) m->body, 
					tolower(p->hotkey), 0, 0, 0); 
			}
			p = p->next;	
		}
	}
	guiCanvasSetBorder((guiCanvasPtr) m->button, 2, guiBorderBackground);
	m->selected = 0;
}

static int bodyEventHandler(guiCanvasPtr c, guiEventPtr e)
{
	guiMenuDataPtr m = (guiMenuDataPtr) guiCanvasGetClientData(c);
	int y;
	guiMenuItemPtr p, invoke;
	int found;
	guiSize s = guiCanvasGetVisibleSize(c);
	guiSize cs = guiFontTextSize(&guiMenuFont, "M", 1);
	switch(e->t)
	{
		case guiEventRepaint:
		guiCanvasSetFont(c, &guiMenuFont);
		guiCanvasSetDrawColor(c,
			guiMenuLabelColor);
		y = 0;
		p = m->first;
		while (p) {
			if (p->selected) {
				guiCanvasSetFillColor(c, 
					guiMenuHighlightColor);
				guiCanvasSetTextBackgroundColor(c,
					guiMenuHighlightColor);
				guiCanvasFillRectangle(c,
					0, y,
					guiCanvasGetVisibleSize(c).w,
					guiMenuFont.ascent + 
						guiMenuFont.descent);
			}
			if (!p->active) {
				guiCanvasSetDim(c, 1);
			}
			guiCanvasDrawText(c, 
				0, y + guiMenuFont.ascent, 
				p->label, strlen(p->label)); 
			if (p->checked) {
				int x = s.w - cs.h / 2 - cs.h;
				guiCanvasSetFillColor(c, 
					guiMenuCheckColor);
				guiCanvasFillRectangle(c,
					x, 2 + y,
					cs.h, cs.h);
				guiCanvasDrawRectangle(c,
					x, 2 + y,
					cs.h, cs.h);
				guiCanvasDrawLine(c,
					x + 1, 2 + 1 + y,
					x + cs.h - 1, y + 1 + cs.h);
				guiCanvasDrawLine(c,
					x + cs.h - 1, 2 + 1 + y,
					x + 1, y + 1 + cs.h);
			}
			if (p->hotkey != guiKeyNone) {
				guiSize s1, s2;
				s1 = guiFontTextSize(
					&guiMenuFont, p->label, 
					p->hotkeyPos);	
				s2 = guiFontTextSize(
					&guiMenuFont, p->label, 
					p->hotkeyPos + 1);	
				guiCanvasDrawLine(c, 
					s1.w, 
					y + guiMenuFont.ascent, 
					s2.w,
					y + guiMenuFont.ascent);
			}
			if (!p->active) {
				guiCanvasSetDim(c, 0);
			}
			y += guiMenuFont.ascent;
			y += guiMenuFont.descent;
			p = p->next;	
		}
		break;
		case guiEventDestroyed:
		/* Make sure the menu isn't already gone. If it is alive,
			tell it the body canvas is gone. */
		if (m) {
			m->body = 0;
		}
		break;
		case guiEventMouseDrag:
		y = 0;
		p = m->first;
		found = 0;
		while (p) {
			int rh = guiMenuFont.ascent + guiMenuFont.descent;
			int paint = 0;
			y += rh;
			if ((e->l.y < y) && (!found)) {
				if (!p->selected) {
					if (p->active) {
						p->selected = 1;
					} else {
						p->selected = 0;
					}
					paint = 1;
				}
				found = 1;
			} else {
				if (p->selected) {
					p->selected = 0;
					paint = 1;
				}
			}
			if (paint) {
				guiRectangle r;
				r.x = 0;
				r.y = y - rh;
				r.w = s.w;
				r.h = rh;
				guiCanvasInvalidate(c, &r);
			}
			p = p->next;	
		}		
		break;
		case guiEventMouseUp:
		y = 0;
		p = m->first;
		found = 0;
		invoke = 0;
		while (p) {
			y += guiMenuFont.ascent;
			y += guiMenuFont.descent;
			if (e->l.y < y) {
				if (!found) {
					if (p->active) {
						invoke = p;
						found = 1;
					}
				}
			}
			p->selected = 0;
			p = p->next;	
		}
		guiMenuPopDown(m);
		if (invoke) {
			if (invoke->actionHandler) {
				invoke->actionHandler(invoke->label, 
					invoke->clientData);
			}
		}
		break;
		case guiEventKeyDown:
		p = m->first;
		while (p) {
			if (p->hotkey != guiKeyNone) {
				if (tolower(e->key) == p->hotkey) {
					if (p->active) {
						guiMenuPopDown(m);
						if (p->actionHandler) {
							p->actionHandler(
								p->label, 
								p->clientData);
						}
						return 1;
					}
				}	
			}
			p = p->next;	
		}		
		if (e->key == guiKeyUp) {
			guiMenuItemPtr prev;
			prev = 0;
			p = m->first;
			while (p) {
				if (p->selected) {
					p->selected = 0;
					if (prev) {
						prev->selected = 1;	
					} else {
						while (p->next) {
							p = p->next;
							if (p->active) {
								prev = p;
							}
						}
						if (prev) {
							prev->selected = 1;
						}
					}
					guiCanvasInvalidate(
						(guiCanvasPtr) m->body, 0);
					return 1;
				}
				if (p->active) {
					prev = p;
				}
				p = p->next;
			}	
		}
		if (e->key == guiKeyDown) {
			p = m->first;
			while (p) {
				if (p->selected) {
					int found = 0;
					p->selected = 0;
					while (p->next) {
						if (p->next->active) {
							found = 1;
							p->next->selected = 1;
							break;
						}
						p = p->next;
					}
					if (!found) {
						p = m->first;
						while (p) {
							if (p->active) {
								p->selected = 1;
								found = 1;
								break;
							}
							p = p->next;
						}
					}	
					guiCanvasInvalidate(
						(guiCanvasPtr) m->body, 0);
					return 1;
				}
				p = p->next;
			}	
		}		
		if (e->key == guiKeyEscape) {
			guiMenuPopDown(m);
			return 1;
		}
		if (e->key == guiKeyEnter) {
			p = m->first;
			while (p) {
				if (p->selected) {
					guiMenuPopDown(m);
					if (p->actionHandler) {
						p->actionHandler(p->label, 
							p->clientData);
					}
					return 1;
				}
				p = p->next;
			}
		}	
		break;
		default:
		break;
	}
	return 0;
}

static int barEventHandler(guiCanvasPtr c, guiEventPtr e)
{
	guiMenuPtr m = (guiMenuPtr) guiCanvasGetClientData(c);
	guiMenuDataPtr p;
	guiMenuDataPtr prev;
	switch (e->t) {
		case guiEventKeyDown:
		if (e->key == guiKeyLeft) {
			prev = 0;
			p = m->first;
			while (p) {
				if (p->selected) {
					if (prev) {
						guiMenuPopUp(prev);
					} else {
						while (p->next) {
							p = p->next;
						}
						guiMenuPopUp(p);
					}
					return 1;
				}
				prev = p;
				p = p->next;
			}	
		}
		if (e->key == guiKeyRight) {
			p = m->first;
			while (p) {
				if (p->selected) {
					if (p->next) {
						guiMenuPopUp(p->next);
					} else {
						guiMenuPopUp(m->first);
					}
					return 1;
				}
				p = p->next;
			}	
		}	
		break;
		default:
		break;
	}
	return 0;
}

guiSize guiMenuGetPhysicalSize(guiMenuPtr m)
{
	return guiCanvasGetPhysicalSize((guiCanvasPtr) m->c);
}

void guiMenuShow(guiMenuPtr m, int flag)
{
	guiCanvasShow((guiCanvasPtr) m->c, flag);
}

