#include "guit.h"
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <math.h>

guiFont guiTextFont = {
        guiTimesRoman,
        12,
        0,
        0,
        0,
        0
};

guiColor guiTextColor = {
	0, 0, 0	
};

static void guiTextScrollbarUpdate(guiTextPtr t);

static void guiTextWarpToCursor(guiTextPtr t);

static guiTextResult guiTextRowInsertCols(guiTextPtr t, 
	int row, int col, char *text, int length);

static void guiTextRowColsTotalSet(guiTextPtr t, 
	int row, int colsTotal);

static int textEventHandler(guiCanvasPtr c, guiEventPtr e);

static guiTextResult guiTextRowInsert(guiTextPtr t, int row);
 
static void guiTextRepaint(guiCanvasPtr w);

static void guiTextMouseDown(guiCanvasPtr w, int x, int y, int button);

static void guiTextMouseDrag(guiCanvasPtr w, int x, int y, int button);

static void guiTextMouseUp(guiCanvasPtr w, int x, int y, int button);

static int guiTextKeyDown(guiCanvasPtr w, guiKey key,
	int shift, int control, int alt);

static guiTextPtr guiTextCreate(guiCanvasPtr parent, int x, int y,
	int width, int height, int colsMax,
	int charsMax, int multiline);

static guiTextResult guiTextInsertAt(
	guiTextPtr t, int row, int col, char *text, int len);

static int guiTextBackOff(guiTextPtr t, int chars);

void guiTextDestroyed(guiCanvasPtr w);

void guiTextRowsTotalSet(guiTextPtr t, int rowsTotal);

static void guiTextRefresh(guiTextPtr t);

static void guiTextRepaintBody(guiTextPtr t, int repaintAll);

static void guiTextSortSelection(guiTextPtr t,
	int *sRowP, int *sColP, int *eRowP, int *eColP);

static void guiTextFetchSelection(guiTextPtr t);

static void guiTextStoreSelection(guiTextPtr t);

static void guiTextDeleteSelection(guiTextPtr t);

static void guiTextBlinkStart(guiCanvasPtr c);

static void guiTextBlinkStop(guiCanvasPtr c);

static void guiTextBlink(guiCanvasPtr c);

static void guiTextCursorDraw(guiTextPtr t);

static void guiTextCursorHide(guiTextPtr t);

static void guiTextCursorDrawBody(guiTextPtr t, int on);

guiTextPtr guiTextCreateSingle(guiCanvasPtr parent, 
	int x, int y, int width, int height, 
	int charsMax)
{
	guiTextPtr t = guiTextCreate(
		parent, x, y, width, height, guiNone,
		charsMax, 0);
	/* No horizontal scrollbar for single-line text areas;
		it always looks awful */
	guiCanvasAllowHorizontalScrollbar(t->c, 0);
	return t;
}

guiTextPtr guiTextCreateMultiple(guiCanvasPtr parent, 
	int x, int y, int width, int height, 
	int colsMax)
{
	return guiTextCreate(
		parent, x, y, width, height, colsMax, guiNone, 
		1);
}

static guiTextPtr guiTextCreate(guiCanvasPtr parent, int x, int y,
	int width, int height, int colsMax, int charsMax, int multiline)
	
{
	guiTextPtr t;
	int rowsAlloc;
	int i;
	/* Get the real text font data pronto */ 
	guiFontFindClosest(&guiTextFont);
	t = (guiTextPtr ) malloc(sizeof(guiText));
	t->c = (guiCanvasPtr ) guiCanvasCreate(
		parent, x, y, width, height);
	if (!t->c) {
		return 0;
	}
	guiCanvasSetBorder(t->c, 2, guiBorderIn);
	guiCanvasSetClientData(t->c, (void *) t);
	if (!multiline) {
		rowsAlloc = 1;
	} else {
		rowsAlloc = 32;
	}
	t->multiline = multiline;
	t->rowsAllocated = rowsAlloc;
	t->text = (char **) malloc(sizeof(char *) * rowsAlloc);
	t->colsAllocated = (int *) malloc(sizeof(int) * rowsAlloc);
	t->colsTotal = (int *) malloc(sizeof(int) * rowsAlloc);
	for (i=0; (i < rowsAlloc); i++) {
		t->text[i] = (char *) malloc(80);
		t->colsAllocated[i] = 80;
		t->colsTotal[i] = 0;	
	}
	t->rowsTotal = 1;
	t->rowsTotalLast = 0;
	t->row = 0;
	t->col = 0;
	t->charsMax = charsMax;
	t->colsMax = colsMax;
	t->wordWrap = 1;
	t->selection = 0;
	t->readOnly = 0;
	t->pixelWidest = 0;
	t->rowWidest = 0;
	t->active = 1;
	t->updateHandler = 0;
	t->clientData = 0;
	t->charsTotal = 0;
	guiCanvasSetEventHandler(t->c, textEventHandler);
	guiCanvasSetEventMask(t->c,
		guiEventMouseDown + guiEventMouseDrag +
		guiEventMouseUp + guiEventKeyDown +
		guiEventDestroyed + guiEventRepaint +
		guiEventFocusIn + guiEventFocusOut +
		guiEventResize);
	guiCanvasSetClearing(t->c, 0);
	guiCanvasWantFocus(t->c, 1);
	t->repaintAll = 1;
	t->blinkOn = 1;
	return t;
}
	
void guiTextDestroyed(guiCanvasPtr w) 
{
	int i;
	guiTextPtr t = (guiTextPtr ) guiCanvasGetClientData(w);
	for (i=0; (i < t->rowsAllocated); i++) {
		free(t->text[i]);
	}
	free(t->colsAllocated);
	free(t->colsTotal);
}

/* Estimate the size for a single-line text area with the
        specified number of columns (visible). */

guiSize guiTextEstimateSingle(int cols)
{
	return guiTextEstimateMultiple(1, cols);
}

/* Estimate the size for a multiple-line text area with the
        specified number of rows and columns (all of them visible). */

guiSize guiTextEstimateMultiple(int rows, int cols)
{
	guiSize s;
	guiFontFindClosest(&guiTextFont);
	s = guiFontTextSize(&guiTextFont, "M", 1);
	s.w *= cols;
	s.h = (guiTextFont.ascent + guiTextFont.descent) * rows;
	s.w += 4;
	s.h += 4;
	return s;
}

void guiTextSetWordWrap(guiTextPtr t, int wrap)
{
	t->wordWrap = wrap;
}

void guiTextSetReadOnly(guiTextPtr t, int readOnly)
{
	t->readOnly = readOnly;
}

void guiTextDestroy(guiTextPtr t)
{
	guiCanvasDestroy(t->c);
}

void guiTextSet(guiTextPtr t, char *text, int len)
{
	guiTextRowsTotalSet(t, 0);
	guiTextInsertAt(t, 0, 0, text, len);
	guiTextRefresh(t);
	if (t->updateHandler) {
		t->updateHandler(t, t->clientData);
	}
}

int breakPointFind(guiTextPtr t,
	char *line, int max);

guiTextResult guiTextInsert(
	guiTextPtr t, char *text, int len)
{
	return guiTextInsertAt(t, t->row, t->col, text, len);
}

guiTextResult guiTextInsertAt(
	guiTextPtr t, int row, int col, char *text, int len)
{
	/*
	**	tail-equals-existing-line-prior-to-insertion;
	**	last-equals-existing-line-following-insertion;
	**	destroy-existing-line;
	**	while(new-lines) {
	**		get-new-line;
	**		line-equals-tail-plus-new-line;
	**		if (last-new-line) {
	**			line += last;
	**		}	
	**		wrap-line-into-sublines;	
	**		tail-equals-nothing;
	**	}
	*/
	int excess = 0;
	int charsRemaining = len;
	char *crat;
	char *line;
	int result = guiTextSucceeded;
	int linePos;
	char *tail = t->text[row];
	int tailLen = col;
	char *last = 0;
	int lastLen = t->colsTotal[row] - col; 
	/* Any insertion invalidates the selection */
	if (t->selection) {
		t->selection = 0;
		t->repaintAll = 1;
	}
	if (!t->multiline) {
		if (t->charsMax == guiNone) {
			excess = 0;
		} else {
			excess = len - (t->charsMax - t->charsTotal);  
		}
		if (excess > 0) {
			len -= excess;
			result = guiTextFull;
		} else {
			result = guiTextSucceeded;
		}
		excess = 0;
		while ((crat = memchr(text, '\n', len))) {
			/* Cursor adjustment */
			t->col += (crat - text);
			t->col++;
			guiTextRowInsertCols(
				t, 0, 
				col, text,
				crat - text);
			t->charsTotal += (crat - text);
			text = crat + 1;
			col += (crat - text);
			len -= (crat - text);
			guiTextRowInsertCols(
				t, 0, col,
				" ",
				1);
			t->charsTotal++;
			col++;
			len--;
			
		}
		guiTextRowInsertCols(
			t, 0,
			col, text,
			len);
		t->col += len;
		t->charsTotal += len;
		return result;
	}
	/* Multiline case */
	if (lastLen) {
		last = (char *) malloc(lastLen);
		memcpy(last, t->text[row] + col, lastLen);
	}
	do {
		char *newLine = text;
		int newLineLen; 
		int lineLen;

		crat = memchr(text, '\n', len);
		if (crat) {
			newLineLen = crat - text;
		} else {
			newLineLen = len;
		} 		
		if (!crat) {
			lineLen = lastLen + tailLen + newLineLen;
			line = (char *) malloc(lineLen);
			charsRemaining += (lastLen + tailLen);
			memcpy(line, tail, tailLen);
			memcpy(line + tailLen, newLine, newLineLen);
			if (lastLen) {
				memcpy(line + tailLen + newLineLen, 
					last, lastLen);
			}
		} else if (tailLen) {
			lineLen = tailLen + newLineLen;
			charsRemaining += tailLen;
			line = (char *) malloc(lineLen);
			memcpy(line, tail, tailLen);
			memcpy(line + tailLen, newLine, newLineLen);
		} else {
			lineLen = newLineLen;
			line = newLine;
		}
		/* Truncate old contents of row */
		guiTextRowColsTotalSet(t, row, 0);
		linePos = 0;
		while (lineLen) {
			if (t->colsMax != guiNone) {
				if (lineLen > t->colsMax) {
					int breakAfter;
					breakAfter = breakPointFind(
						t, line + linePos, t->colsMax);	
					guiTextRowInsertCols(
						t, row, 0, 
						line + linePos,
						breakAfter);
					charsRemaining -= breakAfter;
					lineLen -= breakAfter;
					linePos += breakAfter;
					row++;
					guiTextRowInsert(t, row);
				} else {
					guiTextRowInsertCols(
						t, row, 0, 
						line + linePos,
						lineLen);
					charsRemaining -= lineLen;
					linePos += lineLen;
					lineLen = 0;
				}
			} else {
				guiTextRowInsertCols(
					t, row, 0, 
					line + linePos,
					lineLen);
				charsRemaining -= lineLen;
				linePos = lineLen;
				lineLen = 0;
			}		
		}
		if (line != newLine) {
			free(line);
		}
		tail = "";
		tailLen = 0;
		if (crat) {
			charsRemaining--;
			row++;
			guiTextRowInsert(t, row);
			text = crat + 1;
			len -= newLineLen;
			/* Minus one more for the CR */
			len--;
		}
	} while(crat);
	t->row = row;
	t->col = t->colsTotal[row];		
	/* Back off the cursor */
	guiTextBackOff(t, lastLen);
	if (last) {
		free(last);
	}
	return result;
}

int guiTextGetLength(guiTextPtr t)
{
	int length = 0;
	int i;
	for (i=0; (i < t->rowsTotal); i++) {
		length += t->colsTotal[i];
		if (i != (t->rowsTotal - 1)) {
			/* Account for the LF that will be added */
			length++;
		}
	}
	return length;
}	
	
char *guiTextGet(guiTextPtr t)
{
	int len = guiTextGetLength(t);
	char *result = (char *) malloc(len + 1);
	if (!result) {
		return 0;
	}
	guiTextGetBuffer(t, result, len + 1, 0);
	return result;
}

void guiTextGetBuffer(guiTextPtr t, char *buf, int bufSize, int *textLen)
{
	int i;
	int length = 0;
	for (i=0; (i < t->rowsTotal); i++) {
		int copy = t->colsTotal[i];
		if (length + copy + 1 > bufSize) {
			copy = bufSize - length - 2;
		}
		length += copy;
		memcpy(buf, t->text[i], copy);
		if (i != (t->rowsTotal - 1)) {
			buf[copy] = '\n';
		} else {
			buf[copy] = '\0';
		}
		length++;
		buf += copy;
		buf++;
	}
	if (textLen) {
		*textLen = length;
	}
}

void guiTextRowsTotalSet(guiTextPtr t, int rowsTotal) 
{
	int i;
	if (rowsTotal == 0) {
		/* A reset. */
		rowsTotal = 1;
		t->colsTotal[0] = 0;
		t->charsTotal = 0;
		t->row = 0;
		t->col = 0;
	}	
	/* Don't let the cursor, or the selection, float in the void */
	if (t->row >= rowsTotal) {
		t->row = rowsTotal;
		t->col = 0;
	}	
	if (t->sRow >= rowsTotal) {
		t->sRow = rowsTotal;
		t->sCol = 0;
	}	
	if (rowsTotal > t->rowsAllocated) {
		int rowsAlloc;
		while (rowsTotal > t->rowsAllocated) {
			t->rowsAllocated <<= 1;
		}
		rowsAlloc = t->rowsAllocated;
		t->text = (char **) realloc(t->text, 
			sizeof(char *) * rowsAlloc);
		t->colsAllocated = (int *) realloc(t->colsAllocated,
			sizeof(int) * rowsAlloc);
		t->colsTotal = (int *) realloc(t->colsTotal,
			sizeof(int) * rowsAlloc);
	}
	for (i = t->rowsTotal; (i < rowsTotal); i++) {
		t->text[i] = (char *) malloc(80);
		t->colsAllocated[i] = 80;
		t->colsTotal[i] = 0;	
	}			
	t->rowsTotal = rowsTotal;
	t->repaintAll = 1;
}

guiTextResult guiTextRowInsertCols(guiTextPtr t, 
	int row, int col, char *text, int length)
{
	int i;
	int w;
	if (row >= t->rowsTotal) {
		guiTextRowsTotalSet(t, row+1);
	}
	guiTextRowColsTotalSet(t, row, length + t->colsTotal[row]);
	for (i = (t->colsTotal[row]) - 1; (i > col); i--) {
		t->text[row][i] = t->text[row][i - length];
	}
	/* Copy new text into place */
	memcpy(t->text[row] + col, text, length);
	w = guiFontTextSize(&guiTextFont, t->text[row], t->colsTotal[row]).w;
	return guiTextSucceeded;
}

static void guiTextRowColsTotalSet(guiTextPtr t, 
	int row, int colsTotal)
{
	if (colsTotal > t->colsAllocated[row]) {
		while (colsTotal > t->colsAllocated[row]) {
			t->colsAllocated[row] <<= 1;	
		}
		t->text[row] = (char *) realloc(t->text[row], 
			t->colsAllocated[row]);
	}
	t->colsTotal[row] = colsTotal;
}

static guiTextResult guiTextRowInsert(guiTextPtr t, int row) 
{
	int i;
	char *text;
	int colsTotal;
	int colsAllocated;
	guiTextRowsTotalSet(t, t->rowsTotal+1);
	/* Now the total has been incremented.
		Hold on to the space assigned to the new line,
		and move everything else up... */
	text = t->text[t->rowsTotal-1];
	colsTotal = 0;
	colsAllocated = t->colsAllocated[t->rowsTotal-1];
	for (i = (t->rowsTotal-1); (i > row); i--) {
		t->text[i] = t->text[i-1];
		t->colsAllocated[i] = t->colsAllocated[i-1];
		t->colsTotal[i] = t->colsTotal[i-1];
	} 
	/* Now assign the space for the new line to
		the inserted position. */	
	t->text[row] = text;
	t->colsTotal[row] = colsTotal;
	t->colsAllocated[row] = colsAllocated;
	t->repaintAll = 1;
	return guiTextSucceeded;
}

void guiTextDelete(guiTextPtr t, 
	int row, int col, int toRow, int toCol)
{
	/* Any deletion invalidates the selection */
	if (t->selection) {
		t->selection = 0;
		t->repaintAll = 1;
	}
	if (row < 0) {
		row = 0;
	} else if (row >= t->rowsTotal) {
		row = t->rowsTotal - 1;
	}
	if (col < 0) {
		col = 0;
	} else if (col > t->colsTotal[row]) {
		col = t->colsTotal[row];
	}
	if (toRow < row) {
		return;
	} else if (toRow >= t->rowsTotal) {
		toRow = t->rowsTotal - 1;
		toCol = t->colsTotal[toRow];
	}
	if ((toRow == row) && (toCol < col)) {
		return;
	} else if (toCol < 0) {
		toCol = 0;
	} else if (toCol > t->colsTotal[toRow]) {
		toCol = t->colsTotal[toRow];
	}
	/* Cursor will wind up at the starting point of the deletion */
	t->row = row;
	t->col = col;
	if (row == toRow) {
		/* Simple one-line delete operation */
		int i;
		/* Move stuff */
		int deleted = toCol - col;
		int len = t->colsTotal[row];
		t->charsTotal -= deleted;
		for (i = col; (i < (len - deleted)); i++) {
			t->text[row][i] = 
				t->text[row][i + deleted];
		}
		guiTextRowColsTotalSet(t, row, len - deleted);
	} else {
		int i;
		int deleted = toRow - row;
		int charsInserted;
		/* Multiline; more complicated. */
		/* 1. Shorten the first line. */
		guiTextRowColsTotalSet(t, row, col);
		/* 2. Attach the surviving component of the last line. */
		charsInserted = t->colsTotal[toRow] - toCol;
		guiTextInsertAt(t, row, col,
			t->text[toRow] + toCol, charsInserted);
		/* 3. Adjust row and toRow for any lines
			inserted by line breaks in the insert operation.
			Use the new cursor position to determine this. */
		toRow += (t->row - row);
		row = t->row;
		/* 4. Back the cursor up in a way that
			respects any line breaks that may
			have been made during the insert call. */
		guiTextBackOff(t, charsInserted);

		if (deleted) {
			t->repaintAll = 1;
		}

		/* 5. Free the deleted rows. */
		for (i=0; (i < deleted); i++) {
			free(t->text[row + 1 + i]);
		}

		/* 6. Move the later rows up. */
		for (i = row + 1; (i < (t->rowsTotal - deleted)); i++) {
			t->text[i] = t->text[i + deleted];
			t->colsAllocated[i] = t->colsAllocated[i + deleted];
			t->colsTotal[i] = t->colsTotal[i + deleted];
		}	

		/* 7. Set the new number of total and safely allocated
			rows. Don't call the usual function because it won't
			expect invalid pointers to the highest rows. */
		t->rowsTotal -= deleted;
		t->rowsAllocated = t->rowsTotal;
	}	
	if (t->updateHandler) {
		t->updateHandler(t, t->clientData);
	}
}

int breakPointFind(guiTextPtr t,
	char *line, int max)
{
	if (!(t->wordWrap)) {
		return max;
	} else {
		int i = max-1;
		while (i) {
			if (isspace(line[i])) {
				return i+1;
			}
			i--;
		}
		/* Don't forget the case in which a word
			is too long to be broken elegantly! */
		return max;	
	}
}

static int guiTextRowInSelection(guiTextPtr t, int row,
	int *left, int *right);

static void guiTextRepaint(guiCanvasPtr w)
{
	guiTextPtr t = guiCanvasGetClientData(w);
	/* It's coming from the window system,
		so we repaint it all */
	guiTextRepaintBody(t, 1);
}

static void guiTextRefresh(guiTextPtr t)
{
	guiCanvasPtr w = t->c;
	guiTextScrollbarUpdate(t);
	if (t->repaintAll) {	
		guiCanvasInvalidate(w, 0);
	} else {
		guiTextRepaintBody(t, 0);
	}
}

static void guiTextRepaintBody(guiTextPtr t, int repaintAll)
{
	int r;
	int width;
	int height = 0;
	guiSize s;
	guiCanvasPtr w = t->c;
	guiCanvasSetFont(w, &guiTextFont);
	guiCanvasSetDrawColor(w, guiTextColor);
	if (!t->active) {
		guiCanvasSetDim(w, 1);
	}
	guiCanvasSetTextBackgroundColor(w, guiDefaultBackgroundColor);
	for (r = 0; (r < t->rowsTotal); r++) {
		int left, right;
		int x = 0;
		int y = guiTextFont.ascent + height; 
		if ((!repaintAll) && (r != t->row)) {
			goto advance;
		}
		if (guiTextRowInSelection(t, r, &left, &right)) {
			width = guiFontTextSize(&guiTextFont,
				t->text[r], left).w; 
			if (left) {
				guiCanvasDrawOpaqueText(w,
					x, y, 
					t->text[r], left);
				x += width;
			}
			width = guiFontTextSize(&guiTextFont,
				t->text[r] + left, right - left).w; 
			guiCanvasSetDrawColor(w, guiDefaultBackgroundColor);
			guiCanvasSetTextBackgroundColor(w, guiTextColor);
			guiCanvasDrawOpaqueText(w,
				x, y, 
				t->text[r] + left, right - left);
			guiCanvasSetDrawColor(w, guiTextColor);
			guiCanvasSetTextBackgroundColor(w, 
				guiDefaultBackgroundColor);
			x += width;
			if (right != t->colsTotal[r]) {
				width = guiFontTextSize(&guiTextFont,
					t->text[r] + right, 
					t->colsTotal[r] - right).w; 
				guiCanvasDrawOpaqueText(
					w, x, y, 
					t->text[r] + right, 
					t->colsTotal[r] - right);
				x += width;
			}
		} else {
			width = guiFontTextSize(&guiTextFont,
				t->text[r], t->colsTotal[r]).w; 
			guiCanvasDrawOpaqueText(w,
				x, y,
				t->text[r], t->colsTotal[r]);
			x += width;
		}
		if ((guiCanvasGetVirtualSize(w).w - x) > 0) {
			guiCanvasSetFillColor(w, guiDefaultBackgroundColor);
			guiCanvasFillRectangle(w,
				x, height, 
				guiCanvasGetVirtualSize(w).w - x, 
				guiTextFont.ascent + guiTextFont.descent);
		}
		if ((r == t->row) && (t->blinkOn)) {
			int pos;
			pos = guiFontTextSize(&guiTextFont,
				t->text[r], t->col).w;
			guiCanvasDrawLine(w,
				pos - 2, height,
				pos + 2, height);
			guiCanvasDrawLine(w,
				pos, height, 
				pos, height + 
					guiTextFont.ascent);
			guiCanvasDrawLine(w,
				pos - 2, 
				height + guiTextFont.ascent,
				pos + 2, 
				height + guiTextFont.ascent);
		}
advance:
		height += (guiTextFont.ascent + guiTextFont.descent);
	}	
	if (t->repaintAll) {
		s = guiCanvasGetVisibleSize(t->c);
		guiCanvasSetFillColor(t->c, guiDefaultBackgroundColor);
		guiCanvasFillRectangle(
			t->c,
			0, height,
			s.w, s.h - height);	
	}
	if (repaintAll) {
		t->repaintAll = 0;
	}
	if (!t->active) {
		guiCanvasSetDim(w, 0);
	}
}

static void guiTextWarpToCursor(guiTextPtr t)
{
	int rh = guiTextFont.ascent + guiTextFont.descent;
	int x, y;
	guiRectangle r;
	guiTextScrollbarUpdate(t);
	x = guiFontTextSize(&guiTextFont, 
		t->text[t->row], t->col).w;
	y = t->row * rh;
	r.x = x - 2;
	r.y = y;
	r.w = 4;
	r.h = guiTextFont.ascent + guiTextFont.descent;
	if (r.x < 0) {
		r.x = 0;
	}
	if (r.y < 0) {
		r.y = 0;
	}
	guiCanvasScrollToRectangle(t->c, &r);
}

static void guiTextPixToRowCol(
	guiTextPtr t, int x, int y, int *rowP, int *colP);

static void guiTextMouseDown(
	guiCanvasPtr w, int x, int y, int button)
{
	guiTextPtr t = guiCanvasGetClientData(w);
	if (!guiCanvasQueryFocus(w)) {
		guiCanvasTakeFocus(w);
	}
	if (button == guiButtonLeft) {
		guiTextPixToRowCol(t, x, y, &t->row, &t->col);
		t->sRow = t->row;
		t->sCol = t->col;
		/* But it's not a selection until a drag occurs */
		if (t->selection) {
			t->selection = 0;
			t->repaintAll = 1;
		}
		guiTextWarpToCursor(t);
		guiTextRefresh(t);
	} else if (button == guiButtonMiddle) {
		if ((!t->readOnly) && (t->active)) {
			guiTextFetchSelection(t);
			if (t->updateHandler) {
				t->updateHandler(t, t->clientData);
			}
		} else {
			guiTextRefresh(t);
		}
	}
}

static void guiTextFetchSelection(guiTextPtr t)
{
	char *data;
	int len;
	if (t->selection) {
		t->selection = 0;
		t->repaintAll = 1;
	}
	data = guiGetClipboard(&len);
	if (data) {
		guiTextInsert(t, data, len);
		free(data);
	}
	guiTextWarpToCursor(t);
	guiTextRefresh(t);
}

static void guiTextPixToRowCol(
	guiTextPtr t, int x, int y, int *rowP, int *colP)
{
	int rowHeight = guiTextFont.ascent + guiTextFont.descent;
	int row = ((y-2) / rowHeight);
	int col;
	int width;
	int found = 0;
	if (row < 0) {
		row = 0;
	} else if (row >= t->rowsTotal) {
		row = t->rowsTotal - 1;
	} 
	if (row < 0) {
		row = 0;
	}
	for (col = 0; (col <= t->colsTotal[row]); col++) {
		width = 0;
		if (col) {
			width = guiFontTextSize(&guiTextFont, 
				t->text[row], col).w;
			if (width > x) {
				col--;
				found = 1;
				break;
			}
		}
	}	
	if (!found) {
		col = t->colsTotal[row];
	}
	if (row != t->row) {
		t->repaintAll = 1;
	}
	*rowP = row;
	*colP = col;
}

static void guiTextMouseDrag(
	guiCanvasPtr w, int x, int y, int button)
{
	guiTextPtr t = guiCanvasGetClientData(w);
	if (button != guiButtonLeft) {
		return;
	}
	guiTextPixToRowCol(t, x, y, &t->row, &t->col);
	/* Now a selection exists */
	t->selection = 1;
	guiTextWarpToCursor(t);
	guiTextRefresh(t);
}

static void guiTextMouseUp(
	guiCanvasPtr w, int x, int y, int button)
{
	guiTextPtr t = guiCanvasGetClientData(w);
	if (button != guiButtonLeft) {
		return;
	}
	/* Don't trash the old selection
		unless there is a new one */
	if ((t->col != t->sCol) || (t->row != t->sRow)) {
		guiTextStoreSelection(t);
	}
}

static void guiTextStoreSelection(guiTextPtr t)
{
	int r;
	int first = 1;
	int size = 0, pos = 0;
	char *data;
	/* Store the selection */
	/* First determine the size */
	for (r = 0; (r < t->rowsTotal); r++) {
		int left, right;
		if (guiTextRowInSelection(t, r, &left, &right)) {
			size += (right - left);				
			if (first) {
				first = 0;
			} else {
				/* One more for the CR */
				size++;	
			}
		}
	}
	/* Now do the actual storage */
	data = malloc(size);
	if (!data) {
		return;
	}
	first = 1;
	for (r = 0; (r < t->rowsTotal); r++) {
		int left, right;
		if (guiTextRowInSelection(t, r, &left, &right)) {
			if (first) {
				first = 0;
			} else {
				/* The CR */
				data[pos++] = '\n';	
			}
			memcpy(data + pos, t->text[r] + left, right - left);
			pos += (right - left);				
		}
	}		
	/* Send it */
	guiSetClipboard(data, size);
	/* Free it */
	free(data);	
}

enum {
	keyNone, keyMove, keyPaint
};

static int guiTextKeyDown(guiCanvasPtr w, 
	guiKey key, int shift, int control, int alt)
{
	guiTextPtr t = guiCanvasGetClientData(w);
	int toRow, toCol;
	char ch;
	int v = 0;
	int keyDraw = keyNone;	
	guiTextCursorHide(t);
	switch (key)
	{
		case guiKeyLeft:
		v = 1;
		keyDraw = keyMove;
		if (shift) {
			keyDraw = keyPaint;
			if (!(t->selection)) {
				t->sRow = t->row;
				t->sCol = t->col;
				t->selection = 1;
			}
		} else {
			if (t->selection) {
				t->selection = 0;
				t->repaintAll = 1;
				keyDraw = keyPaint;
			}
		}
		if (t->col) {
			t->col--;
		} else if (t->row) {
			t->row--;
			t->col = t->colsTotal[t->row];
			if (t->selection) {
				t->repaintAll = 1;
			}
		}
		if (shift) {
			guiTextStoreSelection(t);
		}
		break;
		case guiKeyRight:
		v = 1;
		keyDraw = keyMove;
		if (shift) {
			keyDraw = keyPaint;
			if (!(t->selection)) {
				t->sRow = t->row;
				t->sCol = t->col;
				t->selection = 1;
			}
		} else {
			if (t->selection) {
				keyDraw = keyPaint;
				t->selection = 0;
				t->repaintAll = 1;
			}
		}
		if (t->col < t->colsTotal[t->row]) {
			t->col++;
		} else if (t->row < (t->rowsTotal - 1)) {
			t->row++;
			t->col = 0;
			if (t->selection) {
				t->repaintAll = 1;
			}
		}
		if (shift) {
			guiTextStoreSelection(t);
		}
		break;
		case guiKeyUp:
		v = 1;
		keyDraw = keyMove;
		if (shift) {
			keyDraw = keyPaint;
			if (!(t->selection)) {
				t->sRow = t->row;
				t->sCol = t->col;
				t->selection = 1;
			}
		} else {
			if (t->selection) {
				keyDraw = keyPaint;
				t->selection = 0;
				t->repaintAll = 1;
			}
		}
		if (t->row) {
			t->row--;
			if (t->col > t->colsTotal[t->row]) {
				t->col = t->colsTotal[t->row];
			}
			if (t->selection) {
				t->repaintAll = 1;
			}
		}
		if (shift) {
			guiTextStoreSelection(t);
		}
		break;
		case guiKeyDown:
		v = 1;
		keyDraw = keyMove;
		if (shift) {
			keyDraw = keyPaint;
			if (!(t->selection)) {
				t->sRow = t->row;
				t->sCol = t->col;
				t->selection = 1;
			}
		} else {
			if (t->selection) {
				keyDraw = keyPaint;
				t->selection = 0;
				t->repaintAll = 1;
			}
		}
		if (t->row < (t->rowsTotal - 1)) {
			t->row++;
			if (t->col > t->colsTotal[t->row]) {
				t->col = t->colsTotal[t->row];
			}
			if (t->selection) {
				t->repaintAll = 1;
			}
		}
		if (shift) {
			guiTextStoreSelection(t);
		}
		break;
		case guiKeyPageUp:
		v = 1;
		keyDraw = keyMove;
		if (shift) {
			keyDraw = keyPaint;
			if (!(t->selection)) {
				t->sRow = t->row;
				t->sCol = t->col;
				t->selection = 1;
			}
		} else {
			if (t->selection) {
				keyDraw = keyPaint;
				t->selection = 0;
				t->repaintAll = 1;
			}
		}
		if (t->row) {
			t->row -= (guiCanvasGetPhysicalSize(w).h) / 
				(guiTextFont.ascent + guiTextFont.descent);
			if (t->row < 0) {
				t->row = 0;
			}
			if (t->col > t->colsTotal[t->row]) {
				t->col = t->colsTotal[t->row];
			}
			if (t->selection) {
				t->repaintAll = 1;
			}
		}
		if (shift) {
			guiTextStoreSelection(t);
		}
		break;
		case guiKeyPageDown:
		v = 1;
		keyDraw = keyMove;
		if (shift) {
			keyDraw = keyPaint;
			if (!(t->selection)) {
				t->sRow = t->row;
				t->sCol = t->col;
				t->selection = 1;
			}
		} else {
			if (t->selection) {
				keyDraw = keyPaint;
				t->selection = 0;
				t->repaintAll = 1;
			}
		}
		if (t->row < (t->rowsTotal - 1)) {
			t->row += (guiCanvasGetPhysicalSize(w).h) / 
				(guiTextFont.ascent + guiTextFont.descent);
			if (t->row > (t->rowsTotal - 1)) {
				t->row = t->rowsTotal - 1;
				if (t->row < 0) {
					t->row = 0;
				}
			}
			if (t->col > t->colsTotal[t->row]) {
				t->col = t->colsTotal[t->row];
			}
			if (t->selection) {
				t->repaintAll = 1;
			}
		}
		if (shift) {
			guiTextStoreSelection(t);
		}
		break;		
		case guiKeyBackspace:
		if (t->readOnly || (!t->active)) {
			break;
		}
		if (t->selection) {
			guiTextDeleteSelection(t);
			break;
		}
		if (t->col == t->colsTotal[t->row]) {
			/* Try to merge lines first */
			if (t->row < (t->rowsTotal-1)) {
				guiTextDelete(t, 
					t->row, t->col, t->row+1, 0);
			}
		}	
		/* Use the backoff function to make
			sure that we always go back by
			a "real" character, not just a
			line break. */
		if (guiTextBackOff(t, 1) != 1) {
			/* Can't do anything */
			break;
		}
		/* Deliberate fallthrough */
		case guiKeyDelete:
		if (t->readOnly) {
			break;
		}
		if (t->selection) {
			guiTextDeleteSelection(t);
			break;
		}
		toRow = t->row;
		toCol = t->col+1;
		if (toCol > t->colsTotal[t->row]) {
			if (t->row < (t->rowsTotal-1)) {
				toRow++;
				t->repaintAll = 1;
				toCol = 0;
			} else {
				/* Can't do anything */
				break;
			}	
		}
		guiTextDelete(t, t->row, t->col, toRow, toCol);
		break;
		case guiKeyInsert:
		if (t->readOnly || (!t->active)) {
			break;
		}
		guiTextFetchSelection(t);
		break;
		case guiKeyEnter:
		if (t->readOnly || (!t->active)) {
			break;
		}
		ch = '\n';
		guiTextInsert(t, &ch, 1);
		break;
		default:
		if (t->readOnly || (!t->active)) {
			break;
		}
		if (guiKeyIsPrintable(key)) {
			ch = key;
			guiTextDeleteSelection(t);
			guiTextInsert(t, &ch, 1);
		}	
		break;
	}
	guiTextWarpToCursor(t);
	if ((keyDraw == keyPaint) || (keyDraw == keyNone)) {
		guiTextRefresh(t);
	}
	if (t->updateHandler) {
		t->updateHandler(t, t->clientData);
	}
	guiTextCursorDraw(t);
	return v;
}

static int guiTextBackOff(guiTextPtr t, int chars)
{
	int charsB = 0;
	/* Back off by chars characters */
	do {
		int remove = t->col;
		if (remove > chars) {
			remove = chars;
		}
		t->col -= remove;
		chars -= remove;
		charsB += remove;
		if (chars) {
			t->row--;
			if (t->row < 0) {
				t->row = 0;
				t->col = 0;
				return charsB;
			}
			t->repaintAll = 1;
			t->col = t->colsTotal[t->row];	 
		}
	} while (chars);
	return charsB;
}

/* Determines the portion of the row, if any, that is
	inside the current selection; returns true and
	sets *left and *right if any portion is inside
	the current selection */ 

static int guiTextRowInSelection(guiTextPtr t, int row,
	int *left, int *right)
{
	int sRow, eRow, sCol, eCol;
	if (!(t->selection)) {
		return 0;
	}
	/* Sort the selection */
	guiTextSortSelection(t, &sRow, &sCol, &eRow, &eCol);
	if (row < sRow) {
		return 0;
	}
	if (row > eRow) {
		return 0;
	}
	if ((row > sRow) && (row < eRow)) {
		*left = 0;
		*right = t->colsTotal[row];
		return 1;
	} 
	*left = 0;
	*right = t->colsTotal[row];
	if (row == sRow) {
		*left = sCol;	
	}
	if (row == eRow) {
		*right = eCol;
	}	
	return 1;
}

static void guiTextSortSelection(guiTextPtr t,
	int *sRowP, int *sColP, int *eRowP, int *eColP)
{	
	int sRow = *sRowP, sCol = *sColP, 
		eRow = *eRowP, eCol = *eColP;
	if (t->sRow < t->row) {
		sRow = t->sRow;
		sCol = t->sCol;
		eRow = t->row;
		eCol = t->col;
	} else if (t->sRow > t->row) {
		sRow = t->row;
		sCol = t->col;
		eRow = t->sRow;
		eCol = t->sCol;
	} else { /* One line... */
		sRow = eRow = t->sRow;
		if (t->sCol < t->col) {
			sCol = t->sCol;
			eCol = t->col;
		} else {		
			sCol = t->col;
			eCol = t->sCol;
		}
	}
	*sRowP = sRow;
	*sColP = sCol;
	*eRowP = eRow;
	*eColP = eCol;
}

static void guiTextDeleteSelection(guiTextPtr t)
{
	int sRow, sCol, eRow, eCol;
	if (t->selection) {
		guiTextSortSelection(t, &sRow, &sCol,
			&eRow, &eCol);
		guiTextDelete(t, sRow, sCol,
			eRow, eCol); 
	}
}

static void guiTextScrollbarUpdate(guiTextPtr t)
{
	int i;
	int old = t->pixelWidest;
	int w = guiFontTextSize(
			&guiTextFont, t->text[t->row], t->colsTotal[t->row]).w;
	if (w >= t->pixelWidest) {
		t->pixelWidest = w;
		t->rowWidest = t->row;
	}
	if ((w < t->pixelWidest) || t->repaintAll) {
		t->pixelWidest = 0;
		for (i=0; (i < t->rowsTotal); i++) {
			int w = guiFontTextSize(
					&guiTextFont, 
					t->text[i], t->colsTotal[i]).w;
			if (w > t->pixelWidest) {
				t->pixelWidest = w;
				t->rowWidest = i;
			}
		}
	}
	if ((old != t->pixelWidest) || (t->rowsTotal != t->rowsTotalLast)) {
		int r = t->rowsTotal;
		guiCanvasSetVirtualSize(t->c, t->pixelWidest + 10, 
			r * (guiTextFont.ascent + guiTextFont.descent));
		t->rowsTotalLast = t->rowsTotal;
	}
}

static int textEventHandler(guiCanvasPtr c, guiEventPtr e)
{
	guiTextPtr t = (guiTextPtr) guiCanvasGetClientData(c);
        switch(e->t)
	{
		case guiEventMouseDown:
		guiTextMouseDown(c, e->l.x, e->l.y, e->button);
		break;
		case guiEventMouseDrag:
		guiTextMouseDrag(c, e->l.x, e->l.y, e->button);
		break;
		case guiEventMouseUp:
		guiTextMouseUp(c, e->l.x, e->l.y, e->button);
		break;
		case guiEventKeyDown:
		return guiTextKeyDown(c, e->key, e->shift, e->control, e->alt);
		break;
		case guiEventFocusIn:
		if (t->multiline) {
			guiCanvasGrabKey(c, guiKeyEnter, 0, 0, 0);
		}
		guiTextBlinkStart(c);
		break;
		case guiEventFocusOut:
		if (t->multiline) {
			guiCanvasReleaseKey(c, guiKeyEnter, 0, 0, 0);
		}
		guiTextBlinkStop(c);
		break;
		case guiEventTimer:
		guiTextBlink(c);
		break;
		case guiEventDestroyed:
		guiTextDestroyed(c);
		break;
		case guiEventRepaint:
		guiTextRepaint(c);
		break;
		case guiEventResize:
		guiTextWarpToCursor((guiTextPtr) guiCanvasGetClientData(c)); 
		break;
		default:
		break;
	}
	return 0;
}

void guiTextSetPhysicalSize(guiTextPtr t, int w, int h)
{
	guiCanvasSetPhysicalSize(t->c, w, h);
}

void guiTextShow(guiTextPtr t, int flag)
{
	guiCanvasShow(t->c, flag);
}

void guiTextMove(guiTextPtr t, int x, int y)
{
	guiCanvasMove(t->c, x, y);
}

guiSize guiTextGetPhysicalSize(guiTextPtr t)
{
	return guiCanvasGetPhysicalSize(t->c);
}

static void guiTextBlinkStart(guiCanvasPtr c)
{
	guiTextPtr t = (guiTextPtr) guiCanvasGetClientData(c);	
	t->blinkOn = 1;
	/* Set blink timer for the first time */
	guiCanvasSetTimer(c, 0, 500);	
}

static void guiTextBlinkStop(guiCanvasPtr c)
{
	guiTextPtr t = (guiTextPtr) guiCanvasGetClientData(c);	
	/* Clear blink timer */
	guiCanvasClearTimer(c, 0);
	if (!t->blinkOn) {
		/* Make sure the cursor winds up on */
		t->blinkOn = 1;
		guiTextRefresh(t);
	}
}

static void guiTextCursorHide(guiTextPtr t)
{
	guiTextCursorDrawBody(t, 0);
}

static void guiTextCursorDraw(guiTextPtr t)
{
	guiTextCursorDrawBody(t, t->blinkOn);
}

static void guiTextCursorDrawBody(guiTextPtr t, int on)
{
	/* Getting this exactly right is much more painful than you'd think */
	guiColor fore, back;
	int from;
	int left, right;
	int in;
	int len;
	int x, xt, cw = 0;
	int y = t->row * (guiTextFont.ascent + guiTextFont.descent);
	guiCanvasPtr w = (guiCanvasPtr) t->c;
	from = t->col - 1;
	len = 2;
	if (from < 0) {
		from = 0;
		len--;
	}
	if ((from + len) >= t->colsTotal[t->row]) {
		len--;
	}	
	x = guiFontTextSize(&guiTextFont, t->text[t->row], t->col).w;
	xt = guiFontTextSize(&guiTextFont, t->text[t->row], from).w;
	guiCanvasSetDrawColor(w, guiTextColor);
	in = guiTextRowInSelection(t, t->row,
		&left, &right);
	if (in && ((left <= from) && (right > from))) { 
		fore = guiDefaultBackgroundColor;
		back = guiTextColor;
	} else {
		back = guiDefaultBackgroundColor;
		fore = guiTextColor;
	}
	guiCanvasSetDrawColor(w, fore);
	guiCanvasSetTextBackgroundColor(w, back);
	if (len) {
		guiCanvasDrawOpaqueText(w, xt, y + guiTextFont.ascent,
				t->text[t->row] + from, 1);
	}
	if (!on) {
		guiCanvasSetDrawColor(w, back);
	}
	guiCanvasDrawLine(w,
		x - 2, y,
		x - 1, y);
	guiCanvasDrawLine(w,
		x - 2, 
		y + guiTextFont.ascent,
		x - 1, 
		y + guiTextFont.ascent);	
	if (len) {
		guiCanvasSetDrawColor(w, fore);
		guiCanvasDrawText(w, xt, y + guiTextFont.ascent,
				t->text[t->row] + from, 1);
	}
	if (in && ((left <= (from + 1)) && (right > (from + 1)))) { 
		fore = guiDefaultBackgroundColor;
		back = guiTextColor;
	} else {
		back = guiDefaultBackgroundColor;
		fore = guiTextColor;
	}
	guiCanvasSetDrawColor(w, fore);
	guiCanvasSetTextBackgroundColor(w, back);
	if (len > 1) {
		cw = guiFontTextSize(&guiTextFont,
				t->text[t->row] + from, 1).w;
		guiCanvasDrawOpaqueText(w, xt + cw,
			y + guiTextFont.ascent,
			t->text[t->row] + from + 1, 1);	
	}
	if (!on) {
		guiCanvasSetDrawColor(w, back);
	}
	guiCanvasDrawLine(w,
		x, y, 
		x, y + guiTextFont.ascent);
	guiCanvasDrawLine(w,
		x, y,
		x + 2, y);
	guiCanvasDrawLine(w,
		x, 
		y + guiTextFont.ascent,
		x + 2, 
		y + guiTextFont.ascent);
	if (len > 1) {
		guiCanvasSetDrawColor(w, fore);
		guiCanvasDrawText(w, xt + cw,
			y + guiTextFont.ascent,
			t->text[t->row] + from + 1, 1);	
	}
}

static void guiTextBlink(guiCanvasPtr c)
{
	guiTextPtr t = (guiTextPtr) guiCanvasGetClientData(c);	
	t->blinkOn = !t->blinkOn;
	guiTextCursorDraw(t);
	/* Reset blink timer */
	guiCanvasSetTimer(c, 0, 500);	
}

void guiTextSetActive(guiTextPtr t, int flag)
{
	t->active = flag;
	if (!flag) {
		guiTextBlinkStop((guiCanvasPtr) t->c);
	} 
	guiCanvasWantFocus((guiCanvasPtr) t->c, flag);
	guiCanvasInvalidate(t->c, 0);
}

void guiTextTakeFocus(guiTextPtr t)
{
	guiCanvasTakeFocus((guiCanvasPtr) t->c);
}

int guiTextGetActive(guiTextPtr t)
{
	return t->active;
}

void guiTextSetUpdateHandler(guiTextPtr t, 
	void (*updateHandler) (guiTextPtr t, void *clientData),
	void *clientData)
{
	t->updateHandler = updateHandler;	
	t->clientData = clientData;
}

int guiTextGetInt(guiTextPtr t)
{
	char s[32];
	guiTextGetBuffer(t, s, 32, 0);
	return atoi(s);
}

double guiTextGetDouble(guiTextPtr t)
{
	char s[64];
	guiTextGetBuffer(t, s, 64, 0);
	return atof(s);
}

