#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include "jpeglib.h"
#include "png.h"
#include <windows.h>
#include "checkmem.h"
#include "loadimg.h"
#include "gd.h"
#ifdef WIN32
#undef FAR
#endif /* WIN32 */
#include "dither.h"

int gifRowDone(int rowsDone, int rowsTotal);

int LoadGif(FILE *in, HWND context, HBITMAP *hbmP, HPALETTE *hpalP, SIZE *sizeP) {
	LPBITMAPINFO dib = 0;
	HBITMAP hbm = 0;
    HPALETTE hpal = 0;
    gdImagePtr im;
	int x, y, pad;
	LPBITMAPINFOHEADER dibh;
	int i;
	HDC hdc;  
	LOGPALETTE *plgpl = 0;
	int dibStructSize;
	static PALETTEENTRY sysp[16];
	static PALETTEENTRY sourcep[256];
	static unsigned int syspCube[4][4][4];
	int useSysp = 0;
	BYTE *dibBits = 0;
	BYTE *dibBitsC;
    HPALETTE hpalT;

	hdc = WrapGetDC(context);
	im = gdImageCreateFromGifRowhook(in, gifRowDone);
	if (!im) { 
		MessageBox(NULL, "Not a valid GIF file!\r\n"\
			"This image file is corrupted,\r\n"\
			"was downloaded using FTP in ascii mode, or\r\n"\
			"does not contain a GIF-format image.", 
			"GIF Error", 
			MB_ICONEXCLAMATION);
		WrapReleaseDC(context, hdc);
		return 0;
	}
	for (i=0; (i < im->colorsTotal); i++) {
		sourcep[i].peRed = im->red[i];
		sourcep[i].peGreen = im->green[i];
		sourcep[i].peBlue = im->blue[i];
	}	
	if (GetDeviceCaps(hdc, BITSPIXEL) *
		GetDeviceCaps(hdc, PLANES) == 4) {
		int r, g, b;
		useSysp = 1;
		GetSystemPaletteEntries(hdc, 0, 16, sysp);
		BuildSyspCube(sysp, syspCube);
	}
	
	dibBits = (BYTE *) WrapMalloc(im->sx * 3 + 4);
    if (!dibBits) {
		MessageBox(NULL,
			"Out of Memory!",
			"GIF Decoding Error",
			MB_ICONEXCLAMATION);
		WrapFree(dib);
		gdImageDestroy(im);
		WrapReleaseDC(context, hdc);
		return 0;
	}    
	if ((!useSysp) && 
		(GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE)) {
		plgpl = (LOGPALETTE*)
			WrapMalloc(sizeof(LOGPALETTE) + 
				im->colorsTotal * sizeof(PALETTEENTRY));
		if (plgpl) {
			plgpl->palNumEntries = im->colorsTotal;
			plgpl->palVersion = 0x300;
			for (i=0; (i<im->colorsTotal); i++) {
				plgpl->palPalEntry[i].peRed = im->red[i];
				plgpl->palPalEntry[i].peGreen = im->green[i];
				plgpl->palPalEntry[i].peBlue = im->blue[i];
				plgpl->palPalEntry[i].peFlags = 0;
			}
		    hpal = CreatePalette(plgpl);
			WrapFree(plgpl);
			if (hpal) {
			    hpalT = SelectPalette(hdc, hpal, 0);
				RealizePalette(hdc);
			}
		}
	}
	dibStructSize = sizeof(BITMAPINFOHEADER); 
	dib = (LPBITMAPINFO)
		WrapMalloc(dibStructSize + (sizeof(RGBQUAD) * 256));
	if (!dib) {
		MessageBox(NULL,
			"Out of Memory!",
			"GIF Decoding Error",
			MB_ICONEXCLAMATION);
		gdImageDestroy(im);
		if (hpal) {
		    SelectPalette (hdc, hpalT, FALSE);
			RealizePalette (hdc);
			DeleteObject (hpal);
		}
		WrapFree(dibBits);
		WrapReleaseDC(context, hdc);
		return 0;
	}
	dibh = (LPBITMAPINFOHEADER) dib;
	if (hpal) {
		/* Coerce the pointer */
		WORD *index = (WORD *) &(dib->bmiColors);
        /* Build the index */
        int i;
        for (i=0; (i < im->colorsTotal); i++) {
        	index[i] = i;
    	}
	}
	dibh->biSize = (long) dibStructSize;
	dibh->biWidth = im->sx;
	dibh->biHeight = im->sy;
	dibh->biPlanes = 1;
	dibh->biBitCount = hpal ? 8 : 24;
	dibh->biCompression = 0;
	dibh->biXPelsPerMeter = 1000;
	dibh->biYPelsPerMeter = 1000;
	dibh->biClrUsed = 0;
	dibh->biClrImportant = 0;	
	hbm = CreateDIBitmap(hdc,
		dibh, 0L, NULL, NULL, 0);
	pad = 4 - ((im->sx * dibh->biBitCount / 8) & 3);
	if (pad == 4) {
		pad = 0;
	}
	for (y = 0; (y < im->sy); y++) {
	    dibBitsC = dibBits;
		if (hpal) {
			DitherRowPalettePalette(dibBits, im->pixels[y],
				im->sx);
		} else if (useSysp) {
			DitherRowSyspPalette(
				dibBits, im->pixels[y], im->sx, sourcep, 
				syspCube, sysp);
		} else {
			DitherRowTruecolorPalette(dibBits, im->pixels[y],
				im->sx, sourcep);
		}			
		SetDIBits(hdc, hbm, im->sy - y - 1,
			1, dibBits, dib, 
			hpal ? DIB_PAL_COLORS : DIB_RGB_COLORS);
		if (!rowDone(y + im->sy, im->sy * 2)) {
			/* All bets are off */
			WrapFree(dib);
			WrapFree(dibBits);
			dibBits = 0;
			gdImageDestroy(im);
			if (hpal) {
			    SelectPalette (hdc, hpalT, FALSE);
				RealizePalette(hdc);
				DeleteObject (hpal);
			}
			WrapReleaseDC(context, hdc);
			return 0;
		}
	}
    if (hpal) {
    	SelectPalette (hdc, hpalT, FALSE);
	}
	WrapReleaseDC(context, hdc);
	sizeP->cx = im->sx;
	sizeP->cy = im->sy;
	gdImageDestroy(im);
	WrapFree(dib);
	WrapFree(dibBits);
	*hbmP = hbm;
	*hpalP = hpal;
	return 1;
}

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

METHODDEF void
my_error_exit (j_common_ptr cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  (*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}

METHODDEF void
my_output_message (j_common_ptr cinfo)
{
  char buffer[JMSG_LENGTH_MAX];

  /* Create the message */
  (*cinfo->err->format_message) (cinfo, buffer);

  /* Send it to a msgbox */
  MessageBox(NULL, buffer, "JPEG Error", 
  	MB_ICONEXCLAMATION);
}

int LoadJpeg(FILE *in, HWND context, HBITMAP *hbmP, HPALETTE *hpalP, SIZE *sizeP) {
	LPBITMAPINFO dib = 0;
	HBITMAP hbm = 0;
    HPALETTE hpal = 0;
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	int rowsDone = 0;
	JSAMPLE *scanline;
	int pad;
	LPBITMAPINFOHEADER dibh;
	int i;
	BYTE *dibBits = 0;
	BYTE *dibBitsC;
	int dibStructSize;
    long rowBytes;
	LOGPALETTE *plgpl = 0;
	HDC hdc;
    HPALETTE hpalT;
	static PALETTEENTRY sysp[16];
	int useSysp = 0;
	static unsigned int syspCube[4][4][4];
	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.output_message = my_output_message;
	jerr.pub.error_exit = my_error_exit;
	hdc = WrapGetDC(context);
	/* Establish the setjmp return context for my_error_exit to use. */
	if (setjmp(jerr.setjmp_buffer)) {
	    /* If we get here, the JPEG code has signaled an error.
	     * We need to clean up the JPEG object, close the input file, and return.
	     */
	    jpeg_destroy_decompress(&cinfo);
		if (hpal) {
		    SelectPalette(hdc, hpalT, 0);
			RealizePalette(hdc);
			DeleteObject(hpal);
		}
		if (dib) {
			WrapFree(dib);
		}
		if (scanline) {
			WrapFree(scanline);
		}
		if (dibBits) {
			WrapFree(dibBits);
		}		
		WrapReleaseDC(context, hdc);
	    return 0;
	}
	if (GetDeviceCaps(hdc, BITSPIXEL) *
		GetDeviceCaps(hdc, PLANES) == 4) {
		useSysp = 1;
		GetSystemPaletteEntries(hdc, 0, 16, sysp);
		BuildSyspCube(sysp, syspCube);
	}
	if (hpal) {
		DeleteObject(hpal);
	}
	hpal = 0;

	if ((!useSysp) && 
		(GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE)) {
		PALETTEENTRY *p;
		plgpl = (LOGPALETTE*)
			WrapMalloc(sizeof(LOGPALETTE) + 
				125 * sizeof(PALETTEENTRY));
		if (plgpl) {
			plgpl->palNumEntries = 125;
			plgpl->palVersion = 0x300;
			i = 0;
			p = plgpl->palPalEntry;
			Build555Cube(p);
		    hpal = CreatePalette(plgpl);
			WrapFree(plgpl);
			if (hpal) {
			    hpalT = SelectPalette(hdc, hpal, 0);
				RealizePalette(hdc);
			}
		}
	}

	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, in);
	jpeg_read_header(&cinfo, 1);
	jpeg_start_decompress(&cinfo);
	
	scanline = (JSAMPLE *) WrapMalloc(
		cinfo.output_width * cinfo.output_components *
		sizeof(JSAMPLE));

	if (!scanline) {
		MessageBox(NULL,
			"Out of Memory!",
			"JPEG Decoding Error",
			MB_ICONEXCLAMATION);
		jpeg_destroy_decompress(&cinfo);
		if (hpal) {
		    SelectPalette(hdc, hpalT, 0);
			RealizePalette(hdc);
			DeleteObject(hpal);
		}
		WrapReleaseDC(context, hdc);
		return 0;
	}    
	dibStructSize = sizeof(BITMAPINFOHEADER); 
	dib = (LPBITMAPINFO)
		WrapMalloc(dibStructSize + (sizeof(RGBQUAD) * 256));
	if (!dib) {
		MessageBox(NULL,
			"Out of Memory!",
			"JPEG Decoding Error",
			MB_ICONEXCLAMATION);
		WrapFree(scanline);
		jpeg_destroy_decompress(&cinfo);
		if (hpal) {
		    SelectPalette(hdc, hpalT, 0);
			RealizePalette(hdc);
			DeleteObject(hpal);
		}
		WrapReleaseDC(context, hdc);
		return 0;
	}
	dibh = (LPBITMAPINFOHEADER) dib;
	
	if (hpal) {
		pad = 4 - ((cinfo.output_width * 3) & 3);
	} else {
		pad = 4 - (cinfo.output_width & 3);
	}
	if (pad == 4) {
		pad = 0;
	}
	dibh->biSize = (long) dibStructSize;
	dibh->biWidth = cinfo.output_width;
	dibh->biHeight = cinfo.output_height;
	dibh->biPlanes = 1;
	if (hpal) {
		dibh->biBitCount = 8;
	} else {
		dibh->biBitCount = 24;
	}
	dibh->biCompression = 0;
	dibh->biXPelsPerMeter = 1000;
	dibh->biYPelsPerMeter = 1000;
	dibh->biClrUsed = 0;
	dibh->biClrImportant = 0;	

	if (hpal) {
		/* Coerce the pointer */
		WORD *index = (WORD *) &(dib->bmiColors);
        /* Build the index */
        int i;
        for (i=0; (i < 125); i++) {
        	index[i] = i;
    	}
    }
	dibBits = (BYTE  *) WrapMalloc(cinfo.output_width * 3 + 4);
    if (!dibBits) {
		MessageBox(NULL,
			"Out of Memory!",
			"JPEG Decoding Error",
			MB_ICONEXCLAMATION);
		WrapFree(dib);
		WrapFree(scanline);
		jpeg_destroy_decompress(&cinfo);
		dib = 0;
		if (hpal) {
		    SelectPalette(hdc, hpalT, 0);
			RealizePalette(hdc);
			DeleteObject(hpal);
		}
		WrapReleaseDC(context, hdc);
		return 0;
	}     
    rowBytes = (long) cinfo.output_width * 3 + pad;

	hbm = CreateDIBitmap(hdc, dibh, 0L,	NULL, NULL, 0);

    while (jpeg_read_scanlines(&cinfo, &scanline, 1) > 0) {
		int i = cinfo.output_width;
		unsigned char *scanlineC = scanline;
		dibBitsC = dibBits;
		if ((!hpal) && (!useSysp)) {
			/* No palette -- keep it 24 bit */
			if (cinfo.output_components == 3) {
				DitherRowTruecolorRGB(dibBits, scanlineC, i);
			} else {
				DitherRowTruecolorGray(dibBits, scanlineC, i);
			}
		} else {
		    /* Palette in use -- dither */
			if (!useSysp) {
				/* Color Cube */
				if (cinfo.output_components == 3) {
					DitherRow555RGB(dibBits, scanlineC, i);
				} else {
					DitherRow555Gray(dibBits, scanlineC, i); 
				}
			} else {
				if (cinfo.output_components == 3) {
					DitherRowSyspRGB(dibBits, scanlineC, i,
						syspCube, sysp);
				} else {
					DitherRowSyspGray(dibBits, scanlineC, i,
						syspCube, sysp);
				}
			}
		}		    
		SetDIBits(hdc, hbm, cinfo.output_height - rowsDone - 1,
			1, dibBits, dib, hpal ? DIB_PAL_COLORS : DIB_RGB_COLORS);
		rowsDone++;
        if (!rowDone(rowsDone, cinfo.output_height)) {
			WrapFree(dib);
			WrapFree(dibBits);
			WrapFree(scanline);
			jpeg_destroy_decompress(&cinfo);
			if (hpal) {
			    SelectPalette(hdc, hpalT, 0);
				RealizePalette(hdc);
				DeleteObject(hpal);
			}
			WrapReleaseDC(context, hdc);
        	return 0;
        }
   		if ((long) rowsDone == (long) cinfo.output_height) {
			break;
		}
	}	
	sizeP->cx = cinfo.output_width;
	sizeP->cy = cinfo.output_height;
	jpeg_finish_decompress(&cinfo);		
	jpeg_destroy_decompress(&cinfo);
    if (hpal) {
    	SelectPalette (hdc, hpalT, FALSE);
	}
	WrapReleaseDC(context, hdc);
	WrapFree(dibBits);
	WrapFree(scanline);
	WrapFree(dib);
	*hbmP = hbm;
	*hpalP = hpal;
	return 1;
}

extern void PngMessage(char *msg);

void PngMessage(char *msg)
{
	MessageBox(NULL, msg, "PNG Error",
		MB_ICONEXCLAMATION);
}

int LoadPng(FILE *in, HWND context, HBITMAP *hbmP, HPALETTE *hpalP, SIZE *sizeP)
{
	LPBITMAPINFO dib = 0;
    HPALETTE hpal = 0;
	HBITMAP hbm = 0;
	unsigned long rowsDone = 0;
	unsigned long pad;
	LPBITMAPINFOHEADER dibh;
	unsigned long i;
	BYTE **scanlines;
	BYTE *dibBits = 0;
	BYTE *dibBitsC;
	int dibStructSize;
    unsigned long rowBytes;
	int entriesTotal;
	LOGPALETTE *plgpl = 0;
	HDC hdc;
    HPALETTE hpalT;
	int useCube = 0;      
    png_struct *png_ptr;
    png_info *info_ptr;
    float screen_gamma = 2.0;
	int number_passes;
	int pass; 
    png_color_16 my_background;
	static PALETTEENTRY sysp[16];
	static unsigned int syspCube[4][4][4];
	int useSysp = 0;
	static PALETTEENTRY sourcep[256];

   hdc = WrapGetDC(context);
   /* allocate the necessary structures */
   png_ptr = (png_struct *) WrapMalloc(sizeof (png_struct));
   if (!png_ptr)
   {          
      WrapReleaseDC(context, hdc);
		MessageBox(NULL,
			"Out of Memory!",
			"PNG Decoding Error",
			MB_ICONEXCLAMATION);
      return 0;
   }

   info_ptr = (png_info *) WrapMalloc(sizeof (png_info));
   if (!info_ptr)
   {
      WrapFree(png_ptr);
      WrapReleaseDC(context, hdc);
		MessageBox(NULL,
			"Out of Memory!",
			"PNG Decoding Error",
			MB_ICONEXCLAMATION);
      return 0;
   }

   /* set error handling */
   if (setjmp(png_ptr->jmpbuf))
   {
      png_read_destroy(png_ptr, info_ptr, (png_info *)0);
      WrapFree(png_ptr);
      WrapFree(info_ptr);
      WrapReleaseDC(context, hdc);
      /* If we get here, we had a problem reading the file */
      return 0;
   }


   /* initialize the structures, info first for error handling */
   png_info_init(info_ptr);
   png_read_init(png_ptr);

   /* set up the input control */
   png_init_io(png_ptr, in);

   /* read the file information */
   png_read_info(png_ptr, info_ptr);
   
   scanlines = (BYTE **) WrapMalloc(sizeof(BYTE *) * (long)info_ptr->height);
   if (!scanlines) {
      png_read_destroy(png_ptr, info_ptr, (png_info *)0);
      WrapFree(png_ptr);
      WrapFree(info_ptr);
      WrapReleaseDC(context, hdc);
		MessageBox(NULL,
			"Out of Memory!",
			"PNG Decoding Error",
			MB_ICONEXCLAMATION);
      return 0;
   }
   for (i=0; (i < info_ptr->height); i++) {
      if ((info_ptr->color_type == PNG_COLOR_TYPE_PALETTE) ||
          (info_ptr->color_type == PNG_COLOR_TYPE_GRAY) ||
          (info_ptr->color_type == PNG_COLOR_TYPE_GRAY_ALPHA)) {
      	scanlines[i] = (BYTE *) WrapMalloc(info_ptr->width + 4);
      } else {
      	scanlines[i] = (BYTE *) WrapMalloc(info_ptr->width * 3 + 4);
      }
      if (!scanlines[i]) {
      	int j;
      	for (j = 0; (j < i); j++) {
      		WrapFree(scanlines[j]);
      	}
      	WrapFree(scanlines);
        png_read_destroy(png_ptr, info_ptr, (png_info *)0);
        WrapFree(png_ptr);
        WrapFree(info_ptr);
        WrapReleaseDC(context, hdc);
		MessageBox(NULL,
			"Out of Memory!",
			"PNG Decoding Error",
			MB_ICONEXCLAMATION);
        return 0;
      }      	
   }  

   /* set up the transformations you want.  Note that these are
      all optional.  Only call them if you want them */

   /* expand grayscale images to the full 8 bits */
   if (info_ptr->color_type == PNG_COLOR_TYPE_GRAY &&
      info_ptr->bit_depth < 8)
      png_set_expand(png_ptr);

   /* expand images with transparency to full alpha channels */
   if (info_ptr->valid & PNG_INFO_tRNS)
      png_set_expand(png_ptr);

   /* Set the background color to draw transparent and alpha
      images over */
   my_background.red = my_background.green = my_background.blue = 0xFFFF;

   if (info_ptr->valid & PNG_INFO_bKGD)
      png_set_background(png_ptr, &(info_ptr->background),
         PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
   else
      png_set_background(png_ptr, &my_background,
         PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);

   /* tell libpng to handle the gamma conversion for you */
   if (info_ptr->valid & PNG_INFO_gAMA)
      png_set_gamma(png_ptr, screen_gamma, info_ptr->gamma);
   else
      png_set_gamma(png_ptr, screen_gamma, 0.45);

   /* tell libpng to strip 16 bit depth files down to 8 bits */
   if (info_ptr->bit_depth == 16)
      png_set_strip_16(png_ptr);

   /* pack pixels into bytes */
   if (info_ptr->bit_depth < 8)
      png_set_packing(png_ptr);

   /* turn on interlace handling if you are not using png_read_image() */
   if (info_ptr->interlace_type)
      number_passes = png_set_interlace_handling(png_ptr);
   else
      number_passes = 1;

   /* optional call to update palette with transformations */
   png_start_read_image(png_ptr);

   /* optional call to update the info structure */
   png_read_update_info(png_ptr, info_ptr);

	if (GetDeviceCaps(hdc, BITSPIXEL) *
		GetDeviceCaps(hdc, PLANES) == 4) {
		useSysp = 1;
		GetSystemPaletteEntries(hdc, 0, 16, sysp);
		BuildSyspCube(sysp, syspCube);
	}
	if (info_ptr->color_type == PNG_COLOR_TYPE_PALETTE) {
		for (i=0; (i < info_ptr -> num_palette); i++) {
			sourcep[i].peRed = info_ptr -> palette[i].red;
			sourcep[i].peGreen = info_ptr -> palette[i].green;
			sourcep[i].peBlue = info_ptr -> palette[i].blue;
		}	
	}
	if ((!useSysp) && (
			GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE)) {
		PALETTEENTRY *p;
		if (info_ptr->color_type == PNG_COLOR_TYPE_PALETTE) {
			entriesTotal = info_ptr->num_palette;
			useCube = 0;
		} else {
			entriesTotal = 125;
			useCube = 1;
		}
			
		plgpl = (LOGPALETTE*)
			WrapMalloc(sizeof(LOGPALETTE) + 
				entriesTotal * sizeof(PALETTEENTRY));
		if (plgpl) {
			plgpl->palNumEntries = entriesTotal;
			plgpl->palVersion = 0x300;
			i = 0;
			p = plgpl->palPalEntry;
			if (useCube) {
				Build555Cube(p);
			} else {
				for (i=0; (i < info_ptr->num_palette); i++) {
					p->peRed = info_ptr->palette[i].red;
					p->peGreen = info_ptr->palette[i].green;
					p->peBlue = info_ptr->palette[i].blue;
					p->peFlags = 0;
					p++;
				}
			}				
		    hpal = CreatePalette(plgpl);
			WrapFree(plgpl);
			if (hpal) {
			    hpalT = SelectPalette(hdc, hpal, 0);
				RealizePalette(hdc);
			}
		}
	}
	dibStructSize = sizeof(BITMAPINFOHEADER); 
	dib = (LPBITMAPINFO)
		WrapMalloc(dibStructSize + (sizeof(RGBQUAD) * 256));
	if (!dib) {
		int i;
		MessageBox(NULL,
			"Out of Memory!",
			"PNG Decoding Error",
			MB_ICONEXCLAMATION);
		for (i=0; (i < info_ptr->height); i++) {
			WrapFree(scanlines[i]);
		}
		WrapFree(scanlines);
        png_read_destroy(png_ptr, info_ptr, (png_info *)0);
        WrapFree(png_ptr);
        WrapFree(info_ptr);
		if (hpal) {
		    SelectPalette(hdc, hpalT, 0);
			RealizePalette(hdc);
			DeleteObject(hpal);
		}
		WrapReleaseDC(context, hdc);
		return 0;
	}
	dibh = (LPBITMAPINFOHEADER) dib;
	
	if (hpal) {
		pad = 4 - ((info_ptr -> width * 3) & 3);
	} else {
		pad = 4 - (info_ptr -> width & 3);
	}
	if (pad == 4) {
		pad = 0;
	}
	dibh->biSize = (long) dibStructSize;
	dibh->biWidth = info_ptr -> width;
	dibh->biHeight = info_ptr -> height;
	dibh->biPlanes = 1;
	if (hpal) {
		dibh->biBitCount = 8;
	} else {
		dibh->biBitCount = 24;
	}
	dibh->biCompression = 0;
	dibh->biXPelsPerMeter = 1000;
	dibh->biYPelsPerMeter = 1000;
	dibh->biClrUsed = 0;
	dibh->biClrImportant = 0;	

	if (hpal) {
		/* Coerce the pointer */
		WORD *index = (WORD *) &(dib->bmiColors);
        /* Build the index */
        int i;
        for (i=0; (i < entriesTotal); i++) {
        	index[i] = i;
    	}
    }
	dibBits = (BYTE *) WrapMalloc((unsigned int) (info_ptr->width * 3 + pad));
		
    if (!dibBits) {
		MessageBox(NULL,
			"Out of Memory!",
			"PNG Decoding Error",
			MB_ICONEXCLAMATION);
		WrapFree(dib);
        png_read_destroy(png_ptr, info_ptr, (png_info *)0);
        WrapFree(png_ptr);
        WrapFree(info_ptr);
		if (hpal) {
		    SelectPalette(hdc, hpalT, 0);
			RealizePalette(hdc);
			DeleteObject(hpal);
		}
		WrapReleaseDC(context, hdc);
		for (i=0; (i < info_ptr->height); i++) {
			WrapFree(scanlines[i]);
		}
		WrapFree(scanlines);
		return 0;
	}     
    rowBytes = (long) info_ptr->width * 3 + pad;

	hbm = CreateDIBitmap(hdc, dibh, 0L,	NULL, NULL, 0);

   for (pass = 0; pass < number_passes; pass++)
   {
      unsigned long y;
      rowsDone = 0;
      for (y = 0; y < info_ptr->height; y++)
      {
		unsigned long w = info_ptr->width;
		unsigned char *scanlineC = scanlines[y];
        png_read_rows(png_ptr, &scanlineC, NULL, 1);
        if (!rowDone(pass * info_ptr->height + rowsDone, 
        	info_ptr->height * number_passes)) {
			WrapFree(dib);
	        png_read_destroy(png_ptr, info_ptr, (png_info *)0);
	        WrapFree(png_ptr);
	        WrapFree(info_ptr);
			if (hpal) {
			    SelectPalette(hdc, hpalT, 0);
				RealizePalette(hdc);
				DeleteObject(hpal);
			}
			WrapReleaseDC(context, hdc);
			for (i=0; (i < info_ptr->height); i++) {
				WrapFree(scanlines[i]);
			}
			WrapFree(scanlines);
			WrapFree(dibBits);
			return 0;
        }
		/* No need to examine the output of incomplete passes */
		if (pass != (number_passes - 1)) {
			continue;
		}
		dibBitsC = dibBits;
		if ((!hpal) && (!useSysp))  {
			/* No palette -- keep it 24 bit */
			if ((info_ptr -> color_type == PNG_COLOR_TYPE_RGB) ||
				(info_ptr -> color_type == PNG_COLOR_TYPE_RGB_ALPHA))
			{
				DitherRowTruecolorRGB(dibBitsC, scanlineC, w);
			} else if (info_ptr -> color_type == PNG_COLOR_TYPE_PALETTE) {
				/* Expand palette since we're doing true 24-bit. */
				DitherRowTruecolorPalette(dibBits, scanlineC,
					w, sourcep);
			} else if ((info_ptr -> color_type == PNG_COLOR_TYPE_GRAY) ||
				(info_ptr -> color_type == PNG_COLOR_TYPE_GRAY_ALPHA)) {
				DitherRowTruecolorGray(dibBits, scanlineC, w);
			}
		} else {
			if (!useSysp) {
				if ((info_ptr -> color_type == PNG_COLOR_TYPE_RGB) ||
					(info_ptr -> color_type == PNG_COLOR_TYPE_RGB_ALPHA))
				{   
					DitherRow555RGB(dibBitsC, scanlineC, w);
				} else if (info_ptr -> color_type == PNG_COLOR_TYPE_PALETTE) {
					DitherRowPalettePalette(dibBitsC, scanlineC, w);
				} else if ((info_ptr -> color_type == PNG_COLOR_TYPE_GRAY) ||
					(info_ptr -> color_type == PNG_COLOR_TYPE_GRAY_ALPHA))
				{
					DitherRow555Gray(dibBitsC, scanlineC, w);
				}
			} else {
				if ((info_ptr -> color_type == PNG_COLOR_TYPE_RGB) ||
					(info_ptr -> color_type == PNG_COLOR_TYPE_RGB_ALPHA)) {
					DitherRowSyspRGB(dibBitsC, scanlineC, w,
						syspCube, sysp);
				} else if (info_ptr -> color_type == PNG_COLOR_TYPE_PALETTE) {
					DitherRowSyspPalette(dibBitsC, scanlineC, w,
						sourcep, syspCube, sysp);
				} else if ((info_ptr -> color_type == PNG_COLOR_TYPE_GRAY) ||
					(info_ptr -> color_type == PNG_COLOR_TYPE_GRAY_ALPHA)) {
					DitherRowSyspGray(dibBitsC, scanlineC, w,
						syspCube, sysp); 
				}
			}
		}		    
		SetDIBits(hdc, hbm, info_ptr->height - rowsDone - 1,
			1, dibBits, dib, hpal ? DIB_PAL_COLORS : DIB_RGB_COLORS);
        rowsDone++;
      }
   }

   /* read the rest of the file, getting any additional chunks
      in info_ptr */
   png_read_end(png_ptr, info_ptr);

   /* Free the image buffer */
   for (i=0; (i < info_ptr -> height); i++) {
     WrapFree(scanlines[i]);
   }
   WrapFree(scanlines);   

   sizeP->cx = info_ptr -> width;
   sizeP->cy = info_ptr -> height;
   
   /* clean up after the read, and free any memory allocated */
   png_read_destroy(png_ptr, info_ptr, (png_info *)0);

   /* free the structures */
   WrapFree(png_ptr);
   WrapFree(info_ptr);
   WrapFree(dibBits);   
   WrapFree(dib);
   if (hpal) {
      SelectPalette(hdc, hpalT, 0);
   }
   /* that's it */
   WrapReleaseDC(context, hdc);
   *hpalP = hpal;
   *hbmP = hbm;
   return 1;
}

int gifRowDone(int rowsDone, int rowsTotal) {
	/* Fake twice as many rows so the "second 50%"
		can be the rendering pass */
	return rowDone(rowsDone, rowsTotal * 2);
}
	
